/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef CALCULATEPOLAR_SRC_METHODS_LIFTINGLINEPOLAR_H_
#define CALCULATEPOLAR_SRC_METHODS_LIFTINGLINEPOLAR_H_

#include "../flightCondition.h"

#include <moduleBasics/runtimeIO.h>
#include <liftingLineInterface2/liftingLineInterface.h>
#include <liftingLineInterface2/liftingLinePolar.h>
#include <aircraftGeometry2/airfoil_surface.h>

#include <memory>
#include <vector>
#include <string>

class LiLiPolarInitialization {
 public:
    double bestCruiseCL;
    size_t cruiseMachID;
    double initialMachCruise;
    double stepWidthCL;
    double delta_CM0;
    double delta_dCMdCL;
    double indDragCtCorrForCalibration;
    double factorDragCleanPolar;
    std::string CLModeAirfoilSelection; /**< aircraft CL for airfoil selection, 0: CL at ICA, 1: CL at TOC, 2: CL avg cr */

    std::vector <flightCondition> flightConditions;

    std::vector <geom2::MultisectionSurface<geom2::AirfoilSection>> liftingSurfaces;
    LiLiPolarInitialization();
    virtual ~LiLiPolarInitialization() {}
};

class liftingLinePolar{
 public:
    // config variables
    std::string config;
    std::filesystem::path airfoilPath;
    std::string IOFilePath;
    std::string pathToLiftingLine;
    std::string IOFilename;
    // auxillary variables
    std::string CLModeAirfoilSelection; /**< aircraft CL for airfoil selection, 0: CL at ICA, 1: CL at TOC, 2: CL avg cr */
    size_t cruiseMachID;
    double stepWidthCL;
    double initialMachCruise;
    double designCL;
    double bestCruiseCL;
    double indDragCtCorrForCalibration;
    double factorDragCleanPolar;
    double currentIStabSetting;
    double delta_CM0;
    double delta_dCMdCL;
    // geometry
    class aeroSection {
     public:
        /* Constructor and Destructor */
        /** \brief Constructor of aeroSection class
         *  \param &anAcft Reference to aircraft object
         *  \param &y_sec_i Reference to y position of aero section [m]
         *  \param &l_sec_i Reference to length of aero section [m]
         *  \param &s_sec_i Reference to width of aero section [m]
         *  \param &filename_i Reference to airfoil name
         *  \param &anID Reference to identification number within the vector the aero section is stored in
         */
        aeroSection(const geom2::MultisectionSurface<geom2::AirfoilSection> liftingSurface, const double& y_sec_i, const double& l_sec_i,
                    const double& s_sec_i, const std::string& filename_i, const uint16_t& anID, const size_t numberOfFlightConditions);
        /** \brief Virtual destructor of aeroSection class
         */
        virtual ~aeroSection();
            /* Member variables */
        std::string airfoilFilename; /**< Filename of airfoil data [-] */
        int secID; /**< Identification number of areo section [-] */
        std::string airfoilFilename_new; /**< Filename of updated airfoil data [-] */
        bool writePlotData;
        bool rootAirfoil;
        uint8_t theSurfaceType;
        std::vector <std::string> csvFiles_polar_plt_phi;
        std::vector <std::string> csvFiles_airfoil_plt_phi;
        std::vector<double> Cl_max; /**< Vector of maximum lift coefficients of aero section for each single Mach number [-] */

        /* Geometry parameters */
        double y_sec; /**< y-position of the aero section [m] */
        double l_sec; /**< Length of the aero section [m] */
        double s_sec; /**< Width of the aero section [m] */
        double area; /**< Surface area of the aero section [m^2] */

     private:
        /* Constants */
    };
    class LiLiPolarPoint{
     public:
        double AoA; /**< Angle of attack [deg]*/
        double CL;  /**< Lift coefficient [-] */
        double CDind; /**< Induced drag coefficient [-] */
        double CDtotal; /**< Total drag coefficient [-] */
        double CM; /**< Moment coefficient [-]*/
        double LoverD; /**< Lift-to-drag ratio [-] */
        double iStabPolar; /**< current iStab setting[-] */
        LiLiPolarPoint();
        /** \brief Virtual destructor of polarPointLiftSurface class
         */
        virtual ~LiLiPolarPoint() {}
    };

    class LiftDistributionPolarPoint: public LiLiPolarPoint{
     public:
        std::vector <double> yStations;     /**< Vector of y stations of lifting surface [m] */
        std::vector <double> Cl_distr;      /**< Vector of local lift coefficients [-] */
        std::vector <double> Lift_distr;    /**< Vector of lift distribution calculated by multiplying local lift coefficient with the corresponding cord length [-] */
        std::vector <double> y_trans;       /**< Vector of flow transition positions corresponding to the yStations vector [?] */
        std::vector <double> x2c_trans;                 //!< local transition location ???
        std::vector <double> Cd_visc;       /**< Vector of local viscous drag coefficients [-]*/
        std::vector <double> Cd_visc_corr;  /**< Vector of local viscous drag coefficient correction factors [-]*/
        std::vector <double> Cd_wave;       /**< Vector of local wave drag coefficients [-] */
        std::vector <double> Cd_friction;   /**< Vector of local friction drag coefficients [-] */
        std::vector <double> Cd_pressure;   /**< Vector of local pressure drag coefficients [-] */

        /** \brief Constructor of polarPointLiftSurface class
         */
        LiftDistributionPolarPoint();
        /** \brief Virtual destructor of polarPointLiftSurface class
         */
        virtual ~LiftDistributionPolarPoint() {}
    };

    class LiLiPolar {
     public:
        std::string Config;             /**< Vector of calculated mission section configurations [-] */
        double MachNumber;              /**< Mach number [-]*/
        double altitude;                /**< Altitude at the polar point [m] */
        double ReynoldsNumber;          /**< Reynold's number [-] */
        double CLmax;                   /**< Maximum lift coefficient [-] */
        double LoverDmax;               /**< Maximum lift-to-drag ratio [-] */
        double extrapolationMargin; /**< extrapolationMargin */
        bool allowGridChange; /**< Switch whether to allow Mach number grid change when processing final polar data */
        std::vector<LiLiPolarPoint> Points;                 /**< Vector of polarPoint objects representing all points on the polar */
        std::vector <std::vector <LiftDistributionPolarPoint>> liftingSurfacePolars;  /**< Vector over all lifting surfaces, calculated in LL*/
        LiLiPolar();
        virtual ~LiLiPolar() {}
    };
    struct analyticalIncompPolarData {
    std::vector <double> CDmin_inc;
    std::vector <double> kFactor_inc;
    std::vector <double> CLatCDmin_inc;
    std::vector <double> dCMtodCL_inc;
    std::vector <double> CMatCL0_inc;
    std::vector <double> dCMtodAoA;
    std::vector <double> dCLtodAoA;
    std::vector <double> CLatAoA0;
    std::vector <double> CMatAoA0;
    };
    analyticalIncompPolarData theAnalyticalPolarData;
    std::vector <LiftDistributionPolarPoint> aLiftingSurfacePolar;
    std::vector <flightCondition> flightConditions;
    // methods
    void run(const Point& aCoG);
    void buildWing();
    std::vector <std::vector <double> > liftCoefficients;
    liftingLineInterface runLiftingLine(const Point& theCoG, const double& machNumber, const double& altitude);
    void interpolate_LL_results();
    double PrandtlGlauert_interpol(double targetMach, double y_incompr, double Ma_compr, double y_compr);
    void setLiftCoefficientVector();
    double getCL_startFromLiftingLineResults();
    double getCL_endFromLiftingLineResults(const double& aMachNumber);
    double getCLmax_clean(double Mach);
    void setWingDesignPoints();
    void interpolateLiftDistribution(double aLiftCoefficient, const liftingLinePolars::liftingLinePolarWings& aLLwingPolarObject, int machID, int aWingID,
                                           LiftDistributionPolarPoint* aPolarPoint) const;
    void setAeroSections();
    void setAeroSectionsAutomatically();
    void setSectionAreas(std::vector<aeroSection>* theAeroSections);
    std::vector <aeroSection> setGeometricalAeroSections();
    std::vector <aeroSection> setLiftingLineAeroSections();
    std::vector <aeroSection> combineGeometricalAndLiftingLineSections(const std::vector<aeroSection>& theGeometricalAeroSections,
                                                                             const std::vector<aeroSection>& theLiftingLineAeroSections);
    void getCleanPolar();
    void calculateCleanPolarForMach(const double& Mach, const uint16_t Mach_id);
    void calculateInducedDrag(const uint16_t& machID);
    void getMomentCoefficient(const uint16_t& machID);
    // lili stuff
    liftingLineInterface myLiftingLine; /**< liftingLineInterface object containing Lifting Line calculation results */
    liftingLineInterface myLiftingLine_compr; /**< liftingLineInterface object containing Lifting Line calculation results (compressible fluid) */
    std::vector <liftingLineInterface> myLL_vec;  /**< vector of liftingLineinterfaces for the different flight conditions */
    std::string LILI_src_dir; /**< Path to lifting line source directory */
    std::string LILI_bkp_dir; /**< Path to lifting line backup directory */
    std::string LILI_version;
    // aerosections
    std::vector <aeroSection> aeroSections_airfoilSel; /**< Vector of geometrical aero sections */
    std::vector <aeroSection> aeroSections_all; /**< Vector of combined geometrical and Lifting Line aero sections */
    // unclear (for now)
    LiftDistributionPolarPoint *wingPointBestCruiseCL; /**< polarPointLiftSurface object containing data on polar components for best cruise lift coefficient configuration */
    LiftDistributionPolarPoint *wingDesignPoint; /**< polarPointLiftSurface object containing data on lift- and drag distribution for design configuration */
    std::vector <LiftDistributionPolarPoint> liftDistributionPolar;

    std::vector <double> liftDistr_design; /**< Lift coefficient distribution for design configuration [-] */
    std::vector <double> yPositionsLL;  /**< Vector of y positions used in Lifting Line calculations (subdivision of lifting surfaces) [m] */
    /* Objects and functions for lifting line */
    explicit liftingLinePolar(const std::shared_ptr<RuntimeIO>& rtIO, const std::string& liftingLinePath,
                                const LiLiPolarInitialization& LiLiPolarInit);
    virtual ~liftingLinePolar() {}
    // Polars
    LiLiPolar aPolar();
    std::vector <LiLiPolar> theCleanPolars;
    /* Objects and functions for lifting line */
    // liftingLineInterface myLiftingLine; /**< liftingLineInterface object containing Lifting Line calculation results */
    // liftingLineInterface myLiftingLine_compr; /**< liftingLineInterface object containing Lifting Line calculation results (compressible fluid) */
    // std::vector <liftingLineInterface> myLL_vec;
    /* LILI error handling */
    // std::string LILI_src_dir; /**< Path to lifting line source directory */
    // std::string LILI_bkp_dir; /**< Path to lifting line backup directory */
    // std::vector <double> yPositionsLL;  /**< Vector of y positions used in Lifting Line calculations (subdivision of lifting surfaces) [m] */

    /* Aero-Sections */
    /** \todo [T792] (https://unicado.ilr.rwth-aachen.de/T792) create these sections in function setAeroSections() and fill there
     * \warning This vector seems like is never used. Perhaps should be removed. Because it's dead code
     */
    // std::vector <aeroSection> aeroSections_HL;



 private:
    const std::vector <geom2::MultisectionSurface<geom2::AirfoilSection>>& liftingSurfaces;
    const std::shared_ptr<RuntimeIO>& rtIO_;
};

#endif  // CALCULATEPOLAR_SRC_METHODS_LIFTINGLINEPOLAR_H_
