/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "tawDefaultData.h"

#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/module.h>

#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <vector>
#include <memory>
#include <utility>

using std::endl;

tawDefaultData::tawDefaultData(const std::shared_ptr<RuntimeIO>& rtIO)
    :
    rtIO_(rtIO),
    engineType(EndnodeReadOnly<std::string>("type").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/propulsor")).value()),
    initialCruiseMachNumber(EndnodeReadOnly<double>("initial_cruise_mach_number").read(
        rtIO->acxml.at("requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission")).value()),
    bestCruiseCL(EndnodeReadOnly<double>("analysis/aerodynamics/lift_coefficients/C_LoptimumCruise").read(rtIO->acxml).value()),
    designCoGX(EndnodeReadOnly<double>("analysis/masses_cg_inertia/design_mass/mass_properties/center_of_gravity/x").read(rtIO->acxml).value()),
    designCoGY(EndnodeReadOnly<double>("analysis/masses_cg_inertia/design_mass/mass_properties/center_of_gravity/y").read(rtIO->acxml).value()),
    designCoGZ(EndnodeReadOnly<double>("analysis/masses_cg_inertia/design_mass/mass_properties/center_of_gravity/z").read(rtIO->acxml).value()),
    forwardCoGX(EndnodeReadOnly<double>("analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/x").read(rtIO->acxml).value()),
    forwardCoGY(EndnodeReadOnly<double>("analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/y").read(rtIO->acxml).value()),
    forwardCoGZ(EndnodeReadOnly<double>("analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/z").read(rtIO->acxml).value()),
    aftCoGX(EndnodeReadOnly<double>("analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/x").read(rtIO->acxml).value()),
    aftCoGY(EndnodeReadOnly<double>("analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/y").read(rtIO->acxml).value()),
    aftCoGZ(EndnodeReadOnly<double>("analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/z").read(rtIO->acxml).value()),
    MTOM_CoG_x(EndnodeReadOnly<double>("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/center_of_gravity/x").read(rtIO->acxml).value()),
    MTOM_CoG_y(EndnodeReadOnly<double>("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/center_of_gravity/y").read(rtIO->acxml).value()),
    MTOM_CoG_z(EndnodeReadOnly<double>("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/center_of_gravity/z").read(rtIO->acxml).value()),
    OME_CoG_x(EndnodeReadOnly<double>("analysis/masses_cg_inertia/operating_mass_empty/mass_properties/center_of_gravity/x").read(rtIO->acxml).value()),
    OME_CoG_y(EndnodeReadOnly<double>("analysis/masses_cg_inertia/operating_mass_empty/mass_properties/center_of_gravity/y").read(rtIO->acxml).value()),
    OME_CoG_z(EndnodeReadOnly<double>("analysis/masses_cg_inertia/operating_mass_empty/mass_properties/center_of_gravity/z").read(rtIO->acxml).value()),
    C_L_optimumCruise(Endnode<double>("analysis/aerodynamics/lift_coefficients/C_LoptimumCruise", "C_LoptimumCruise", 0, "1")),
    CL_max_landing(Endnode<double>("analysis/aerodynamics/lift_coefficients/C_LmaxLanding", "C_LmaxLanding", 0, "1")),
    C_LmaxTO(Endnode<double>("analysis/aerodynamics/lift_coefficients/C_LmaxT-O", "C_LmaxTO", 0, "1")),
    C_LgroundRoll(Endnode<double>("analysis/aerodynamics/lift_coefficients/C_LgroundRoll", "C_LgroundRoll", 0, "1")),
    span(Endnode<double>("analysis/aerodynamics/reference_values/b", "Wing half span", 0, "m")),
    MAC(Endnode<double>("analysis/aerodynamics/reference_values/MAC", "Mean aerodynamic chord", 0, "m")),
    S_ref(Endnode<double>("analysis/aerodynamics/reference_values/S_ref", "Wing reference area", 0, "m^2")),
    neutral_point_xCoordinate(Endnode<double>("analysis/aerodynamics/reference_values/neutral_point/x", "x component of neutral point", 0., "m")),
    max_spoiler_factor(Endnode<double>("analysis/aerodynamics/max_spoiler_factor", "max_spoiler_factor", 0, "1")),
    polar_file(Endnode<std::string>("analysis/aerodynamics/polar/polar_file", "Name of polar file")) {
    // check for reasonable values for the CoG values
    if (accuracyCheck(designCoGX, 0., ACCURACY_HIGH)) {
        designCoGX = 16.0;
    }
    if (accuracyCheck(forwardCoGX, 0., ACCURACY_HIGH)) {
        forwardCoGX = 15.0;
    }
    if (accuracyCheck(aftCoGX, 0., ACCURACY_HIGH)) {
        aftCoGX = 17.0;
    }
    if (accuracyCheck(MTOM_CoG_x, 0., ACCURACY_HIGH)) {
        MTOM_CoG_x = 16.0;
    }
    if (accuracyCheck(OME_CoG_x, 0., ACCURACY_HIGH)) {
        OME_CoG_x = 16.0;
    }
    if (accuracyCheck(bestCruiseCL, 0., ACCURACY_HIGH)) {
        bestCruiseCL = 0.5;
    }
    designCoG.xCoordinate = designCoGX;
    designCoG.yCoordinate = designCoGY;
    designCoG.zCoordinate = designCoGZ;
    forwardCoG.xCoordinate = forwardCoGX;
    forwardCoG.yCoordinate = forwardCoGY;
    forwardCoG.zCoordinate = forwardCoGZ;
    aftCoG.xCoordinate = aftCoGX;
    aftCoG.yCoordinate = aftCoGY;
    aftCoG.zCoordinate = aftCoGZ;
    MTOM_CoG.xCoordinate = MTOM_CoG_x;
    MTOM_CoG.yCoordinate = MTOM_CoG_y;
    MTOM_CoG.zCoordinate = MTOM_CoG_z;
    OME_CoG.xCoordinate = OME_CoG_x;
    OME_CoG.yCoordinate = OME_CoG_y;
    OME_CoG.zCoordinate = OME_CoG_z;
    numberDevices = rtIO_->acxml.getVector(("wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device"), 3).size();
    // read in device names to later distinguish between control and high lift device
    for (size_t deviceID(0); deviceID < numberDevices; ++deviceID) {
        wingDevice aWingDevice;
        aWingDevice.type =
                    std::string(rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                    + num2Str(deviceID)).getStringAttrib("description"));
        aWingDevice.maxNegativeDeflection =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/deflection/full_negative_deflection/value");
        aWingDevice.maxPositiveDeflection =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/deflection/full_positive_deflection/value");
        aWingDevice.innerRelSpanPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/inner_position/spanwise/value");
        aWingDevice.innerRelFwdChordPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/inner_position/chord/from/value");
        aWingDevice.innerRelAftChordPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/inner_position/chord/to/value");
        aWingDevice.outerRelSpanPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/outer_position/spanwise/value");
        aWingDevice.outerRelFwdChordPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/outer_position/chord/from/value");
        aWingDevice.outerRelAftChordPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/outer_position/chord/to/value");
        wingDevices.push_back(aWingDevice);
    }
}

void tawDefaultData::updateNodes(std::string mode) {
    // reference values
    span.update(rtIO_->acxml);
    MAC.update(rtIO_->acxml);
    S_ref.update(rtIO_->acxml);
    neutral_point_xCoordinate.update(rtIO_->acxml);
    if (mode == "mode_0") {
        // lift coefficients
        C_L_optimumCruise.update(rtIO_->acxml);
        CL_max_landing.update(rtIO_->acxml);
        C_LmaxTO.update(rtIO_->acxml);
        C_LgroundRoll.update(rtIO_->acxml);
        // maximum spoiler factor
        max_spoiler_factor.update(rtIO_->acxml);
        // polar file stuff
        polar_file.update(rtIO_->acxml);
    }
}
