/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "tawDefaultStrategy.h"

#include <aircraftGeometry2/fuselage.h>
#include <aircraftGeometry2/hull_surface.h>
#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>
#include <unitConversion/unitConversion.h>
#include <CGAL/Surface_mesh/IO/PLY.h>
#include <moduleBasics/module.h>

#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <memory>
#include <vector>
#include <filesystem>
#include <utility>  // make_pair

#include "../polar.h"
#include "../methods/generalMethods/auxFunctions.h"
#include "../methods/liftingLinePolar.h"
#include "../methods/liftingLineForTAW.h"
#include "../methods/viscDragRaymer.h"
#include "../methods/waveDragMason.h"
#include "../methods/semiEmpiricalHighLiftAdaption.h"
#include "../methods/trimInterpolation.h"

using std::endl;
using std::exception;
using std::fixed;
using std::ofstream;
using std::setprecision;
using std::string;
using std::stringstream;
using std::pair;
using std::vector;

/*  TODO:   - check if acft xml has entries before building the geometry comopnents */
tawDefaultStrategy::tawDefaultStrategy(const std::shared_ptr<RuntimeIO>& rtIO, const std::shared_ptr<tawCalculatePolarConfig>& config)
    :
    CoGnames{"design", "forward", "aft"},
    HLconfigs{"takeoff", "takeoff_landing_gear_retracted", "climb", "approach", "approach_landing_gear_out", "landing"},
    C_LmaxLanding(0.0),
    C_LmaxTO(0.0),
    C_LoptimumCruise(0.0),
    rtIO_(rtIO),
    data_(std::make_shared<tawDefaultData>(rtIO_)),
    config_(config) {
}

void tawDefaultStrategy::run() {
    // define center of gravity positions for trim calculations
    Point Design_CoG_Position;
    Design_CoG_Position.xCoordinate = config_->design_trim_MTOM_fraction * data_->MTOM_CoG.xCoordinate + (1 - config_->design_trim_MTOM_fraction) * data_->MTOM_CoG.xCoordinate;
    Design_CoG_Position.yCoordinate = config_->design_trim_MTOM_fraction * data_->MTOM_CoG.yCoordinate + (1 - config_->design_trim_MTOM_fraction) * data_->MTOM_CoG.yCoordinate;
    Design_CoG_Position.zCoordinate = config_->design_trim_MTOM_fraction * data_->MTOM_CoG.zCoordinate + (1 - config_->design_trim_MTOM_fraction) * data_->MTOM_CoG.zCoordinate;
    CoGPositions["design"] = Design_CoG_Position;
    if (config_->CGPositions.size() > 1) {
        CoGPositions["forward"] = data_->forwardCoG;
        CoGPositions["aft"] = data_->aftCoG;
    }

    // loop over all cg positions and stabilizer angles
    for (size_t CGID(0); CGID < config_->CGPositions.size(); ++CGID) {
        for (size_t stabID(0); stabID < config_->stabAngles.size(); ++stabID) {
            this->clearDataForNextCalculationRun();
            myRuntimeInfo-> out << endl;
            myRuntimeInfo->out << "Run aero calculations for " << config_->CGPositions.at(CGID) <<
                            " CoG and stabilizer angle of " << config_->stabAngles.at(stabID) << " deg..." << endl;
            this->createAircraftGeometry(config_->stabAngles.at(stabID));
            this->runAeroCalculations(CoGPositions.at(config_->CGPositions.at(CGID)), config_->stabAngles.at(stabID));
            this->copyPolarsToCGMap(CGID, stabID);
        }
    }
    // interpolate trimed polars, or just copy polars if trim is disabled
    if (config_->doTrim.value()) {
        this->doTrimInterpolation();
        if (config_->trimHighLiftPolars.value()) {
            this->doTrimInterpolationHighLift();
        } else {
            this->copyHighLiftPolars();
        }
    } else {
        this->copyCleanPolars();
        this->copyHighLiftPolars();
    }
    this->calibratePolars();
}

void tawDefaultStrategy::createAircraftGeometry(const double stabAngle) {
    auto get_position = [](const node &component) -> geom2::Point_3 {
        return geom2::Point_3{double{component.at("position/x/value")},
                              double{component.at("position/y/value")},
                              double{component.at("position/z/value")}};
    };
    std::filesystem::path file{rtIO_->acxml.getName()};
    std::shared_ptr<node> AcXML = aixml::openDocument(file);
    // create wing
    this->aircraft_xml = aixml::openDocument(rtIO_->acxmlAccess);
    const auto wing_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/wing"));
    geom2::WingFactory wingFactory{AcXML, rtIO_->getAirfoilDataDir()};
    geom2::MultisectionSurface<geom2::AirfoilSection> mainWing = wingFactory.create("wing/specific/geometry/aerodynamic_surface@0");
    auto offset = wing_position - CGAL::ORIGIN;
    mainWing.origin += offset;
    liftingSurfaces.push_back(mainWing);
    // create HTP
    const auto htp_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/empennage"));
    geom2::WingFactory HTPFactory{AcXML, rtIO_->getAirfoilDataDir()};
    geom2::MultisectionSurface<geom2::AirfoilSection> HTP = HTPFactory.create("empennage/specific/geometry/aerodynamic_surface@1");
    offset = htp_position - CGAL::ORIGIN;
    HTP.origin += offset;
    double designStabAngle = HTP.sections.front().get_twist_angle();
    HTP.sections.front().set_twist_angle(designStabAngle + convertUnit(DEGREE, RADIAN, stabAngle));
    liftingSurfaces.push_back(HTP);
    // create fuselage
    geom2::FuselageFactory fuselageFactory{AcXML, rtIO_->getGeometryDir()};
    fuselage = fuselageFactory.create("fuselage/specific/geometry/fuselage@0");
    // create VTP
    const auto vtp_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/empennage"));
    geom2::WingFactory VTPFactory{AcXML, rtIO_->getAirfoilDataDir()};
    VTP = VTPFactory.create("empennage/specific/geometry/aerodynamic_surface@0");
    offset = vtp_position - CGAL::ORIGIN;
    VTP.origin += offset;
    geom2::Mesh wing_mesh = geom2::transform::to_mesh(mainWing);
    geom2::Mesh vtp_mesh = geom2::transform::to_mesh(VTP);
    vtp_mesh += wing_mesh;
    geom2::Mesh htp_mesh = geom2::transform::to_mesh(HTP);
    htp_mesh += vtp_mesh;
    geom2::Mesh fuselage_mesh = geom2::transform::to_mesh(fuselage);
    fuselage_mesh += htp_mesh;
    // create nacelles
    const auto nacelle_position = get_position(this->aircraft_xml->at("aircraft_exchange_file/component_design/propulsion"));
    offset = nacelle_position - CGAL::ORIGIN;
    size_t numberOfPropulsors = rtIO_->acxml.getVector(("aircraft_exchange_file/component_design/propulsion/specific/propulsion"), 1).size();
    geom2::HullFactory nacelleFactory{AcXML, rtIO_->getGeometryDir()};
    for (size_t i(0); i < numberOfPropulsors; ++i) {
        geom2::MultisectionSurface<geom2::PolygonSection> aNacelle = nacelleFactory.create("propulsion/specific/propulsion@" + num2Str(i) + "/nacelle");
        aNacelle.origin += offset;
        geom2::Mesh aNacelle_mesh = geom2::transform::to_mesh(aNacelle);
        fuselage_mesh += aNacelle_mesh;
        nacelles.push_back(aNacelle);
    }
    CGAL::IO::write_STL("TAW_mesh.stl", fuselage_mesh);
}

void tawDefaultStrategy::runAeroCalculations(const Point CoGPosition, const double currentIStabAngle) {
    // lifting line for TAW
    LiLiforTAWInitialization myLiLiInit;
    this->initializeLiLiforTAW(&myLiLiInit);
    liftingLineForTAW* myLiftingLineForTAW = new liftingLineForTAW(rtIO_, config_->liftingLineFolderPath, myLiLiInit);
    myLiftingLineForTAW->run(CoGPosition);
    this->getPolarDataFromLiftingLine(*myLiftingLineForTAW);
    // viscous drag by raymer
    viscDragRaymerInitialization myViscDragInit;
    this->initializeViscDragRaymer(&myViscDragInit);
    viscDragRaymer* myViscDragRamyer = new viscDragRaymer(rtIO_, myViscDragInit);
    this->calculateViscDragRaymer(myViscDragRamyer);
    // wave drag by mason
    waveDragInitialization myWaveDragInit;
    this->initializeWaveDragMason(&myWaveDragInit);
    waveDragMason* myWaveDragMason = new waveDragMason(rtIO_, myWaveDragInit);
    this->calculateWaveDragMason(myWaveDragMason);
    // summation of the drag components
    this->sumUpDragComponents();
    // high lift calculation
    semiEmpiricalHighLiftAdaption* mySemiEmpHLAdaption = new semiEmpiricalHighLiftAdaption(rtIO_, liftingSurfaces.front());
    this->copyWingDevicesAndStabilizer(mySemiEmpHLAdaption);
    this->copyHLPolarCalibrationData(mySemiEmpHLAdaption);
    this->copyCleanPolarForHLCalculation(mySemiEmpHLAdaption, myLiftingLineForTAW);
    mySemiEmpHLAdaption->runHighLiftAdaption();
    this->calculateHighLiftPolars(mySemiEmpHLAdaption, CoGPosition.xCoordinate, currentIStabAngle);
}

void tawDefaultStrategy::initializeLiLiforTAW(LiLiforTAWInitialization *myLiLiInit) {
    myLiLiInit->CLModeAirfoilSelection = config_->CLModeAirfoilSelection.value();
    myLiLiInit->stepWidthCL = config_->stepWidthCL.value();
    myLiLiInit->CM_corr_fuselage_mode = config_->CM_corr_fuselage_mode.value();
    myLiLiInit->CM_corr_nacelle_mode = config_->CM_corr_nacelle_mode.value();
    myLiLiInit->delta_CM0 = config_->delta_CM0.value();
    myLiLiInit->delta_dCMdCL = config_->delta_dCMdCL.value();
    myLiLiInit->indDragCtCorrForCalibration = config_->indDragCtCorrForCalibration.value();
    myLiLiInit->factorDragCleanPolar = config_->factorIndDragCleanPolar.value();
    myLiLiInit->bestCruiseCL = data_->bestCruiseCL;
    myLiLiInit->initialMachCruise = data_->initialCruiseMachNumber;
    myLiLiInit->engineType = data_->engineType;
    for (size_t flightConditionID(0); flightConditionID < config_->flightConditions.size(); ++flightConditionID) {
        flightCondition* aLiLiFlightCondition = new flightCondition;
        aLiLiFlightCondition->Mach = config_->flightConditions.at(flightConditionID).Mach;
        aLiLiFlightCondition->Altitude = config_->flightConditions.at(flightConditionID).Altitude;
        aLiLiFlightCondition->extrapolationMargin = config_->flightConditions.at(flightConditionID).extrapolationMargin;
        aLiLiFlightCondition->allowGridChange = config_->flightConditions.at(flightConditionID).allowGridChange;
        myLiLiInit->flightConditions.push_back(*aLiLiFlightCondition);
        if (accuracyCheck(aLiLiFlightCondition->Mach, data_->initialCruiseMachNumber, ACCURACY_HIGH)) {
            myLiLiInit->cruiseMachID = flightConditionID;
        }
        delete aLiLiFlightCondition;
    }
    myLiLiInit->liftingSurfaces = liftingSurfaces;
    myLiLiInit->theFuselage = fuselage;
    myLiLiInit->theNacelles = nacelles;
}

void tawDefaultStrategy::initializeViscDragRaymer(viscDragRaymerInitialization *myViscDragInit) {
    // copy geometry
    myViscDragInit->theWing = liftingSurfaces.at(0);
    myViscDragInit->calculateWing = true;
    myViscDragInit->theHTP = liftingSurfaces.at(1);
    myViscDragInit->calculateHTP = true;
    myViscDragInit->theVTP = VTP;
    myViscDragInit->calculateVTP = true;
    myViscDragInit->theFuselage = fuselage;
    myViscDragInit->calculateFuselage = true;
    myViscDragInit->theNacelles = nacelles;
    myViscDragInit->calculateNacelles = true;
    myViscDragInit->calculateMisc = true;
    // initialize transition parameters
    myViscDragInit->manualTransitionLocation = config_->manualTransition.value();
    myViscDragInit->transitionLocationWing = config_->manualTransitionWing.value();
    myViscDragInit->transitionLocationHTP = config_->manualTransitionStabilizer.value();
    myViscDragInit->transitionLocationVTP = config_->manualTransitionFin.value();
    // initialize calibration parameters
    myViscDragInit->doCalibration = config_->doViscDragCalibration.value();
    myViscDragInit->CDvisc_CDsum = config_->CDvisc_CDsum.value();
    myViscDragInit->CDvisc_CLexp = config_->CDvisc_CLexp.value();
    myViscDragInit->CDvisc_CLfact = config_->CDvisc_CLfact.value();
    myViscDragInit->CDvisc_CDsum_lowMa = config_->CDvisc_CDsum_lowMa.value();
    myViscDragInit->CDvisc_CLexp_lowMa = config_->CDvisc_CLexp_lowMa.value();
    myViscDragInit->CDvisc_CLfact_lowMa = config_->CDvisc_CLfact_lowMa.value();
    // calculate parameters that remain static over all polars
}

void tawDefaultStrategy::initializeWaveDragMason(waveDragInitialization *myWaveDragInit) {
    // fill configuration parameters and values in the wave drag method
    myWaveDragInit->CLExp = config_->CDwave_CLExp.value();
    myWaveDragInit->CLFact = config_->CDwave_CLFact.value();
    myWaveDragInit->factorKa = config_->Ka.value();
    myWaveDragInit->initialMachCruise = data_->initialCruiseMachNumber;
    myWaveDragInit->numberOfSegments = config_->numberWingStrips.value();
    myWaveDragInit->waveDragCtCorrForCalibration = config_->deltaWaveDragCountWing.value();
    myWaveDragInit->maxAllowedSweep = convertUnit(RADIAN, DEGREE, config_->maximumSegmentSweep.value());
    myWaveDragInit->wFuselage = {geom2::measure::width_max(fuselage) * 0.5, 0};
    myWaveDragInit->liftingSurfaces = liftingSurfaces;
}

void tawDefaultStrategy::getPolarDataFromLiftingLine(const liftingLineForTAW& liftingLineForTAW) {
    for (size_t machID(0); machID < config_->flightConditions.size(); ++machID) {
        tawDefaultPolar aCleanPolar;
        aCleanPolar.extrapolationMargin = liftingLineForTAW.theCleanPolars.at(machID).extrapolationMargin;
        aCleanPolar.allowGridChange = liftingLineForTAW.theCleanPolars.at(machID).allowGridChange;
        aCleanPolar.altitude = liftingLineForTAW.theCleanPolars.at(machID).altitude;
        aCleanPolar.MachNumber = liftingLineForTAW.theCleanPolars.at(machID).MachNumber;
        aCleanPolar.ReynoldsNumber = liftingLineForTAW.theCleanPolars.at(machID).ReynoldsNumber;
        for (size_t pointID(0); pointID < liftingLineForTAW.theCleanPolars.at(machID).Points.size(); ++pointID) {
            aCleanPolar.Points.push_back(defaultPolarPoint());
            aCleanPolar.Points.back().CL = liftingLineForTAW.theCleanPolars.at(machID).Points.at(pointID).CL;
            aCleanPolar.Points.back().AoA = liftingLineForTAW.theCleanPolars.at(machID).Points.at(pointID).AoA;
            aCleanPolar.Points.back().CD_ind = liftingLineForTAW.theCleanPolars.at(machID).Points.at(pointID).CDind;
            aCleanPolar.Points.back().CM = liftingLineForTAW.theCleanPolars.at(machID).Points.at(pointID).CM;
            aCleanPolar.Points.back().iStabPolar = liftingLineForTAW.theCleanPolars.at(machID).Points.at(pointID).iStabPolar;
            aCleanPolar.PointsWing.push_back(defaultPolarPointLiftSurface());
            aCleanPolar.PointsWing.back().yStations = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).yStations;
            aCleanPolar.PointsWing.back().Cl_distr = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).Cl_distr;
            aCleanPolar.PointsWing.back().Lift_distr = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).Lift_distr;
            aCleanPolar.PointsWing.back().CL = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).CL;
            aCleanPolar.PointsWing.back().CD_ind = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).CDind;
            aCleanPolar.PointsStab.push_back(defaultPolarPointLiftSurface());
            aCleanPolar.PointsStab.back().yStations = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).yStations;
            aCleanPolar.PointsStab.back().Cl_distr = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).Cl_distr;
            aCleanPolar.PointsStab.back().Lift_distr = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).Lift_distr;
            aCleanPolar.PointsStab.back().CL = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).CL;
            aCleanPolar.PointsStab.back().CD_ind = liftingLineForTAW.theCleanPolars.at(machID).liftingSurfacePolars.at(1).at(pointID).CDind;
            aCleanPolar.PointsNacelles.push_back(defaultPolarPoint());
            aCleanPolar.PointsNacelles.back().CD_ind = liftingLineForTAW.theCleanPolars.at(machID).PointsNacelles.at(pointID).CDind;
            aCleanPolar.PointsFuselage.push_back(defaultPolarPoint());
            aCleanPolar.PointsFuselage.back().CD_ind = liftingLineForTAW.theCleanPolars.at(machID).PointsFuselage.at(pointID).CDind;
            aCleanPolar.PointsFin.push_back(defaultPolarPoint());
            aCleanPolar.PointsMisc.push_back(defaultPolarPoint());
        }
        currentPolarSet.cleanPolars.insert(std::make_pair(machID, aCleanPolar));
    }
    LILI_Version = liftingLineForTAW.LILI_version;
}

void tawDefaultStrategy::calculateViscDragRaymer(viscDragRaymer *myViscDragRaymer) {
    myViscDragRaymer->preprocessGeometry();
    // run viscous drag calculation by Raymer over the current polar set
    for (std::pair<size_t, tawDefaultPolar> polarIterator : currentPolarSet.cleanPolars) {
        size_t polarID = polarIterator.first;
        vector<double> theComponentDrags = myViscDragRaymer->calculateViscDrag(currentPolarSet.cleanPolars.at(polarID), 0.0, 0.0, 0.0);
        // loop over all points to calculate CL dependent visc drag correction
        for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(polarID).Points.size(); ++pointID) {
            vector <double> theCorrectedComponentDrags = myViscDragRaymer->calculateCorrectedViscDrag(theComponentDrags, currentPolarSet.cleanPolars.at(polarID).MachNumber,
                                                                                                        currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CL);
            // assign total and componentwise viscous drag to the polar map
            currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_visc = std::accumulate(theCorrectedComponentDrags.begin(), theCorrectedComponentDrags.end(), 0.0);
            currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).CD_visc = theCorrectedComponentDrags.at(0);
            currentPolarSet.cleanPolars.at(polarID).PointsStab.at(pointID).CD_visc = theCorrectedComponentDrags.at(1);
            currentPolarSet.cleanPolars.at(polarID).PointsFin.at(pointID).CD_visc = theCorrectedComponentDrags.at(2);
            currentPolarSet.cleanPolars.at(polarID).PointsFuselage.at(pointID).CD_visc = theCorrectedComponentDrags.at(3);
            currentPolarSet.cleanPolars.at(polarID).PointsNacelles.at(pointID).CD_visc = theCorrectedComponentDrags.at(4);
            currentPolarSet.cleanPolars.at(polarID).PointsMisc.at(pointID).CD_visc = theCorrectedComponentDrags.at(5);
        }
    }
    myRuntimeInfo->out<< "... Viscous drag has been calculated sucessfully!" << endl;
}

void tawDefaultStrategy::calculateWaveDragMason(waveDragMason *myWaveDragMason) {
    // run wave drag calculation by Mason over the current polar set
    for (size_t polarID(0); polarID < currentPolarSet.cleanPolars.size(); ++polarID) {
        double sweepPosition(0.0);
        if (currentPolarSet.cleanPolars.at(polarID).MachNumber < 0.7) {
            sweepPosition = 0.25;
        } else {
            sweepPosition = 0.5;
        }
        for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(polarID).Points.size(); ++pointID) {
            currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).CD_wave
                                            = myWaveDragMason->calculateWaveDrag(currentPolarSet.cleanPolars.at(polarID).Points.at(pointID),
                                                                                currentPolarSet.cleanPolars.at(polarID).MachNumber,
                                                                                0, sweepPosition, currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).yStations,
                                                                                currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).Cl_distr, true);
            currentPolarSet.cleanPolars.at(polarID).PointsStab.at(pointID).CD_wave
                                            = myWaveDragMason->calculateWaveDrag(currentPolarSet.cleanPolars.at(polarID).Points.at(pointID),
                                                                                currentPolarSet.cleanPolars.at(polarID).MachNumber,
                                                                                1, sweepPosition, currentPolarSet.cleanPolars.at(polarID).PointsStab.at(pointID).yStations,
                                                                                currentPolarSet.cleanPolars.at(polarID).PointsStab.at(pointID).Cl_distr, false);
            currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_wave = currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).CD_wave
                                                                               + currentPolarSet.cleanPolars.at(polarID).PointsStab.at(pointID).CD_wave;
        }
    }
    myRuntimeInfo->out << "... Wave drag has been calculated sucessfully!" << endl;
}

void tawDefaultStrategy::sumUpDragComponents() {
    myRuntimeInfo->out << "Sum up drag components" << endl;
    for (size_t polarID(0); polarID < currentPolarSet.cleanPolars.size(); ++polarID) {
        for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(polarID).Points.size(); ++pointID) {
            currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD =  currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_ind
                                                                          +  currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_visc
                                                                          +  currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_wave;
        }
    }
}

void tawDefaultStrategy::copyWingDevicesAndStabilizer(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption) {
    myRuntimeInfo->debug << "Copy wing devices" << endl;
    for (size_t deviceID(0); deviceID < data_->numberDevices; ++deviceID) {
        semiEmpiricalHighLiftAdaption::WingDevice* aWingDevice = new semiEmpiricalHighLiftAdaption::WingDevice;
        aWingDevice->name = data_->wingDevices.at(deviceID).type;
        aWingDevice->innerRelFwdChordPosition = data_->wingDevices.at(deviceID).innerRelFwdChordPosition;
        aWingDevice->innerRelAftChordPosition = data_->wingDevices.at(deviceID).innerRelAftChordPosition;
        aWingDevice->outerRelFwdChordPosition = data_->wingDevices.at(deviceID).outerRelFwdChordPosition;
        aWingDevice->outerRelAftChordPosition = data_->wingDevices.at(deviceID).outerRelAftChordPosition;
        aWingDevice->innerSpan_rel = data_->wingDevices.at(deviceID).innerRelSpanPosition;
        aWingDevice->outerSpan_rel = data_->wingDevices.at(deviceID).outerRelSpanPosition;
        mySemiEmpHLAdaption->wingDevices.push_back(*aWingDevice);
    }
    // stabilizer has to be optional
    bool stabilizer = true;
    if (stabilizer) {
    mySemiEmpHLAdaption->hasStabilizer = true;
    mySemiEmpHLAdaption->theStabilizer = liftingSurfaces.at(1);
    }
}

void tawDefaultStrategy::copyHLPolarCalibrationData(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption) {
    myRuntimeInfo->debug << "Copy high lift polar calibration data" << endl;
    for (size_t configID(0); configID < HLconfigs.size(); ++configID) {
        mySemiEmpHLAdaption->hl_polar_calibration_data[HLconfigs.at(configID)] = mySemiEmpHLAdaption->aHLPolarCalibrationContainer;
        mySemiEmpHLAdaption->hl_polar_calibration_data.at(HLconfigs.at(configID)).CL_max = config_->hl_polar_calibration_data.at(HLconfigs.at(configID)).CL_max;
        mySemiEmpHLAdaption->hl_polar_calibration_data.at(HLconfigs.at(configID)).delta_CD = config_->hl_polar_calibration_data.at(HLconfigs.at(configID)).delta_CD;
        mySemiEmpHLAdaption->hl_polar_calibration_data.at(HLconfigs.at(configID)).factor_CD = config_->hl_polar_calibration_data.at(HLconfigs.at(configID)).factor_CD;
    }
}

void tawDefaultStrategy::copyCleanPolarForHLCalculation(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption, liftingLineForTAW *myLiftingLineForTAW) {
    // find polar with minimum mach number
    double minMachNumber(1.0);
    size_t minMachID;
    for (std::pair<size_t, tawDefaultPolar> polarIterator : currentPolarSet.cleanPolars) {
        if (polarIterator.second.MachNumber < minMachNumber) {
            minMachID = polarIterator.first;
            minMachNumber = polarIterator.second.MachNumber;
        }
    }
    // copy clean polar to HL method
    mySemiEmpHLAdaption->theCleanPolar.MachNumber = currentPolarSet.cleanPolars.at(minMachID).MachNumber;
    mySemiEmpHLAdaption->theCleanPolar.altitude = currentPolarSet.cleanPolars.at(minMachID).altitude;
    mySemiEmpHLAdaption->theCleanPolar.extrapolationMargin = currentPolarSet.cleanPolars.at(minMachID).extrapolationMargin;
    mySemiEmpHLAdaption->theCleanPolar.allowGridChange = currentPolarSet.cleanPolars.at(minMachID).allowGridChange;
    mySemiEmpHLAdaption->theCleanPolar.ReynoldsNumber = currentPolarSet.cleanPolars.at(minMachID).ReynoldsNumber;
    mySemiEmpHLAdaption->theCleanPolar.Config = currentPolarSet.cleanPolars.at(minMachID).Config;
    for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(minMachID).Points.size(); ++pointID) {
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.push_back(PolarPoint());
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().AoA = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).AoA;
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().CL = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).CL;
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().CD = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).CD;
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().CM = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).CM;
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().iStabPolar = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).iStabPolar;
    }
    // copy other neccesary values from lifting line calculation
    mySemiEmpHLAdaption->do_individual_hl_calibration = config_->do_individual_hl_calibration;
    mySemiEmpHLAdaption->apply_manual_CLmax_hl_values = config_->apply_manual_CLmax_hl_values;
    mySemiEmpHLAdaption->CLmaxTO_req = 1.0;
    mySemiEmpHLAdaption->CLmaxLDG_req = 1.0;
    mySemiEmpHLAdaption->stepWidthCL = config_->stepWidthCL.value();
    mySemiEmpHLAdaption->reductionDragCountsHighLift = config_->deltaTotalDragHL.value();
    mySemiEmpHLAdaption->factorDragHighLift = config_->factorDragHL.value();
    mySemiEmpHLAdaption->CDmin_inc = myLiftingLineForTAW->theAnalyticalPolarData.CDmin_inc.at(0);
    mySemiEmpHLAdaption->kFactor_inc = myLiftingLineForTAW->theAnalyticalPolarData.kFactor_inc.at(0);
    mySemiEmpHLAdaption->CLatCDmin_inc = myLiftingLineForTAW->theAnalyticalPolarData.CLatCDmin_inc.at(0);
    mySemiEmpHLAdaption->dCMtodCL_inc = myLiftingLineForTAW->theAnalyticalPolarData.dCMtodCL_inc.at(0);
    mySemiEmpHLAdaption->CMatCL0_inc = myLiftingLineForTAW->theAnalyticalPolarData.CMatCL0_inc.at(0);
    mySemiEmpHLAdaption->dCMtodAoA = myLiftingLineForTAW->theAnalyticalPolarData.dCMtodAoA.at(0);
    mySemiEmpHLAdaption->dCLtodAoA = myLiftingLineForTAW->theAnalyticalPolarData.dCLtodAoA.at(0);
    mySemiEmpHLAdaption->CLatAoA0 = myLiftingLineForTAW->theAnalyticalPolarData.CLatAoA0.at(0);
    mySemiEmpHLAdaption->CMatAoA0 = myLiftingLineForTAW->theAnalyticalPolarData.CMatAoA0.at(0);
}

void tawDefaultStrategy::calculateHighLiftPolars(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption, const double CoG_X_Position, const double currentIStabAngle) {
    for (size_t configID(0); configID < HLconfigs.size(); ++configID) {
        myRuntimeInfo->out << "Calculate high lift polar for " << HLconfigs.at(configID) << endl;
        Polar aPolar = mySemiEmpHLAdaption->calculateHighLiftPolar(HLconfigs.at(configID), CoG_X_Position, currentIStabAngle);
        currentPolarSet.highLiftPolars[HLconfigs.at(configID)] = tawDefaultPolar();
        // copy polar data
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Config = HLconfigs.at(configID);
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).extrapolationMargin = aPolar.extrapolationMargin;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).allowGridChange = aPolar.allowGridChange;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).altitude = aPolar.altitude;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).MachNumber = aPolar.MachNumber;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).ReynoldsNumber = aPolar.ReynoldsNumber;
        for (size_t pointID(0); pointID < aPolar.PolarPoints.size(); ++pointID) {
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.push_back(defaultPolarPoint());
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().AoA = aPolar.PolarPoints.at(pointID).AoA;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().CL = aPolar.PolarPoints.at(pointID).CL;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().CD = aPolar.PolarPoints.at(pointID).CD;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().CM = aPolar.PolarPoints.at(pointID).CM;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().iStabPolar = aPolar.PolarPoints.at(pointID).iStabPolar;
        }
    }
}

void tawDefaultStrategy::copyPolarsToCGMap(const size_t CGID, const size_t stabID) {
    polarDataCGMap[CGID][stabID] = currentPolarSet;
}

void tawDefaultStrategy::clearDataForNextCalculationRun() {
    currentPolarSet.cleanPolars.clear();
    currentPolarSet.highLiftPolars.clear();
    liftingSurfaces.clear();
}

void tawDefaultStrategy::doTrimInterpolation() {
    for (size_t CGID(0); CGID < config_->CGPositions.size(); ++CGID) {
        vector<vector<trimInterpolation::interpolationPoint>> polarSet;
        // copy polar data into the generalized interpolation container
        for (size_t polarID(0); polarID < polarDataCGMap.at(CGID).at(0).cleanPolars.size(); ++polarID) {
            myRuntimeInfo->out << "Interpolate trim for clean polars at " << config_->CGPositions.at(CGID) << " CoG at Ma " << config_->machNumbers.at(polarID) << "..." << endl;
            for (size_t stabID(0); stabID < config_->stabAngles.size(); ++stabID) {
                vector<trimInterpolation::interpolationPoint> aStabPolar;
                for (size_t pointID(0); pointID < polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).Points.size(); ++pointID) {
                    trimInterpolation::interpolationPoint anInterpolationPoint;
                    anInterpolationPoint.CL = polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).Points.at(pointID).CL;
                    anInterpolationPoint.CM = polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).Points.at(pointID).CM;
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).Points.at(pointID).CD);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).Points.at(pointID).AoA);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).Points.at(pointID).iStabPolar);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).Points.at(pointID).CD_ind);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).Points.at(pointID).CD_visc);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).Points.at(pointID).CD_wave);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).PointsWing.at(pointID).CD_wave);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).PointsStab.at(pointID).CD_ind);
                    anInterpolationPoint.values.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).PointsStab.at(pointID).CD_wave);
                    // interpolate additional values for cruise mach polar
                    if (accuracyCheck(data_->initialCruiseMachNumber, polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).MachNumber, ACCURACY_HIGH)) {
                        anInterpolationPoint.vectorValues.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).PointsWing.at(pointID).Cl_distr);
                        anInterpolationPoint.vectorValues.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).PointsWing.at(pointID).Lift_distr);
                        anInterpolationPoint.vectorValues.push_back(polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(polarID).PointsWing.at(pointID).yStations);
                    }
                    aStabPolar.push_back(anInterpolationPoint);
                }
                polarSet.push_back(aStabPolar);
            }
            // interpolate the current polar set
            trimInterpolation* myTrimInterpolation = new trimInterpolation();
            vector<trimInterpolation::interpolationPoint> interpolatedTrimPolar = myTrimInterpolation->interpolateTrim(polarSet, true);
            polarSet.clear();
            // copy data to polarDataTrimMap
            polarDataTrimMap[CGID].cleanPolars[polarID].MachNumber = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).MachNumber;
            polarDataTrimMap.at(CGID).cleanPolars.at(polarID).ReynoldsNumber = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).ReynoldsNumber;
            polarDataTrimMap.at(CGID).cleanPolars.at(polarID).altitude = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).altitude;
            polarDataTrimMap.at(CGID).cleanPolars.at(polarID).extrapolationMargin = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).extrapolationMargin;
            polarDataTrimMap.at(CGID).cleanPolars.at(polarID).allowGridChange = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID).allowGridChange;
            for (size_t pointID(0); pointID < interpolatedTrimPolar.size(); ++pointID) {
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.push_back(defaultPolarPoint());
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsWing.push_back(defaultPolarPointLiftSurface());
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().CL = interpolatedTrimPolar.at(pointID).CL;
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().CM = interpolatedTrimPolar.at(pointID).CM;
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().CD = interpolatedTrimPolar.at(pointID).values.at(0);
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().AoA = interpolatedTrimPolar.at(pointID).values.at(1);
                polarDataTrimMap.at(CGID).cleanPolars.at(polarID).Points.back().iStabPolar = interpolatedTrimPolar.at(pointID).values.at(2);
                if (accuracyCheck(data_->initialCruiseMachNumber, polarDataTrimMap.at(CGID).cleanPolars.at(polarID).MachNumber, ACCURACY_HIGH)) {
                    polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsWing.back().Cl_distr = interpolatedTrimPolar.at(pointID).vectorValues.at(0);
                    polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsWing.back().Lift_distr = interpolatedTrimPolar.at(pointID).vectorValues.at(1);
                    polarDataTrimMap.at(CGID).cleanPolars.at(polarID).PointsWing.back().yStations = interpolatedTrimPolar.at(pointID).vectorValues.at(2);
                }
            }
        }
    }
    myRuntimeInfo->out << "All polars are trimmed!" << endl;
}

void tawDefaultStrategy::doTrimInterpolationHighLift() {
    // TBA
}

void tawDefaultStrategy::copyCleanPolars() {
    for (size_t CGID(0); CGID < config_->CGPositions.size(); ++CGID) {
        for (size_t polarID(0); polarID < polarDataCGMap.at(CGID).at(0).cleanPolars.size(); ++polarID) {
            polarDataTrimMap[CGID].cleanPolars[polarID] = polarDataCGMap.at(CGID).at(0).cleanPolars.at(polarID);
        }
    }
}

void tawDefaultStrategy::copyHighLiftPolars() {
    for (size_t CGID(0); CGID < config_->CGPositions.size(); ++CGID) {
        for (size_t configID(0); configID < HLconfigs.size(); ++configID) {
            polarDataTrimMap[CGID].highLiftPolars[HLconfigs.at(configID)] = polarDataCGMap.at(CGID).at(config_->neutralStabAngleID).highLiftPolars.at(HLconfigs.at(configID));
        }
    }
}

void tawDefaultStrategy::calibratePolars() {
    if (!(accuracyCheck(config_->factorDragCleanPolar.value(), 1.0, ACCURACY_HIGH) && accuracyCheck(config_->deltaTotalDragCleanPolar.value(), 0.0, ACCURACY_HIGH))) {
        calibrateCleanPolars();
    }
    if (!(accuracyCheck(config_->factorDragHighLiftPolar.value(), 1.0, ACCURACY_HIGH) && accuracyCheck(config_->deltaTotalDragHighLift.value(), 0.0, ACCURACY_HIGH))) {
        calibrateHighLiftPolars();
    }
}

void tawDefaultStrategy::calibrateCleanPolars() {
    myRuntimeInfo->out << "Apply calibration factors for clean polars ..." << endl;
    for (size_t CGID(0); CGID < polarDataTrimMap.size(); ++CGID) {
        for (size_t machID(0); machID < polarDataTrimMap.at(CGID).cleanPolars.size(); ++machID) {
            for (size_t pointID(0); pointID < polarDataTrimMap.at(CGID).cleanPolars.at(machID).Points.size(); ++pointID) {
                polarDataTrimMap.at(CGID).cleanPolars.at(machID).Points.at(pointID).CD =
                                polarDataTrimMap.at(CGID).cleanPolars.at(machID).Points.at(pointID).CD * config_->factorDragCleanPolar.value();
                polarDataTrimMap.at(CGID).cleanPolars.at(machID).Points.at(pointID).CD += config_->deltaTotalDragCleanPolar.value();
            }
        }
    }
}

void tawDefaultStrategy::calibrateHighLiftPolars() {
    myRuntimeInfo->out << "Apply calibration factors for high lift polars ..." << endl;
    for (size_t CGID(0); CGID < polarDataTrimMap.size(); ++CGID) {
        for (size_t configID(0); configID < HLconfigs.size(); ++configID) {
            for (size_t pointID(0); pointID < polarDataTrimMap.at(CGID).highLiftPolars.at(HLconfigs.at(configID)).Points.size(); ++pointID) {
                polarDataTrimMap.at(CGID).highLiftPolars.at(HLconfigs.at(configID)).Points.at(pointID).CD =
                                polarDataTrimMap.at(CGID).highLiftPolars.at(HLconfigs.at(configID)).Points.at(pointID).CD * config_->factorDragHighLiftPolar.value();
                polarDataTrimMap.at(CGID).highLiftPolars.at(HLconfigs.at(configID)).Points.at(pointID).CD += config_->deltaTotalDragHighLift.value();
            }
        }
    }
}

void tawDefaultStrategy::writePolarXML() {
    myRuntimeInfo->out << "Create XML-Polar file ..." << endl;
    // Construct the file path and convert to absolute path
    std::filesystem::path filePath = std::filesystem::absolute(this->rtIO_->getAeroDataDir() + "/");
    // Ensure the directory exists
    if (!std::filesystem::exists(filePath.parent_path())) {
        std::filesystem::create_directories(filePath.parent_path());
    }
    for (size_t CGID(0); CGID < polarDataTrimMap.size(); ++CGID) {
        polarXMLFilePath = rtIO_->getAeroDataDir() + "/";
        polarXMLFilePath += rtIO_->aircraft_model();
        if (!config_->doTrim.value()) {
            this->polarXMLFilePath += "_polar.xml";
        } else {
            this->polarXMLFilePath += "_" + config_->CGPositions.at(CGID) + "_trimmed_polar.xml";
        }
        // file name for design case and sizing loop
        if (CGID == 0) {
            this->designPolarXMLFilePath = polarXMLFilePath;
            this->designPolarXMLFilePath = designPolarXMLFilePath.substr(designPolarXMLFilePath.find_last_of("/") + 1);
        }
        ofstream XML;
        XML.open(polarXMLFilePath.c_str());
        if (!XML) {
            myRuntimeInfo->err << polarXMLFilePath + " could not be opened!" << endl;
            exit(1);
        } else {
            XML << "<?xml version=\"1.0\"?>" << endl;
            XML << "<polar-exchange-file version=\"1.0\">" << endl;
            XML << "    <comment>The 'PEF-Browser' application, a polar file converter is available in the software section at:";
            XML << " http://www.MH-AeroTools.de/airfoils/</comment>" << endl;
            XML << "    <liftingLineVersion>" << LILI_Version << "</liftingLineVersion>" << endl;
            XML << "    <dragComponentMethods>" << endl;
            XML << "        <inducedDrag>Lifting Line + Corrections for fuselage and nacelles</inducedDrag>" << endl;
            XML << "        <viscousDrag>" << "Raymer" << "</viscousDrag>" << endl;
            XML << "        <waveDrag>" << "Mason" << "</waveDrag>" << endl;
            XML << "    </dragComponentMethods>" << endl;
            XML << "    <configurations>" << endl;
            XML << "        <name>" << rtIO_->aircraft_model() << "</name>" << endl;
            XML << "        <comment>Polars for wing configuration</comment>" << endl;
            XML << "        <configuration ID=\"clean\">" << endl;
            XML << "            <name>Clean</name>" << endl;
            XML << "            <comment>Clean Configuration</comment>" << endl;
            XML << "            <polars>" << endl;
            for (size_t MachID(0); MachID < polarDataTrimMap.at(CGID).cleanPolars.size(); MachID++) {
                XML << "                <polar ID=\"" << MachID + 1 << "\" ExtrapolationMargin=\"" << polarDataTrimMap.at(CGID).cleanPolars.at(MachID).extrapolationMargin
                                    << "\" AllowGridChange=\"" << polarDataTrimMap.at(CGID).cleanPolars.at(MachID).allowGridChange <<"\">" << endl;
                XML << "                    <name>" << "---" << "</name>" << endl;
                XML << "                    <reynoldsnumber>" <<polarDataTrimMap.at(CGID).cleanPolars.at(MachID).ReynoldsNumber << "</reynoldsnumber>" << endl;
                XML << "                    <altitude>" <<polarDataTrimMap.at(CGID).cleanPolars.at(MachID).altitude << "</altitude>" << endl;
                XML << "                    <machnumber>" << polarDataTrimMap.at(CGID).cleanPolars.at(MachID).MachNumber << "</machnumber>" << endl;
                std::string Alpha_str(""), Cl_str(""), Cd_str(""), Cd_turb_str(""), Cm_str(""), origin_polar_str(""), i_stab_str("");
                for (size_t i(0); i < polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.size(); ++i) {
                    Alpha_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).AoA);
                    Cl_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CL);
                    Cd_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CD);
                    Cd_turb_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CD);
                    Cm_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CM);
                    origin_polar_str += "seg0-0.0";
                    i_stab_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).iStabPolar);
                    if (i < polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.size() - 1) {
                        Alpha_str += ";";
                        Cl_str += ";";
                        Cd_str += ";";
                        Cd_turb_str += ";";
                        Cm_str += ";";
                        origin_polar_str += ";";
                        i_stab_str += ";";
                    }
                }
                XML << "                    <Alpha>" << Alpha_str << "</Alpha>" << endl;
                XML << "                    <Cl>" << Cl_str << "</Cl>" << endl;
                XML << "                    <Cd>" <<Cd_str << "</Cd>" << endl;
                XML << "                    <Cd_turb>" << Cd_turb_str << "</Cd_turb>" << endl;
                XML << "                    <Cm>" << Cm_str << "</Cm>" << endl;
                XML << "                    <origin_polar>"<< origin_polar_str << "</origin_polar>" << endl;
                XML << "                    <iStab_polar>"<< i_stab_str << "</iStab_polar>" << endl;
                XML << "                </polar>" << endl;
            }
            XML << "            </polars>" << endl;
            XML << "        </configuration>" << endl;
            for (const pair<const string, tawDefaultPolar>& thePolarIterator : polarDataTrimMap.at(CGID).highLiftPolars) {
                XML << "        <configuration ID=\"" << thePolarIterator.first << "\">" << endl;
                XML << "            <name>" << thePolarIterator.first << "</name>" << endl;
                XML << "            <comment>" << thePolarIterator.first << " Configuration</comment>" << endl;
                XML << "            <polars>" << endl;
                // Inner config loop: Since the myAcftPt->highLiftPolar elements are ordered by config name, configs can be gathered here until the next one appears.
                XML << "                <polar ID=\"" << num2Str(1) << "\" ExtrapolationMargin =\"" << thePolarIterator.second.extrapolationMargin
                                        << "\" AllowGridChange=\"" << thePolarIterator.second.allowGridChange << "\">" << endl;
                XML << "                    <name>" << "---" << "</name>" << endl;
                XML << "                    <reynoldsnumber>" << thePolarIterator.second.ReynoldsNumber << "</reynoldsnumber>" << endl;
                XML << "                    <altitude>" << thePolarIterator.second.altitude << "</altitude>" << endl;
                XML << "                    <machnumber>" <<thePolarIterator.second.MachNumber << "</machnumber>" << endl;
                std::string Alpha_str(""), Cl_str(""), Cd_str(""), Cm_str(""), origin_polar_str(""), i_stab_str("");
                for (size_t j(0); j < thePolarIterator.second.Points.size(); ++j) {
                    Alpha_str += num2Str(thePolarIterator.second.Points.at(j).AoA);
                    Cl_str += num2Str(thePolarIterator.second.Points.at(j).CL);
                    Cd_str += num2Str(thePolarIterator.second.Points.at(j).CD);
                    Cm_str += num2Str(thePolarIterator.second.Points.at(j).CM);
                    origin_polar_str += "seg0-0.0";
                    i_stab_str += num2Str(thePolarIterator.second.Points.at(j).iStabPolar);
                    if (j < thePolarIterator.second.Points.size() - 1) {
                        Alpha_str += ";";
                        Cl_str += ";";
                        Cd_str += ";";
                        Cm_str += ";";
                        origin_polar_str += ";";
                        i_stab_str += ";";
                    }
                }
                XML << "                    <Alpha>" << Alpha_str << "</Alpha>" << endl;
                XML << "                    <Cl>" << Cl_str << "</Cl>" << endl;
                XML << "                    <Cd>" <<Cd_str << "</Cd>" << endl;
                XML << "                    <Cm>" << Cm_str << "</Cm>" << endl;
                XML << "                    <origin_polar>"<< origin_polar_str << "</origin_polar>" << endl;
                XML << "                    <iStab_polar>"<< i_stab_str << "</iStab_polar>" << endl;
                XML << "                </polar>" << endl;
                XML << "            </polars>" << endl;
                XML << "        </configuration>" << endl;
                }
            XML << "    </configurations>" << endl;
            XML << "</polar-exchange-file>" << endl;
        }
        XML.close();
    }
}

void tawDefaultStrategy::writePolarData() {
     ofstream plot;
    /* plot all untrimmed polars */
    for (size_t CGID(0); CGID < polarDataCGMap.size(); ++CGID) {
        for (const pair<const double, PolarData>& thePolarIterator : polarDataCGMap.at(CGID)) {
            /* plot all clean polars */
            for (const pair<const int, tawDefaultPolar>& theMachIDIterator : thePolarIterator.second.cleanPolars) {
                stringstream machNumber;
                stringstream iStabStream;
                machNumber << setprecision(2) << fixed << theMachIDIterator.second.MachNumber;
                iStabStream << setprecision(2) << fixed << Rounding(config_->stabAngles.at(thePolarIterator.first), 2);
                std::string csvFiles = "tawCalculatePolar_M" + replaceAll(machNumber.str(), ".", "") + "_"
                                       + CoGnames.at(CGID) + "_" + replaceAll(iStabStream.str(), ".", "") +"_plot.csv";
                plot.open(csvFiles.c_str());
                if (!plot) {
                    myRuntimeInfo->err << csvFiles + " could not be opened!" << endl;
                    exit(1);
                } else {
                    plot << "# " + csvFiles << endl;
                    plot << "# Clean Polar" << endl;
                    plot << "# Mach Number " << theMachIDIterator.second.MachNumber << endl;
                    plot << "AoA; C_L; C_Dtotal; C_Dind; C_Dvisc; C_Dwave; C_M; L/D; i_stab; C_Dvisc_wing; "
                    << "C_Dwave_wing; C_Dvisc_stab; C_Dwave_stab; C_Dind_stab; C_L_stab" << endl;
                    for (size_t i(0); i < theMachIDIterator.second.Points.size(); ++i) {
                        plot << setprecision(3) << fixed << theMachIDIterator.second.Points.at(i).AoA << ";";
                        plot << setprecision(4) << fixed << theMachIDIterator.second.Points.at(i).CL << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD_ind << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD_wave << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CM << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CL / theMachIDIterator.second.Points.at(i).CD << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).iStabPolar << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsWing.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsWing.at(i).CD_wave << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsStab.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsStab.at(i).CD_wave << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsStab.at(i).CD_ind << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.PointsStab.at(i).CL << ";";
                        plot << endl;
                    }
                }
                plot.close();
            }
        }
    }
    for (size_t CGID(0); CGID < polarDataCGMap.size(); ++CGID) {
        for (const pair<const double, PolarData>& thePolarIterator : polarDataCGMap.at(CGID)) {
            /* plot all clean polars */
            for (const pair<const std::string, tawDefaultPolar>& theConfigIterator : thePolarIterator.second.highLiftPolars) {
                stringstream machNumber;
                stringstream iStabStream;
                machNumber << setprecision(2) << fixed << theConfigIterator.second.MachNumber;
                iStabStream << setprecision(2) << fixed << Rounding(config_->stabAngles.at(thePolarIterator.first), 2);
                std::string csvFiles = "tawCalculatePolar_" + theConfigIterator.first + "_"
                                       + CoGnames.at(CGID) + "_" + replaceAll(iStabStream.str(), ".", "") +"_plot.csv";
                plot.open(csvFiles.c_str());
                if (!plot) {
                    myRuntimeInfo->err << csvFiles + " could not be opened!" << endl;
                    exit(1);
                } else {
                    plot << "# " + csvFiles << endl;
                    plot << "# Clean Polar" << endl;
                    plot << "# Mach Number " << theConfigIterator.second.MachNumber << endl;
                    plot << "AoA; C_L; C_Dtotal; C_M; i_stab;" << endl;
                    for (size_t i(0); i < theConfigIterator.second.Points.size(); ++i) {
                        plot << setprecision(3) << fixed << theConfigIterator.second.Points.at(i).AoA << ";";
                        plot << setprecision(4) << fixed << theConfigIterator.second.Points.at(i).CL << ";";
                        plot << setprecision(6) << fixed << theConfigIterator.second.Points.at(i).CD << ";";
                        plot << setprecision(6) << fixed << theConfigIterator.second.Points.at(i).CM << ";";
                        plot << setprecision(6) << fixed << theConfigIterator.second.Points.at(i).CL / theConfigIterator.second.Points.at(i).CD << ";";
                        plot << setprecision(6) << fixed << theConfigIterator.second.Points.at(i).iStabPolar << ";";
                        plot << endl;
                    }
                }
                plot.close();
            }
        }
    }
    size_t cruiseMachID(0);
    stringstream machNumber;
    machNumber << setprecision(2) << fixed << data_->initialCruiseMachNumber;
    for (size_t machID(0); machID < polarDataTrimMap.at(0).cleanPolars.size(); ++machID) {
        if (accuracyCheck(polarDataTrimMap.at(0).cleanPolars.at(machID).MachNumber, data_->initialCruiseMachNumber, ACCURACY_HIGH)) {
            cruiseMachID = machID;
            break;
        }
    }
    std::string csvFiles = "tawCalculatePolar_liftDistribution_" + replaceAll(machNumber.str(), ".", "") +"_plot.csv";
    plot.open(csvFiles.c_str());
    if (!plot) {
        myRuntimeInfo->err << csvFiles + " could not be opened!" << endl;
        exit(1);
    } else {
        plot << "# " + csvFiles << endl;
        plot << "# (1)C_L; (2)C_Lw; (3)y, m; (4)C_l; (5)Cl*l, m" << endl;
        for (size_t pointID(0); pointID < polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsWing.size(); ++pointID) {
            vector <double> clDistribution_tmp = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsWing.at(pointID).Cl_distr;
            vector <double> yStations_tmp      = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsWing.at(pointID).yStations;
            reverse(clDistribution_tmp.begin(), clDistribution_tmp.end());
            reverse(yStations_tmp.begin(), yStations_tmp.end());
            // for Cl distr analysis at single CL
            double CL_tmp(0.51);
            if (true || fabs(polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL - CL_tmp) < 1e-5) {  // uncomment "true || " for analysis
                string delimiter(";");
                for (size_t j(0); j < clDistribution_tmp.size(); ++j) {
                    plot << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL << delimiter;
                    plot << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsWing.at(pointID).CL << delimiter;
                    plot << yStations_tmp.at(j) << delimiter;
                    plot << clDistribution_tmp.at(j) << delimiter;
                    plot << getLocalChordLength(liftingSurfaces.front(), yStations_tmp.at(j)) << endl;
                }
                plot << endl << endl;
            }
        }
        plot.close();
    }
    std::string csvFile = "trimmed_designCruisePolar_" + replaceAll(machNumber.str(), ".", "") +"_plot.csv";
    plot.open(csvFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << csvFile + " could not be opened!" << endl;
        exit(1);
    } else {
        plot << "# " + csvFile << endl;
                            plot << "AoA; C_L; C_Dtotal; C_Dind; C_Dvisc; C_Dwave; C_M; L/D; i_stab" << endl;
        for (size_t pointID(0); pointID < polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.size(); ++pointID) {
            plot << setprecision(3) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).AoA << ";";
            plot << setprecision(4) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD_ind << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD_visc << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD_wave << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CM << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL
                                        / polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD << ";";
            plot << setprecision(6) << fixed << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).iStabPolar << ";";
            plot << endl;
        }
        plot.close();
    }
}

void tawDefaultStrategy::writeViscDragData() {
    ofstream plot;
    for (size_t CGID(0); CGID < polarDataCGMap.size(); ++CGID) {
        for (size_t stabID(0); stabID < polarDataCGMap.at(CGID).size(); ++stabID) {
            for (size_t machID(0); machID < polarDataCGMap.at(CGID).at(stabID).cleanPolars.size(); ++machID) {
                stringstream machNumber;
                stringstream iStabStream;
                machNumber << setprecision(2) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).MachNumber;
                iStabStream << setprecision(2) << fixed << Rounding(config_->stabAngles.at(stabID), 2);
                std::string csvFile = "viscDrag_M" + replaceAll(machNumber.str(), ".", "") + "_"
                                       + CoGnames.at(CGID) + "_" + replaceAll(iStabStream.str(), ".", "") +"_plot.csv";
                plot.open(csvFile.c_str());
                if (!plot) {
                    myRuntimeInfo->err << csvFile + " could not be opened!" << endl;
                    exit(1);
                } else {
                    plot << "# " + csvFile << endl;
                    plot << "# Reynolds Number" << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).ReynoldsNumber << endl;
                    plot << "# Clean Polar" << endl;
                    plot << "# Mach Number " << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).MachNumber << endl;
                    plot << "AoA; C_L; C_D_visc; C_D_visc_Wing; C_D_visc_Stab; C_D_visc_Fin; C_D_visc_Fuselage; C_D_visc_Nacelles; C_D_visc_Misc;" << endl;
                    for (size_t i(0); i < polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).Points.size(); ++i) {
                        plot << setprecision(3) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).Points.at(i).AoA << ";";
                        plot << setprecision(3) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).Points.at(i).CL << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).Points.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).PointsWing.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).PointsStab.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).PointsFin.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).PointsFuselage.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).PointsNacelles.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << polarDataCGMap.at(CGID).at(stabID).cleanPolars.at(machID).PointsMisc.at(i).CD_visc << ";";
                        plot << endl;
                    }
                }
                plot.close();
            }
        }
    }
}

void tawDefaultStrategy::setReferenceValues() {
    data_->span.set_value(getHalfSpan(liftingSurfaces.at(0)) * 2);
    data_->MAC.set_value(geom2::measure::mean_aerodynamic_chord(liftingSurfaces.at(0)));
    data_->S_ref.set_value(geom2::measure::reference_area(liftingSurfaces.at(0)));
    data_->neutral_point_xCoordinate.set_value(getNeutralPointXposition(getTrimmedCruisePolar(polarDataCGMap.at(0).at(0).cleanPolars, data_->initialCruiseMachNumber),
                                                                        data_->MAC.value(), CoGPositions["design"].xCoordinate));
}

void tawDefaultStrategy::setOutputValues() {
    data_->C_L_optimumCruise.set_value(getCLoptCruise());
    data_->CL_max_landing.set_value(getMaxCL(polarDataTrimMap.at(0).highLiftPolars.at("landing")));
    data_->C_LmaxTO.set_value(getMaxCL(polarDataTrimMap.at(0).highLiftPolars.at("takeoff")));
    data_->C_LgroundRoll.set_value(0.4);
    data_->max_spoiler_factor.set_value(2.5);
    // handle the axcml entries
    data_->polar_file.set_value(this->designPolarXMLFilePath);
    data_->polar_file.update(rtIO_->acxml);
    vector<std::string>types{"cruise", "departure", "departure", "departure", "approach", "approach", "approach"};
    vector<std::string>names{"clean", "takeoff", "takeoff_landing_gear_retracted", "climb", "approach", "approach_landing_gear_out", "landing"};
    //Write aerodynamic node with description and tool level
    rtIO_->acxml.at("analysis/aerodynamics").addAttrib("description", "Aerodynamical analysis");
    rtIO_->acxml.at("analysis/aerodynamics").addAttrib("tool_level", EndnodeReadOnly<std::string>("control_settings/own_tool_level").read(rtIO_->moduleConfig).value());
    std::string subPath = "analysis/aerodynamics/polar";
    Endnode<int> configurations(Endnode<int>(subPath + "/configurations", "Number of configurations in the polar file", types.size()));
    configurations.update(rtIO_->acxml);
    // setup configuration nodes
    for (size_t ID = 0; ID < types.size(); ID++) {
        rtIO_->acxml.at(subPath)["configuration@" + num2Str(ID)];
        rtIO_->acxml.at(subPath + "/configuration@" + num2Str(ID)).addAttrib("description", "Configuration in polar file marked with ID - name can vary");
        Endnode<string>type(Endnode<string>(subPath + "/configuration@" + num2Str(ID) + "/type", " Type", "cruise"));
        Endnode<string>name(Endnode<string>(subPath + "/configuration@" + num2Str(ID) + "/name", " Name", "clean"));
        type.set_value(types.at(ID));
        name.set_value(names.at(ID));
        type.update(rtIO_->acxml);
        name.update(rtIO_->acxml);
    }
}

void tawDefaultStrategy::update_xml_entries(std::string mode) {
    data_->updateNodes(mode);
}

double tawDefaultStrategy::getCLoptCruise() {
    double LoverDmax(0.0);
    double CLopt(0.0);
    std::vector<defaultPolarPoint> theTrimmedCruisePolar(getTrimmedCruisePolar(polarDataTrimMap.at(0).cleanPolars, data_->initialCruiseMachNumber));
    for (size_t pointID(0); pointID < theTrimmedCruisePolar.size(); ++pointID) {
        double LoverD = theTrimmedCruisePolar.at(pointID).CL / theTrimmedCruisePolar.at(pointID).CD;
            if (LoverD > LoverDmax) {
                LoverDmax = LoverD;
                CLopt = theTrimmedCruisePolar.at(pointID).CL;
            }
    }
    return CLopt;
}

double tawDefaultStrategy::getMaxCL(const tawDefaultPolar& aPolar) {
    double maxCL = -10000;
    for (size_t ID = 0; ID < aPolar.Points.size(); ++ID) {
        if (aPolar.Points.at(ID).CL > maxCL) {
            maxCL = aPolar.Points.at(ID).CL;
        }
    }
    return maxCL;
}
