# Import standard libraries.
from numpy import linspace, concatenate


def prepare_payload_range_diagram_data(dict_ac_data, dict_operating_costs, runtime_output):
    """Prepare data for payload range diagram (PRD).

    The payload range diagram shows four characteristic points:
        - Point A: Maximum payload mass at no range
        - Point B: Maximum range at maximum payload mass
        - Point C: Maximum range at maximum fuel mass
        - Point D: Ferry range
    Knowing these four points, arrays are created for the payload and the range, which have n points each per section
    (point A to point B, point B to point C, etc.). In the following process, the arrays are necessary to determine the
    cost trajectories and plot them against the range.

    This function adds the 'payload_range_diagram_data' dict to the 'dict_operating_costs' that contains the following
    parameters:
        - int idx_design_point: Index of design point of current aircraft in 1
        - array n_pax_for_prd:  Number of PAX (range dependent) in 1
        - list n_points: Number of points per PRD segment (equals range in km) in 1
        - array payloads_for_prd: Payload mass (running from max. payload mass to zero, range dependent) in kg
        - array ranges_for_prd: Range (running from 0 to ferry range) in kg

    :param dict dict_ac_data: Dict containing parameters and values from aircraft exchange and module configuration file
    :param dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    """

    # Extract necessary variables from 'dict_ac_data'.
    payload_point_a = dict_ac_data['m_payload_max']
    payload_point_b = dict_ac_data['m_payload_max']
    payload_point_c = dict_ac_data['m_payload_at_max_fuel']
    payload_point_d = 0
    range_point_a = 0
    range_point_b = dict_ac_data['range_at_max_payload']
    range_point_c = dict_ac_data['range_at_max_fuel']
    range_point_d = dict_ac_data['range_ferry']
    m_cargo = dict_ac_data['m_cargo']
    m_passenger_incl_luggage = dict_ac_data['m_passenger'] + dict_ac_data['m_luggage']
    stage_length = dict_ac_data['stage_length']

    # Determine number of points (equals stage_length in m).
    n_points = [round(range_point_b), round(range_point_c) - round(range_point_b),
                round(range_point_d) - round(range_point_c)]
    # Payloads for PRD.
    payloads_for_prd = concatenate([
        linspace(payload_point_a, payload_point_b, n_points[0], endpoint=False),
        linspace(payload_point_b, payload_point_c, n_points[1], endpoint=False),
        linspace(payload_point_c, payload_point_d, n_points[2])
    ], axis=None)
    # Ranges for PRD.
    ranges_for_prd = concatenate([
        linspace(range_point_a, range_point_b, n_points[0], endpoint=False),
        linspace(range_point_b, range_point_c, n_points[1], endpoint=False),
        linspace(range_point_c, range_point_d, n_points[2])
    ], axis=None)
    # Determine number of PAX.
    n_pax_for_prd = (payloads_for_prd - m_cargo) / m_passenger_incl_luggage
    n_pax_for_prd[n_pax_for_prd < 1] = 0
    # Determine index of design point (equals range of design point in m, rounded).
    idx_design_point = round(stage_length)

    # Prepare output data.
    dict_operating_costs['payload_range_diagram_data'] = {}
    dict_operating_costs['payload_range_diagram_data']['idx_design_point'] = idx_design_point
    dict_operating_costs['payload_range_diagram_data']['n_pax_for_prd'] = n_pax_for_prd
    dict_operating_costs['payload_range_diagram_data']['n_points'] = n_points
    dict_operating_costs['payload_range_diagram_data']['payloads_for_prd'] = payloads_for_prd
    dict_operating_costs['payload_range_diagram_data']['ranges_for_prd'] = ranges_for_prd

    return dict_operating_costs
