/**
 * @file standard_mission.cpp
 * @author Gerrit Pauls (gerrit.pauls@tuhh.de)
 * @brief 
 * @version 0.1
 * @date 2024-08-30
 * 
 * @copyright Copyright (c) 2024
 * 
 */
#include "standard_mission.h"

/**
 * @brief Construct a new standard_mission:: standard_mission object
 * 
 * @param toolName entry from toolinfo.h
 * @param toolVersion entry from toolinfo.h
 */
standard_mission::standard_mission(const std::shared_ptr<RuntimeIO>& rtIO) :
    rtIO(rtIO),
    design_mission(nullptr),
    study_mission(nullptr),
    requirements_mission(nullptr),
    use_delta_max_operating_mach(EndnodeReadOnly<bool>("program_settings/maximum_operating_mach_number/delta").read(rtIO->moduleConfig).value()),
    adapt_climb_speed(EndnodeReadOnly<bool>("program_settings/adapt_climb_speed_schedule/enable").read(rtIO->moduleConfig).value()),
    use_design_mission_mach(EndnodeReadOnly<bool>("program_settings/study_mission/copy_mach_number").read(rtIO->moduleConfig).value()),
    use_design_mission_ICO(EndnodeReadOnly<bool>("program_settings/study_mission/copy_initial_cruise_altitude").read(rtIO->moduleConfig).value()),
    delta_max_operating_mach(EndnodeReadOnly<double>("program_settings/maximum_operating_mach_number/delta").read(rtIO->moduleConfig).value()),
    crossover_altitude_specs(EndnodeReadOnly<double>("program_settings/adapt_climb_speed_schedule/crossover_altitude").read(rtIO->moduleConfig).value()) {
    }

void standard_mission::initialize() {
    myRuntimeInfo->info << "Initialize Mission Parameters ..." << std::endl;
    std::string subpath("requirements_and_specifications/design_specification/transport_task/passenger_definition/");
    double single_PAX_mass = EndnodeReadOnly<double>(subpath + "mass_per_passenger").read(rtIO->acxml).value();
    single_PAX_mass += EndnodeReadOnly<double>(subpath + "luggage_mass_per_passenger").read(rtIO->acxml).value();

    design_mission = new mission_parameters("design_mission", single_PAX_mass);
    this->initialize_mission(design_mission);
    this->write_shell_output(design_mission);

    study_mission = new mission_parameters("study_mission", single_PAX_mass);
    this->initialize_mission(study_mission);
    this->write_shell_output(study_mission);

    /* Copy design mission for requirements mission and alter paths, name and cruise segment */
    requirements_mission = new mission_parameters("design_mission", single_PAX_mass);
    this->initialize_mission(requirements_mission);
    requirements_mission->mission_type = "requirements_mission";
    requirements_mission->mission_xml_name =
        EndnodeReadOnly<std::string>("requirements_and_specifications/mission_files/requirements_mission_file").read(rtIO->acxml).value();
    requirements_mission->write_mission_file =
        EndnodeReadOnly<bool>("program_settings/write_requirement_mission").read(rtIO->moduleConfig).value();
    requirements_mission->cruise_step_segments.clear(); // Clear cruise steps for requirement mission
    requirements_mission->cruise_step_segments.push_back(design_mission->cruise_step_segments.front());
    requirements_mission->cruise_step_segments.front().relative_segment_length = 1.; // Set the relative segment length for requirement mission
    requirements_mission->cruise_step_segments.front().altitude = // Set flight level for requirement mission
        convertUnit(HECTO, FOOT, NOPREFIX, METER, requirements_mission->initial_cruise_altitude);
}

void standard_mission::run() {
    if (design_mission->write_mission_file == true) {
        set_mission_profiles(design_mission);
    }
    if (study_mission->write_mission_file == true) {
        set_mission_profiles(study_mission);
    }
    if (requirements_mission->write_mission_file == true) {
        set_mission_profiles(requirements_mission);
    }
}

void standard_mission::update() {
    this->setup_mission_file(design_mission);
    this->fill_mission_file(design_mission);

    this->setup_mission_file(study_mission);
    this->fill_mission_file(study_mission);

    this->setup_mission_file(requirements_mission);
    this->fill_mission_file(requirements_mission);
    // std::string subPath("AcftExchangeFile/Performance/RequirementsChecks/TLARs/"); // TODO(Gerrit): Check if still needed!
    // /* Mission type: requirement mission */
    // if (myRequirementMissionSettingsPt->getWriteMissionFile() &&
    //         (!this->myMissionXMLPt->acXML.at(subPath + "h_initialCruise_Design").getIntAttrib("Checked") ||
    //          !this->myMissionXMLPt->acXML.at(subPath + "h_maxOperating").getIntAttrib("Checked") ||
    //          !this->myMissionXMLPt->acXML.at(subPath + "h_maxOEI").getIntAttrib("Checked") ||
    //          !this->myMissionXMLPt->acXML.at(subPath + "M_initialCruise_Design").getIntAttrib("Checked") ||
    //          !this->myMissionXMLPt->acXML.at(subPath + "TTC_Design").getIntAttrib("Checked"))) {
    //     this->myMissionXMLPt->acXML.at("AcftExchangeFile/Performance/RequirementsChecks/MissionFile").setAttrib("OfftakesWritten", 0);
    //     this->generateMissionXMLFile(myRequirementMissionProfilePt, myRequirementMissionSettingsPt, myRequirementMissionPt);
    // }
}

void standard_mission::report() {
}

void standard_mission::save() {
}

void standard_mission::initialize_mission(standard_mission::mission_parameters* mission) {
    this->set_module_settings(mission);
    this->set_range_and_payload(mission);
    this->set_cruise(mission);
    this->set_climb_and_descent(mission);
    this->set_start_and_landing(mission);
    this->set_performance_limits(mission);
    // TODO(GERRIT): Soll das nochmal rein?
    /* Adjust maximum operating Mach number by delta to Mach_Cruise if necessary (for parameter studies with Mach variation) */
    // if (mySettingsPt->getUseDeltaMaxOperatingMachNumber()) {
    //     double maxOperatingMachNumber(designMission->getMachInitialCruise() + mySettingsPt->getDeltaMaxOperatingMachNumber());
    //     acXML.at("AcftExchangeFile/Requirements/DesignFlightEnvelope/MMO") = maxOperatingMachNumber;
    // }
}

/* Write the shell output */
void standard_mission::write_shell_output(standard_mission::mission_parameters* mission) {
    myRuntimeInfo->out << "***************** Gathered input *****************" << std::endl;
    myRuntimeInfo->out << "* Mission type: " + mission->mission_type << std::endl;
    /* Range */
    myRuntimeInfo->out << "* Range                    = " << convertUnit(METER, NAUTICALMILE, mission->range.value()) << " NM" << std::endl;
    /* Payload */
    myRuntimeInfo->out << "* Payload                  = " << mission->payload.value() << " kg" << std::endl;
    /* Initial cruise parameters */
    myRuntimeInfo->out << "* Cruise Mach number       = " << mission->initial_cruise_mach.value() << std::endl; // Mach number
    myRuntimeInfo->out << "* Initial cruise altitude  = " << mission->initial_cruise_altitude << " FL" << std::endl; // Altitude
    /* Speed schedule */
    myRuntimeInfo->out << "* Climb speed under FL100  = " << mission->CAS_ATC_limit_climb << " m/s" << std::endl;
    myRuntimeInfo->out << "* Climb speed above FL100  = " << mission->CAS_over_flight_level_100_climb << " m/s" << std::endl;
    myRuntimeInfo->out << "* Climb Mach number        = " << mission->mach_climb << std::endl;
    myRuntimeInfo->out << "**************************************************" << std::endl;
    myRuntimeInfo->out << std::endl;
}

void standard_mission::set_mission_profiles(standard_mission::mission_parameters* mission) {
    this->set_departure_profile(mission);
    this->set_cruise_profile(mission);
    this->set_approach_profile(mission);
}

void standard_mission::setup_mission_file(standard_mission::mission_parameters* mission) {
    myRuntimeInfo->out << "Create " << mission->mission_xml_name << "..." << std::endl;
    /* Open file */
    std::ofstream outstream;
    outstream.open(rtIO->getMissionDataDir() + "/" + mission->mission_xml_name.c_str());
    /* Check if the file can be opened */
    if (!outstream) { // File can not be opened -> create directory
        rtIO->createMissionDataDir();
        outstream.open(rtIO->getMissionDataDir() + "/" + mission->mission_xml_name.c_str());
    }
    if (!outstream) { // File can still not be opened, error message
        myRuntimeInfo->err << "Could not open " << rtIO->getMissionDataDir() + "/" +  mission->mission_xml_name << ". Abort progrm." << std::endl;
        exit(1);
    } else { // File can be opened, the generate mission XML output is created
        /* Version and encoding */
        outstream << "<?xml version=\"1.0\" encoding=\"utf-8\" ?>" << std::endl;
        outstream << "<mission>" << std::endl;
        outstream << "</mission>" << std::endl;
        outstream.close();
    }
}

void standard_mission::fill_mission_file(standard_mission::mission_parameters* mission) {
    node& file(aixml::openDocument(rtIO->getMissionDataDir() + "/" + mission->mission_xml_name));
    mission->range.update(file.at("mission"));
    mission->payload.update(file.at("mission"));
    mission->pax_number.update(file.at("mission"));
    mission->cargo_mass.update(file.at("mission"));
    mission->initial_cruise_mach.update(file.at("mission"));
    mission->alternate_distance.update(file.at("mission"));
    mission->taxi_time_origin.update(file.at("mission"));
    mission->taxi_time_destination.update(file.at("mission"));
    mission->engine_warmup_time.update(file.at("mission"));
    mission->terminal_operation_time.update(file.at("mission"));
    mission->taxiing_procedure.update(file.at("mission"));
    if (!mission->origin_airport.value().empty() &&  !mission->origin_airport.value().empty()) {
        mission->origin_airport.update(file.at("mission"));
        mission->destination_airport.update(file.at("mission"));
    }

    /* Setup departure blocks */
    std::string subpath("mission/departure");
    std::string description("Departure procedure; mode Takeoff: %N1 (double), aerodynamic configuration (string).");
    description += " -- mode Climb: %N1 (double), aerodynamic configuration (string), End Point Altitude [m] (double).";
    description += " -- mode Accelerate: %N1 (double), aerodynamic configuration (string), Rate of climb [m/s] (double), End point CAS [m/s] (double)";
    description += " -- Alternatively, use the names from the ANP database";

    file[subpath];
    file.at(subpath).addAttrib("description", description);

    subpath += "/departure_step@";
    for (uint16_t i(0); i < mission->departure_steps.size(); i++) {
        file[(subpath + num2Str(i))];
        file.at(subpath + num2Str(i)).addAttrib("description", "Single departure step");
        mission->departure_steps.at(i).write_departure_steps(subpath + num2Str(i), &file);
    }

    /* Setup cruise blocks */
    subpath = "mission/cruise";
    description =  "Cruise procedure: mode change_speed_to_CAS:  %N1 (double), aerodynamic configuration (string), Rate of climb [m/s] (double), End point CAS [m/s] (double)";
    description += " -- mode change_speed_to_Mach: %N1 (double), aerodynamic configuration (string), Rate of climb [m/s] (double), End point CAS [m/s] (double)";
    description += " -- mode climb_to_cruise:  %N1 (double), aerodynamic configuration (string), End Point Altitude [m] (double), Mach [-] (double).";
    description += " -- mode cruise: aerodynamic configuration (string), Range [%] (double, relative distance at the end of the cruise segment without climb and descent)";
    description += " -- mode change_flight_level: %N1 (double), aerodynamic configuration (string), End Point Altitude [m] (double).";
    description += " -- mode descriptionent_to_approach: aerodynamic configuration (string), End Point Altitude [m] (double).";

    file[subpath];
    file.at(subpath).addAttrib("description", description);

    subpath += "/cruise_step@";
    for (uint16_t i(0); i < mission->cruise_steps.size(); i++) {
        file[(subpath + num2Str(i))];
        file.at(subpath + num2Str(i)).addAttrib("description", "Single cruise step");
        mission->cruise_steps.at(i).write_cruise_steps(subpath + num2Str(i), &file);
    }

    /* Setup approach blocks */
    subpath = "mission/approach";
    description =  "Approach procedure: mode descent: aerodynamic configuration (string), End Point Altitude [m] (double), glidepath [deg] (double).";
    description += " -- mode change_speed: aerodynamic configuration (string), Ground distance [m] (double), End point CAS [m/s] (double), glidepath [deg] (double).";
    description += " -- mode landing: aerodynamic configuration (string), End Point Altitude [m] (double), glidepath [deg] (double).";

    file[subpath];
    file.at(subpath).addAttrib("description", description);

    subpath += "/approach_step@";
    for (uint16_t i(0); i < mission->approach_steps.size(); i++) {
        file[(subpath + num2Str(i))];
        file.at(subpath + num2Str(i)).addAttrib("description", "Single approach step");
        mission->approach_steps.at(i).write_approach_steps(subpath + num2Str(i), &file);
    }

    aixml::saveDocument(file);
    aixml::closeDocument(file);
}
