/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <algorithm>
#include "production.h"
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>

Production::Production()
    :
    myEcoDatabase(ecoDatabase()),
    materialEnergy(0.),
    testFlightFuel(0.),
    totalRC(0.),
    E_manuf_engine(0.),
    E_manuf_total(0.),
    Energy_engine(0.),
    Energy_total(0.) {
}

void Production::calculateProductionResources(const std::shared_ptr<SchaeferIOData>& data, const std::shared_ptr<RuntimeIO>& rtIO) {
    /* Raw material and production */
    //Distribution of components
    double energyRecycling = 0;
    double energyLandfill = 0;
    recovery.clear();
    if (data->configuration.primaryMaterialRecycling) {
        myRuntimeInfo->info << " Primary Recycling on" << std::endl;
        for (int i(0); i < numberOfMaterials; i++)
            recovery.push_back(recoveryRate[i]);
    } else {
        myRuntimeInfo->info << " No Primary Recycling" << std::endl;
        for (int i(0); i < numberOfMaterials; i++)
            recovery.push_back(0);
    }
    double wingRC(0.);
    for (int i(0); i < data->configuration.partsNo; i++) {
        theAcComp.push_back(mainComponent(data->mainPartsName[i]));
        theAcComp.back().mass = data->mainPartsMass[i];
        if (data->configuration.calculateRC) {// TODO: only wingRC costs needed in production, maybe shift rest to operation?
            myRuntimeInfo->info << "Calculate RC of main part: " << data->mainPartsName[i] << std::endl;
            double CPI75(4.05); // Factor to convert 1975-USD to 2010-USD
            if (data->mainPartsName[i] == "Wing") {
                theAcComp.back().RC = CPI75 * getNASACost(theAcComp.back().mass, data->configuration.producedAC, 1730., 0.766, -0.218);
                wingRC = theAcComp.back().RC;
            } else if (data->mainPartsName[i] == "Fuselage") {
                theAcComp.back().RC = CPI75 * getNASACost(theAcComp.back().mass, data->configuration.producedAC, 2060., 0.766, -0.218);
            } else if (data->mainPartsName[i] == "Fin") {
                theAcComp.back().RC = data->finMassRatio * CPI75 * getNASACost(data->empennageMass, data->configuration.producedAC, 1820., 0.766, -0.218);
            } else if (data->mainPartsName[i] == "Tailplane") {
                theAcComp.back().RC = data->stabilizerMassRatio * CPI75 * getNASACost(data->empennageMass, data->configuration.producedAC, 1820., 0.766, -0.218);
            } else if (data->mainPartsName[i] == "LandingGear") {
                const double structureWeightPercentage = 0.55;
                const double controlWeightPercentage = 0.1;
                const double wheelsBrakesWeightPercentage = 0.25;
                const double tiresPercentage = 0.1;
                double gearStructureCost = 0.;
                double gearControlsCost = 0.;
                double gearWheelsAndBrakesCost = 0.;
                double gearTiresCost = 0.;
                if ((convertUnit(KILO, GRAM, POUND, theAcComp.back().mass)* structureWeightPercentage) <= 10000.) {
                    gearStructureCost = CPI75 * getNASACost(structureWeightPercentage * theAcComp.back().mass, data->configuration.producedAC, 1180., 0.766, -0.218);
                } else {
                    gearStructureCost = CPI75 * getNASACost(structureWeightPercentage * theAcComp.back().mass, data->configuration.producedAC, 136., 1.0, -0.218);
                }
                gearControlsCost = CPI75 * getNASACost(controlWeightPercentage * theAcComp.back().mass, data->configuration.producedAC, 157., 1.0, -0.0896);
                gearWheelsAndBrakesCost = CPI75 * getNASACost(wheelsBrakesWeightPercentage * theAcComp.back().mass, data->configuration.producedAC, 23.8, 1.0, -0.0896);
                gearTiresCost = CPI75 * getNASACost(tiresPercentage *theAcComp.back().mass, data->configuration.producedAC, 2., 1.0, 0.0);
                theAcComp.back().RC = gearStructureCost + gearControlsCost + gearWheelsAndBrakesCost + gearTiresCost;
            } else if (data->mainPartsName[i] == "Pylons") {
                theAcComp.back().RC = data->pylonMassRatio * CPI75 * static_cast<double>(data->numberOfEngines) * getNASACost((data->nacelleMass
                                + data->pylonMass) / static_cast<double>(data->numberOfEngines), data->configuration.producedAC, 3470., 0.766, -0.218);
            } else if (data->mainPartsName[i] == "PowerUnit") {
                /*Engine Price after Stephan Langhans Dissertation 2013, S.50; Attention: here all cost values in Mio USD, assumption: Langhans formula for Dry Engine
                 (without systems and nacelles); see values Airbus NSR-01 and calculation CSR-01 in documentation*/
                double engineWeightin1000LBS = convertUnit(KILO, GRAM, POUND, data->myEngine.get_dry_mass()) / 1000.; //mass of one engine
                double equippedEnginesRC = data->numberOfEngines * (2.941 + 0.2603 * engineWeightin1000LBS + 0.04765
                                                 * pow(engineWeightin1000LBS, 2)) * 1000000;
                //from here again NASA formula
                //Engine Systems: Assumption Engine Dry and Engine Systems = Equipped Engines; from here again NASA Formula
                double otherEngineSystemsMass = data->engineMass / data->numberOfEngines - data->myEngine.get_dry_mass(); // Other systems mass eg. hydraulics, starter, etc.;
                equippedEnginesRC += static_cast<double>(data->numberOfEngines) * CPI75 * getNASACost(otherEngineSystemsMass,
                                        data->configuration.producedAC, 159., 1., -0.0896);
                //Nacelles
                double nacellesRC = data->nacelleMassRatio * static_cast<double>(data->numberOfEngines) * CPI75 * getNASACost((data->nacelleMass
                                        + data->pylonMass) / static_cast<double>(data->numberOfEngines), data->configuration.producedAC, 3470., 0.766, -0.218);
                //Bleed Air
                double bleedAirSystemRC(0.);
                if (convertUnit(KILO, GRAM, POUND, data->bleedAirSystemMass) <= 400.) {
                    bleedAirSystemRC = CPI75 * getNASACost(data->bleedAirSystemMass, data->configuration.producedAC, 151., 1.0, -0.0896);
                } else {
                    bleedAirSystemRC = CPI75 * getNASACost(data->bleedAirSystemMass, data->configuration.producedAC, 201., 1.0, -0.0896);
                }
                //Fuel System
                double fuelSystemRC = CPI75 * getNASACost(data->fuelSystemMass, data->configuration.producedAC, 61.9, 1.0, -0.0896);
                // Total engine
                theAcComp.back().RC = equippedEnginesRC + nacellesRC + bleedAirSystemRC + fuelSystemRC;
            } else {
                myRuntimeInfo->err << "No recurring cost method available for main component " << data->mainPartsName[i] << " !"<< std::endl;
            }
        } else { // ATTENTION: path to RC of old exchange file version, adapt before use!
            double costRatio(1.);
            if (data->mainPartsName[i] == "Fin") {
                costRatio = data->finMassRatio;
            } else if (data->mainPartsName[i] == "Tailplane") {
                costRatio = data->stabilizerMassRatio;
            }
            theAcComp.back().RC = costRatio * checkBoundaries(rtIO->acxml.at("/aircraft_exchange_file/MonetaryValues/LCC/RecurringCosts/RCComponents/"
                                + data->mainPartsRCPath[i]), 0.0, true, 1e9, true);
        }
        myRuntimeInfo->info << "mass [kg]: " << theAcComp.back().mass << std::endl;
        myRuntimeInfo->info << "RC [$]: " << theAcComp.back().RC << std::endl;
        /* Energy calculation: Distribution according to aircraft components (Main Parts Modus) */
        //Material distribution. For numeration see production.h: const double matDistr
        double manufacturingEnergy(0.);
        for (int j(0); j < numberOfMaterials; j++) {
            theAcComp.back().r_material.push_back(matDistr[i][j]);
            double massMaterialComp = theAcComp.back().r_material.back() * theAcComp.back().mass;
            double massMFGMaterialComp = massMaterialComp / (1 - clippings[j]);
            double massOreMaterialComp = massMFGMaterialComp * (1 - clippings[j] * recovery[j]);
            theAcComp.back().oreMaterial.push_back(massOreMaterialComp);
            double energyMaterial = massMaterialComp * specificE_manuf[j];
            if (data->configuration.primaryMaterialRecycling) {
                if (j != 1) {
                    //Composites will be recycled.
                    /*Note: Specific energy is wrong described; rather Primary Energy=CED; but here process energy is looked for,
                      i.e. divide by PrimaryEnergy from electricity consumption; see Notes Annex Diss Schaefer*/
                    energyRecycling = 1 / recyclingEnergyRatio[j] * myEcoDatabase.matSpecificEnergy.at(j)
                                      / myEcoDatabase.elecEmData.specificEnergy * massMFGMaterialComp * clippings[j];
                }
                energyLandfill = myEcoDatabase.disposalSpecificEnergy.at(j) / myEcoDatabase.elecEmData.specificEnergy * massMFGMaterialComp * clippings[j] * (1 - recovery[j]);
            }
            manufacturingEnergy += (energyMaterial + energyRecycling + energyLandfill);
            myRuntimeInfo->info << std::endl;
            myRuntimeInfo->info << "Material " << j + 1 << "..." << std::endl;
            myRuntimeInfo->info << "ER: " << recyclingEnergyRatio[j] << std::endl;
            myRuntimeInfo->info << "clippings[j]: " << clippings[j] << std::endl;
            myRuntimeInfo->info << "Mass share: " << theAcComp.back().r_material.back() << std::endl;
            myRuntimeInfo->info << "Mass material: " << massMaterialComp << std::endl;
            myRuntimeInfo->info << "Mass Ore: " << theAcComp.back().oreMaterial.back() << std::endl;
            myRuntimeInfo->info << "spec. energy consumption: " << specificE_manuf[j] << std::endl;
            myRuntimeInfo->info << "energyMaterial: " << energyMaterial << std::endl;
            myRuntimeInfo->info << "energyRecycling: " << energyRecycling << std::endl;
            myRuntimeInfo->info << "energyLandfill: " << energyLandfill << std::endl;
            myRuntimeInfo->info << "manufacturingEnergy: " << manufacturingEnergy << std::endl;
        }
        theAcComp.back().E_manuf = manufacturingEnergy;
        E_manuf_total += theAcComp.back().E_manuf;
        myRuntimeInfo->debug << "Masse:   " << theAcComp.back().mass << std::endl;
        myRuntimeInfo->debug << "Energie: " << theAcComp.back().E_manuf << std::endl;
    }
    E_manuf_engine = E_manuf_total - theAcComp.back().E_manuf;
    /* Energy Calculation: Distribution by Materials (Material Mode) */
    myRuntimeInfo->debug << "E_manuf_total: " << E_manuf_total << std::endl;
    for (int i(0); i < numberOfMaterials; i++) {
        theMaterial.push_back(Production::material());
        theMaterial.back().matName = data->materialName[i];
        for (int j(0); j < data->configuration.partsNo; j++) {
            theMaterial.back().mass += theAcComp.at(j).r_material.at(i) * theAcComp.at(j).mass;
            theMaterial.back().massOre += theAcComp.at(j).r_material.at(i) * theAcComp.at(j).mass / (1 - clippings[i]) * (1 - clippings[i] * recovery[i]); //
        }
    }
    myRuntimeInfo->debug << "E_manuf_total: " << E_manuf_total << std::endl;
    /* TEST Output */
    for (int i(0); i < numberOfMaterials; i++) {
        myRuntimeInfo->debug << std::endl;
        myRuntimeInfo->debug << theMaterial.at(i).matName << std::endl;
        myRuntimeInfo->debug << "Material mass:  " << theMaterial.at(i).mass << std::endl;
        myRuntimeInfo->debug << "Raw material masse  " << theMaterial.at(i).mass*(1 + clippings[i]) << std::endl;
        myRuntimeInfo->debug << "Energy:            " << theMaterial.at(i).E_manuf << std::endl;
    }
    /* personnel expenditure */
    if (data->configuration.calculateRC) {
        totalRC = wingRC / 0.0627; // Assumption for single aisle aircraft: wingRC = 7% * totalRC [Lam12]
        theFinalAssembly.costs = 0.13 * totalRC; // Assumption for single aisle aircraft: finalAssemblyRC = 20% * totalRC [Lam12]
    } else { // from Micado aircraft.xml
        totalRC = checkBoundaries(rtIO->acxml.at("/aircraft_exchange_file/MonetaryValues/LCC/RecurringCosts/RCTotal"), 0.0, true, 1e9, true);
        theFinalAssembly.costs = checkBoundaries(rtIO->acxml.at("aircraft_exchange_file/MonetaryValues/LCC/RecurringCosts/RCComponents/FinalAssembly"), 0.0, true, 1e9, true);
    }
//41% Labour of RC, s. Markish, s. 64 (33% Materials, Rest Others! (do not count here since also engineering work that has been neglected, s. Diss. Schaefer)
    theManufacturingWork.labourCost = 0.41 * (totalRC - theFinalAssembly.costs);
    myRuntimeInfo->debug << "theFinalAssembly.costs: " << theFinalAssembly.costs << std::endl;
    myRuntimeInfo->debug << "totalRC: " << totalRC << std::endl;
    myRuntimeInfo->debug << "theManufacturingWork.labourCost: " << theManufacturingWork.labourCost << std::endl;
    theManufacturingWork.E_elec_eng = E_PerWorker_pr[0] * theManufacturingWork.labourCost / (C_labourHour_pr * t_labourYear_pr) * (data->configuration.engineEngineeringFactor);
    theManufacturingWork.E_heat_eng = E_PerWorker_pr[1] * theManufacturingWork.labourCost / (C_labourHour_pr * t_labourYear_pr) * (data->configuration.engineEngineeringFactor);
    myRuntimeInfo->debug << "E_heat_eng: " << theManufacturingWork.E_heat_eng << std::endl;
    theManufacturingWork.E_elec = E_PerWorker_pr[0] * theManufacturingWork.labourCost / (C_labourHour_pr * t_labourYear_pr) * (1 + data->configuration.engineEngineeringFactor);
    theManufacturingWork.E_heat = E_PerWorker_pr[1] * theManufacturingWork.labourCost / (C_labourHour_pr * t_labourYear_pr) * (1 + data->configuration.engineEngineeringFactor);
    myRuntimeInfo->debug << "Man years " << theManufacturingWork.labourCost / (C_labourHour_pr * t_labourYear_pr) << std::endl;
    //final assembly
    theFinalAssembly.E_elec = E_PerWorker_pr[0] * theFinalAssembly.costs / (C_labourHour_pr * t_labourYear_pr);
    theFinalAssembly.E_heat = E_PerWorker_pr[1] * theFinalAssembly.costs / (C_labourHour_pr * t_labourYear_pr);
    Energy_total = E_manuf_total + theFinalAssembly.E_elec + theFinalAssembly.E_heat + theManufacturingWork.E_elec +
                   theManufacturingWork.E_heat; //still without raw material and without "energy expenditure for energy production")
    Energy_engine = E_manuf_engine + (theManufacturingWork.E_elec + theManufacturingWork.E_heat) * data->configuration.engineEngineeringFactor
                        / (1 + data->configuration.engineEngineeringFactor);
    myRuntimeInfo->info << "Assembly costs " << theFinalAssembly.costs << std::endl;
    myRuntimeInfo->info << "Assembly C_labourHour_pr " << C_labourHour_pr << std::endl;
    myRuntimeInfo->info << "Assembly t_labourYear_pr " << t_labourYear_pr << std::endl;
    /* Transport */
    for (unsigned int i(0.); i < numberOfTransportationTypes; i++) {
        std::string name_tmp;
        name_tmp = data->transportationName[i];
        theTransportation.push_back(TransportLogistics(name_tmp));
        myRuntimeInfo->debug << theTransportation.back().transpType << "...:" << std::endl;
        for (int j(0.); j < data->configuration.compTransportNo; j++) {
            double distance_tmp(0.);
            double mass_tmp(0.);
            if (data->componentTransportation[j] == theTransportation.back().transpType) {
                std::string compName_tmp = data->componentName[j];
                distance_tmp = transportationDistance[j];
                mass_tmp = data->transportedCompMass[j];
                theTransportation.back().name_component.push_back(compName_tmp);
                theTransportation.back().distance_component.push_back(distance_tmp);
                theTransportation.back().mass_component.push_back(data->transportedCompMass[j]);
                myRuntimeInfo->debug << theTransportation.back().name_component.back() << "... " << std::endl;
                myRuntimeInfo->debug << "Distance " << theTransportation.back().distance_component.back() << std::endl;
                myRuntimeInfo->debug << "Mass (pro rata) " << theTransportation.back().mass_component.back() << std::endl;
            }
            theTransportation.back().mass_total += mass_tmp;
            theTransportation.back().distance_total += distance_tmp;
        }
        theTransportation.back().fuel = theTransportation.back().mass_total * theTransportation.back().distance_total * transportationFuelCons[i] / 1000.;
        myRuntimeInfo->debug << "Total distance: " << theTransportation.back().distance_total << std::endl;
        myRuntimeInfo->debug << "Total mass: " << theTransportation.back().mass_total << std::endl;
        myRuntimeInfo->debug << "Total fuel: " << theTransportation.back().fuel << std::endl;
    }
}

void Production::calculateProductionEmissions(const std::shared_ptr<SchaeferIOData>& data, const std::shared_ptr<RuntimeIO>& rtIO) {
    /* Manufacturing and raw material */
    if (data->configuration.productionMode == "mode_0") { //in case with engines: always mode_1! s. settings
        //Material mode
        for (int i(0); i < numberOfMaterials; i++) {
            std::string materialEntry(theMaterial.at(i).matName);
            std::vector<std::string> matName_tmp;
            for (unsigned int k(0.); k < myEcoDatabase.matEmData.size(); k++) {
                matName_tmp.push_back(myEcoDatabase.matEmData.at(k).specification);
            }
            std::vector<std::string>::iterator entry = find(matName_tmp.begin(), matName_tmp.end(), materialEntry); //Iterator to find the configuration input
            if (entry == matName_tmp.end()) { //Input does not exist!
                myRuntimeInfo->err << "Material " << materialEntry << " not found. Abort!" << std::endl;
                exit(1);
            } else {
                unsigned int pos(0);
                pos = static_cast<int>(entry - matName_tmp.begin());
                /* emission calculation */
                materialProduction.push_back(Emissions());
                materialProcessing.push_back(Emissions());
                materialProduction.back().CO2 = theMaterial.at(i).massOre * myEcoDatabase.matEmData.at(pos).CO2; //raw material
                materialProcessing.back().CO2 = theMaterial.at(i).E_manuf * myEcoDatabase.elecEmData.CO2; //manufacturing
                materialProduction.back().H2O = theMaterial.at(i).massOre * myEcoDatabase.matEmData.at(pos).H2O;
                materialProcessing.back().H2O = theMaterial.at(i).E_manuf * myEcoDatabase.elecEmData.H2O;
                materialProduction.back().NOx = theMaterial.at(i).massOre * myEcoDatabase.matEmData.at(pos).NOx;
                materialProcessing.back().NOx = theMaterial.at(i).E_manuf * myEcoDatabase.elecEmData.NOx;
                materialProduction.back().SO2 = theMaterial.at(i).massOre * myEcoDatabase.matEmData.at(pos).SO2;
                materialProcessing.back().SO2 = theMaterial.at(i).E_manuf * myEcoDatabase.elecEmData.SO2;
                materialProduction.back().CO = theMaterial.at(i).massOre * myEcoDatabase.matEmData.at(pos).CO;
                materialProcessing.back().CO = theMaterial.at(i).E_manuf * myEcoDatabase.elecEmData.CO;
                materialProduction.back().CH4 = theMaterial.at(i).massOre * myEcoDatabase.matEmData.at(pos).CH4;
                materialProcessing.back().CH4 = theMaterial.at(i).E_manuf * myEcoDatabase.elecEmData.CH4;
                materialProduction.back().PM10 = theMaterial.at(i).massOre * myEcoDatabase.matEmData.at(pos).PM10;
                materialProcessing.back().PM10 = theMaterial.at(i).E_manuf * myEcoDatabase.elecEmData.PM10;
                materialProduction.back().CED = theMaterial.at(i).massOre * myEcoDatabase.matSpecificEnergy.at(pos);
            }
            materialProduction_total.CO2  += materialProduction.back().CO2;
            materialProduction_total.H2O  += materialProduction.back().H2O;
            materialProduction_total.NOx  += materialProduction.back().NOx;
            materialProduction_total.SO2  += materialProduction.back().SO2;
            materialProduction_total.CO   += materialProduction.back().CO;
            materialProduction_total.CH4  += materialProduction.back().CH4;
            materialProduction_total.PM10 += materialProduction.back().PM10;
            materialProduction_total.N2O += materialProduction.back().N2O;
            materialProduction_total.PFC += materialProduction.back().PFC;
            materialProduction_total.NMVOC += materialProduction.back().NMVOC;
            materialProduction_total.HC += materialProduction.back().HC;
            materialProduction_total.CED += materialProduction.back().CED;
            materialProcessing_total.CO2  += materialProcessing.back().CO2;
            materialProcessing_total.H2O  += materialProcessing.back().H2O;
            materialProcessing_total.NOx  += materialProcessing.back().NOx;
            materialProcessing_total.SO2  += materialProcessing.back().SO2;
            materialProcessing_total.CO   += materialProcessing.back().CO;
            materialProcessing_total.CH4  += materialProcessing.back().CH4;
            materialProcessing_total.PM10 += materialProcessing.back().PM10;
            materialProcessing_total.N2O += materialProcessing.back().N2O;
            materialProcessing_total.PFC += materialProcessing.back().PFC;
            materialProcessing_total.NMVOC += materialProcessing.back().NMVOC;
            materialProcessing_total.HC += materialProcessing.back().HC;
            materialProcessing_total.CED = E_manuf_total * myEcoDatabase.elecEmData.specificEnergy;
        }
        myRuntimeInfo->info << "    Material Manufacturing Energy Total: " << materialProcessing_total.CED + materialProduction_total.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing Energy: " << E_manuf_total << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing Primary Energy: " << materialProcessing_total.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing CO2: " << materialProcessing_total.CO2 << " kg" << std::endl;
        myRuntimeInfo->info << "      Material Extraction Energy: " << materialProduction_total.CED << " MJ" << std::endl;
    } else if (data->configuration.productionMode == "mode_1") {
        //Main Parts mode
        for (int i(0); i < data->configuration.partsNo; i++) {
            /* Emission calculation */
            //manufacturing
            materialProduction.push_back(Emissions());
            materialProcessing.push_back(Emissions());
            materialProcessing.back().CO2 = theAcComp.at(i).E_manuf * myEcoDatabase.elecEmData.CO2;
            materialProcessing.back().H2O = theAcComp.at(i).E_manuf * myEcoDatabase.elecEmData.H2O;
            materialProcessing.back().NOx = theAcComp.at(i).E_manuf * myEcoDatabase.elecEmData.NOx;
            materialProcessing.back().SO2 = theAcComp.at(i).E_manuf * myEcoDatabase.elecEmData.SO2;
            materialProcessing.back().CO = theAcComp.at(i).E_manuf * myEcoDatabase.elecEmData.CO;
            materialProcessing.back().CH4 = theAcComp.at(i).E_manuf * myEcoDatabase.elecEmData.CH4;
            materialProcessing.back().PM10 = theAcComp.at(i).E_manuf * myEcoDatabase.elecEmData.PM10;
            materialProcessing.back().CED = theAcComp.at(i).E_manuf * myEcoDatabase.elecEmData.specificEnergy;
            //raw material extraction
            for (int j(0); j < numberOfMaterials; j++) {
                std::string materialEntry(theMaterial.at(j).matName);
                std::vector<std::string> matName_tmp;
                for (unsigned int k(0.); k < myEcoDatabase.matEmData.size(); k++) {
                    matName_tmp.push_back(myEcoDatabase.matEmData.at(k).specification);
                }
                std::vector<std::string>::iterator entry = find(matName_tmp.begin(), matName_tmp.end(), materialEntry); //Iterator to find the configuration input
                if (entry == matName_tmp.end()) { //Input not available!
                    myRuntimeInfo->err << "Material " << materialEntry << " in EcoDatabase not found. Abort!" << std::endl;
                    exit(1);
                } else {
                    unsigned int pos(0);
                    pos = static_cast<int>(entry - matName_tmp.begin());
                    myRuntimeInfo->info << "Ore Mat. at " << j << " :" << theAcComp.at(i).oreMaterial.at(j) << std::endl;
                    myRuntimeInfo->info << "CED at " << j << " :" << myEcoDatabase.matSpecificEnergy.at(pos) << std::endl;
                    /* emission calculation */
                    materialProduction.back().CO2 += theAcComp.at(i).oreMaterial.at(j) * myEcoDatabase.matEmData.at(pos).CO2;
                    materialProduction.back().H2O += theAcComp.at(i).oreMaterial.at(j) * myEcoDatabase.matEmData.at(pos).H2O;
                    materialProduction.back().NOx += theAcComp.at(i).oreMaterial.at(j) * myEcoDatabase.matEmData.at(pos).NOx;
                    materialProduction.back().SO2 += theAcComp.at(i).oreMaterial.at(j) * myEcoDatabase.matEmData.at(pos).SO2;
                    materialProduction.back().CO += theAcComp.at(i).oreMaterial.at(j) * myEcoDatabase.matEmData.at(pos).CO;
                    materialProduction.back().CH4 += theAcComp.at(i).oreMaterial.at(j) * myEcoDatabase.matEmData.at(pos).CH4;
                    materialProduction.back().PM10 += theAcComp.at(i).oreMaterial.at(j) * myEcoDatabase.matEmData.at(pos).PM10;
                    materialProduction.back().CED += theAcComp.at(i).oreMaterial.at(j) * myEcoDatabase.matSpecificEnergy.at(pos);
                }
            }
            materialProduction_total.CO2 += materialProduction.back().CO2;
            materialProduction_total.H2O += materialProduction.back().H2O;
            materialProduction_total.NOx += materialProduction.back().NOx;
            materialProduction_total.SO2 += materialProduction.back().SO2;
            materialProduction_total.CO += materialProduction.back().CO;
            materialProduction_total.CH4 += materialProduction.back().CH4;
            materialProduction_total.PM10 += materialProduction.back().PM10;
            materialProduction_total.N2O += materialProduction.back().N2O;
            materialProduction_total.PFC += materialProduction.back().PFC;
            materialProduction_total.NMVOC += materialProduction.back().NMVOC;
            materialProduction_total.HC += materialProduction.back().HC;
            materialProduction_total.CED += materialProduction.back().CED;
            materialProcessing_total.CO2 += materialProcessing.back().CO2;
            materialProcessing_total.H2O += materialProcessing.back().H2O;
            materialProcessing_total.NOx += materialProcessing.back().NOx;
            materialProcessing_total.SO2 += materialProcessing.back().SO2;
            materialProcessing_total.CO += materialProcessing.back().CO;
            materialProcessing_total.CH4 += materialProcessing.back().CH4;
            materialProcessing_total.PM10 += materialProcessing.back().PM10;
            materialProcessing_total.N2O += materialProcessing.back().N2O;
            materialProcessing_total.PFC += materialProcessing.back().PFC;
            materialProcessing_total.NMVOC += materialProcessing.back().NMVOC;
            materialProcessing_total.HC += materialProcessing.back().HC;
            materialProcessing_total.CED += materialProcessing.back().CED;
        }
        if (data->configuration.engineMode) {
            materialProduction_engine.CO2 = materialProduction.back().CO2;
            materialProduction_engine.H2O = materialProduction.back().H2O;
            materialProduction_engine.NOx = materialProduction.back().NOx;
            materialProduction_engine.SO2 = materialProduction.back().SO2;
            materialProduction_engine.CO = materialProduction.back().CO;
            materialProduction_engine.CH4 = materialProduction.back().CH4;
            materialProduction_engine.PM10 = materialProduction.back().PM10;
            materialProduction_engine.N2O = materialProduction.back().N2O;
            materialProduction_engine.PFC = materialProduction.back().PFC;
            materialProduction_engine.NMVOC = materialProduction.back().NMVOC;
            materialProduction_engine.HC = materialProduction.back().HC;
            materialProduction_engine.CED = materialProduction.back().CED;
            materialProcessing_engine.CO2 = materialProcessing.back().CO2;
            materialProcessing_engine.H2O = materialProcessing.back().H2O;
            materialProcessing_engine.NOx = materialProcessing.back().NOx;
            materialProcessing_engine.SO2 = materialProcessing.back().SO2;
            materialProcessing_engine.CO = materialProcessing.back().CO;
            materialProcessing_engine.CH4 = materialProcessing.back().CH4;
            materialProcessing_engine.PM10 = materialProcessing.back().PM10;
            materialProcessing_engine.N2O = materialProcessing.back().N2O;
            materialProcessing_engine.PFC = materialProcessing.back().PFC;
            materialProcessing_engine.NMVOC = materialProcessing.back().NMVOC;
            materialProcessing_engine.HC = materialProcessing.back().HC;
            materialProcessing_engine.CED = materialProcessing.back().CED;
        }
        myRuntimeInfo->info << "    Material Manufacturing Energy Total: " << materialProcessing_total.CED + materialProduction_total.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Extraction Energy: " << materialProduction_total.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing Primary Energy: " << materialProcessing_total.CED << " MJ" << std::endl;
        myRuntimeInfo->info << "      Material Processing Energy: " << E_manuf_total << " MJ" << std::endl;
        myRuntimeInfo->info << "    Material Processing CO2: " << materialProcessing_total.CO2 << " kg" << std::endl;
    } else {
        myRuntimeInfo->err << "Unknown production mode " << data->configuration.productionMode << ". (config-entry \"productionMode\") " << std::endl;
        myRuntimeInfo->err << "Abort program!" << std::endl;
        exit(1);
    }
    /* Personnel expenses for production */
    /*Manufacturing Labour*/
    /* emission calculation */
    labourProduction_elec.CO2  = myEcoDatabase.elecEmData.CO2 *  (theManufacturingWork.E_elec);
    labourProduction_elec.H2O  = myEcoDatabase.elecEmData.H2O *  (theManufacturingWork.E_elec);
    labourProduction_elec.NOx  = myEcoDatabase.elecEmData.NOx *  (theManufacturingWork.E_elec);
    labourProduction_elec.SO2  = myEcoDatabase.elecEmData.SO2 *  (theManufacturingWork.E_elec);
    labourProduction_elec.CO  = myEcoDatabase.elecEmData.CO *  (theManufacturingWork.E_elec);
    labourProduction_elec.CH4  = myEcoDatabase.elecEmData.CH4 *  (theManufacturingWork.E_elec);
    labourProduction_elec.PM10  = myEcoDatabase.elecEmData.PM10 *  (theManufacturingWork.E_elec);
    labourProduction_elec.CED = myEcoDatabase.elecEmData.specificEnergy *  (theManufacturingWork.E_elec);
    labourProduction_heat.CO2  = myEcoDatabase.heatEmData.CO2 *  (theManufacturingWork.E_heat);
    labourProduction_heat.H2O  = myEcoDatabase.heatEmData.H2O *  (theManufacturingWork.E_heat);
    labourProduction_heat.NOx  = myEcoDatabase.heatEmData.NOx *  (theManufacturingWork.E_heat);
    labourProduction_heat.SO2  = myEcoDatabase.heatEmData.SO2 *  (theManufacturingWork.E_heat);
    labourProduction_heat.CO  = myEcoDatabase.heatEmData.CO *  (theManufacturingWork.E_heat);
    labourProduction_heat.CH4  = myEcoDatabase.heatEmData.CH4 *  (theManufacturingWork.E_heat);
    labourProduction_heat.PM10  = myEcoDatabase.heatEmData.PM10 *  (theManufacturingWork.E_heat);
    labourProduction_heat.CED = myEcoDatabase.heatEmData.specificEnergy *  (theManufacturingWork.E_heat);
    //Total assembly emissions
    myRuntimeInfo->info << "    Labour Production Energy: " << theManufacturingWork.E_elec + theManufacturingWork.E_heat << " MJ" << std::endl;
    labourProduction_total.CO2 = labourProduction_elec.CO2 + labourProduction_heat.CO2;
    labourProduction_total.H2O = labourProduction_elec.H2O + labourProduction_heat.H2O;
    labourProduction_total.NOx = labourProduction_elec.NOx + labourProduction_heat.NOx;
    labourProduction_total.SO2 = labourProduction_elec.SO2 + labourProduction_heat.SO2;
    labourProduction_total.CO = labourProduction_elec.CO + labourProduction_heat.CO;
    labourProduction_total.CH4 = labourProduction_elec.CH4 + labourProduction_heat.CH4;
    labourProduction_total.PM10 = labourProduction_elec.PM10 + labourProduction_heat.PM10;
    labourProduction_total.N2O = labourProduction_elec.N2O + labourProduction_heat.N2O;
    labourProduction_total.PFC = labourProduction_elec.PFC + labourProduction_heat.PFC;
    labourProduction_total.NMVOC = labourProduction_elec.NMVOC + labourProduction_heat.NMVOC;
    labourProduction_total.HC = labourProduction_elec.HC + labourProduction_heat.HC;
    labourProduction_total.CED = labourProduction_elec.CED + labourProduction_heat.CED;
    myRuntimeInfo->info << "    Labour Production Primary Energy: " << labourProduction_total.CED << " MJ" << std::endl;
    if (data->configuration.engineMode) {
        labourProduction_elec_engine.CO2  = myEcoDatabase.elecEmData.CO2 *  (theManufacturingWork.E_elec_eng);
        labourProduction_elec_engine.H2O  = myEcoDatabase.elecEmData.H2O *  (theManufacturingWork.E_elec_eng);
        labourProduction_elec_engine.NOx  = myEcoDatabase.elecEmData.NOx *  (theManufacturingWork.E_elec_eng);
        labourProduction_elec_engine.SO2  = myEcoDatabase.elecEmData.SO2 *  (theManufacturingWork.E_elec_eng);
        labourProduction_elec_engine.CO  = myEcoDatabase.elecEmData.CO *  (theManufacturingWork.E_elec_eng);
        labourProduction_elec_engine.CH4  = myEcoDatabase.elecEmData.CH4 *  (theManufacturingWork.E_elec_eng);
        labourProduction_elec_engine.PM10  = myEcoDatabase.elecEmData.PM10 *  (theManufacturingWork.E_elec_eng);
        labourProduction_elec_engine.CED = myEcoDatabase.elecEmData.specificEnergy *  (theManufacturingWork.E_elec_eng);
        labourProduction_heat_engine.CO2  = myEcoDatabase.heatEmData.CO2 *  (theManufacturingWork.E_heat_eng);
        labourProduction_heat_engine.H2O  = myEcoDatabase.heatEmData.H2O *  (theManufacturingWork.E_heat_eng);
        labourProduction_heat_engine.NOx  = myEcoDatabase.heatEmData.NOx *  (theManufacturingWork.E_heat_eng);
        labourProduction_heat_engine.SO2  = myEcoDatabase.heatEmData.SO2 *  (theManufacturingWork.E_heat_eng);
        labourProduction_heat_engine.CO  = myEcoDatabase.heatEmData.CO *  (theManufacturingWork.E_heat_eng);
        labourProduction_heat_engine.CH4  = myEcoDatabase.heatEmData.CH4 *  (theManufacturingWork.E_heat_eng);
        labourProduction_heat_engine.PM10  = myEcoDatabase.heatEmData.PM10 *  (theManufacturingWork.E_heat_eng);
        labourProduction_heat_engine.CED = myEcoDatabase.heatEmData.specificEnergy *  (theManufacturingWork.E_heat_eng);
        //Total assembly emissions
        labourProduction_engine.CO2 = labourProduction_elec_engine.CO2 + labourProduction_heat_engine.CO2;
        labourProduction_engine.H2O = labourProduction_elec_engine.H2O + labourProduction_heat_engine.H2O;
        labourProduction_engine.NOx = labourProduction_elec_engine.NOx + labourProduction_heat_engine.NOx;
        labourProduction_engine.SO2 = labourProduction_elec_engine.SO2 + labourProduction_heat_engine.SO2;
        labourProduction_engine.CO = labourProduction_elec_engine.CO + labourProduction_heat_engine.CO;
        labourProduction_engine.CH4 = labourProduction_elec_engine.CH4 + labourProduction_heat_engine.CH4;
        labourProduction_engine.PM10 = labourProduction_elec_engine.PM10 + labourProduction_heat_engine.PM10;
        labourProduction_engine.N2O = labourProduction_elec_engine.N2O + labourProduction_heat_engine.N2O;
        labourProduction_engine.PFC = labourProduction_elec_engine.PFC + labourProduction_heat_engine.PFC;
        labourProduction_engine.NMVOC = labourProduction_elec_engine.NMVOC + labourProduction_heat_engine.NMVOC;
        labourProduction_engine.HC = labourProduction_elec_engine.HC + labourProduction_heat_engine.HC;
        labourProduction_engine.CED = labourProduction_elec_engine.CED + labourProduction_heat_engine.CED;
        myRuntimeInfo->info << "    Labour Production Primary Energy Engine: " <<  labourProduction_engine.CED << " MJ" << std::endl;
    }
    /*Final Assembly*/
    /* emission calculation */
    finalAssembly_elec.CO2  = myEcoDatabase.elecEmData.CO2 *  (theFinalAssembly.E_elec);
    finalAssembly_elec.H2O  = myEcoDatabase.elecEmData.H2O *  (theFinalAssembly.E_elec);
    finalAssembly_elec.NOx  = myEcoDatabase.elecEmData.NOx *  (theFinalAssembly.E_elec);
    finalAssembly_elec.SO2  = myEcoDatabase.elecEmData.SO2 *  (theFinalAssembly.E_elec);
    finalAssembly_elec.CO  = myEcoDatabase.elecEmData.CO *  (theFinalAssembly.E_elec);
    finalAssembly_elec.CH4  = myEcoDatabase.elecEmData.CH4 *  (theFinalAssembly.E_elec);
    finalAssembly_elec.PM10  = myEcoDatabase.elecEmData.PM10 *  (theFinalAssembly.E_elec);
    finalAssembly_elec.CED = myEcoDatabase.elecEmData.specificEnergy *  (theFinalAssembly.E_elec);
    finalAssembly_heat.CO2  = myEcoDatabase.heatEmData.CO2 *  (theFinalAssembly.E_heat);
    finalAssembly_heat.H2O  = myEcoDatabase.heatEmData.H2O *  (theFinalAssembly.E_heat);
    finalAssembly_heat.NOx  = myEcoDatabase.heatEmData.NOx *  (theFinalAssembly.E_heat);
    finalAssembly_heat.SO2  = myEcoDatabase.heatEmData.SO2 *  (theFinalAssembly.E_heat);
    finalAssembly_heat.CO  = myEcoDatabase.heatEmData.CO *  (theFinalAssembly.E_heat);
    finalAssembly_heat.CH4  = myEcoDatabase.heatEmData.CH4 *  (theFinalAssembly.E_heat);
    finalAssembly_heat.PM10  = myEcoDatabase.heatEmData.PM10 *  (theFinalAssembly.E_heat);
    finalAssembly_heat.CED = myEcoDatabase.heatEmData.specificEnergy *  (theFinalAssembly.E_heat);
    //Final Assembly emissions
    myRuntimeInfo->info << "    Final Assembly Energy: " << theFinalAssembly.E_elec + theFinalAssembly.E_heat << " MJ" << std::endl;
    finalAssembly_total.CO2 = finalAssembly_elec.CO2 + finalAssembly_heat.CO2;
    finalAssembly_total.H2O = finalAssembly_elec.H2O + finalAssembly_heat.H2O;
    finalAssembly_total.NOx = finalAssembly_elec.NOx + finalAssembly_heat.NOx;
    finalAssembly_total.SO2 = finalAssembly_elec.SO2 + finalAssembly_heat.SO2;
    finalAssembly_total.CO = finalAssembly_elec.CO + finalAssembly_heat.CO;
    finalAssembly_total.CH4 = finalAssembly_elec.CH4 + finalAssembly_heat.CH4;
    finalAssembly_total.PM10 = finalAssembly_elec.PM10 + finalAssembly_heat.PM10;
    finalAssembly_total.N2O = finalAssembly_elec.N2O + finalAssembly_heat.N2O;
    finalAssembly_total.PFC = finalAssembly_elec.PFC + finalAssembly_heat.PFC;
    finalAssembly_total.NMVOC = finalAssembly_elec.NMVOC + finalAssembly_heat.NMVOC;
    finalAssembly_total.HC = finalAssembly_elec.HC + finalAssembly_heat.HC;
    finalAssembly_total.CED = finalAssembly_elec.CED + finalAssembly_heat.CED;
    myRuntimeInfo->info << "    Final Assembly Primary Energy: " << finalAssembly_total.CED << " MJ" << std::endl;
    /* transport logistics */
/*TODO(franz#1#): T141 here assumption that values WITHOUT fuel prod., therefore additional addition of fuel prod. here!
 (not completely comprehensible in GaBi, but probably correct)*/
    int fuelTypeNo(99);
    for (unsigned int i(0.); i < numberOfTransportationTypes; i++) {
        for (unsigned int j(0.); j < theTransportation.at(i).mass_component.size(); j++) {
            double mass_tmp(theTransportation.at(i).mass_component.at(j));
            double distance_tmp(theTransportation.at(i).distance_component.at(j));
            /* emission calculation */
            transport.CO2 += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).CO2 / 1000.;
            transport.H2O += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).H2O / 1000.;
            transport.NOx += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).NOx / 1000.;
            transport.SO2 += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).SO2 / 1000.;
            transport.CO += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).CO / 1000.;
            transport.CH4 += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).CH4 / 1000.;
            transport.PM10 += mass_tmp * distance_tmp * myEcoDatabase.transpEmData.at(i).PM10 / 1000.;
            transport.CED += mass_tmp * distance_tmp * myEcoDatabase.transportSpecificEnergy.at(i) / 1000.;
        }
        // TODO(franz#1#): T140 should be programmed more generically... settings must be set differently, that means of transport+spec.consumption+fuel type belong together
        if  (theTransportation.at(i).transpType == "Aircraft") {
            fuelTypeNo = 0;
        } else if  (theTransportation.at(i).transpType == "Truck" || theTransportation.at(i).transpType == "Train") {
            fuelTypeNo = 1;
        } else if  (theTransportation.at(i).transpType == "Ship") {
            fuelTypeNo = 2;
        } else {
            myRuntimeInfo->err << "No transportation type for transport of the production materials specified."
                                  " Please specify 0 for aircraft, 1 for truck, 2 for ship." << std::endl;
            myRuntimeInfo->err << "Program termination!" << std::endl;
            exit(1);
        }
        transport.CO2 += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).CO2;
        transport.H2O += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).H2O;
        transport.NOx += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).NOx;
        transport.SO2 += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).SO2;
        transport.CO += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).CO;
        transport.CH4 += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).CH4;
        transport.PM10 += theTransportation.at(i).fuel * myEcoDatabase.fuelProdEmData.at(fuelTypeNo).PM10;
        transport.CED += theTransportation.at(i).fuel * myEcoDatabase.fuelSpecificEnergy.at(fuelTypeNo);
    }
    myRuntimeInfo->info << "    Transport to Final Assembly Primary Energy: " << transport.CED << " MJ" << std::endl;
    /* test flights incl. Fuel Prod. */
    /* emission calculation */
    double testFlightFactor = finalFlightTestHours / data->flightTimeDesign;
    testFlightFuel = (data->blockfuelDesign) * testFlightFactor;
    myRuntimeInfo->info << "    Final Test Flight Fuel: " << testFlightFuel << " kg" << std::endl;
    finalFlightTests.CO2 = data->theTestMission.LTO.CO2 + data->theTestMission.Cruise.CO2 * testFlightFactor + testFlightFuel * myEcoDatabase.fuelProdEmData.at(0).CO2;;
    finalFlightTests.H2O = data->theTestMission.LTO.H2O + data->theTestMission.Cruise.H2O * testFlightFactor + testFlightFuel * myEcoDatabase.fuelProdEmData.at(0).H2O;
    finalFlightTests.NOx = data->theTestMission.LTO.NOx + data->theTestMission.Cruise.NOx * testFlightFactor + testFlightFuel * myEcoDatabase.fuelProdEmData.at(0).NOx;
    finalFlightTests.SO2 = data->theTestMission.LTO.SO2 + data->theTestMission.Cruise.SO2 * testFlightFactor + testFlightFuel * myEcoDatabase.fuelProdEmData.at(0).SO2;
    finalFlightTests.CO = data->theTestMission.LTO.CO + data->theTestMission.Cruise.CO * testFlightFactor + testFlightFuel * myEcoDatabase.fuelProdEmData.at(0).CO;
    finalFlightTests.CH4 = testFlightFuel * myEcoDatabase.fuelProdEmData.at(0).CH4;
    finalFlightTests.PM10 = data->theTestMission.LTO.soot + data->theTestMission.Cruise.soot * testFlightFactor + testFlightFuel * myEcoDatabase.fuelProdEmData.at(0).PM10;
    finalFlightTests.HC = data->theTestMission.LTO.HC + data->theTestMission.Cruise.HC * testFlightFactor;
    finalFlightTests.CED = testFlightFuel * myEcoDatabase.fuelSpecificEnergy.at(0);
    myRuntimeInfo->info << "  Production Energy (w/o Material): " << Energy_total << " MJ" << std::endl;
    myRuntimeInfo->info << "  Production Fuel (w/o Transport): " << testFlightFuel << " kg" << std::endl;
    production_total.CO2 = materialProduction_total.CO2 + materialProcessing_total.CO2 + labourProduction_total.CO2 + finalAssembly_total.CO2 + transport.CO2 +
                           finalFlightTests.CO2;
    production_total.H2O = materialProduction_total.H2O + materialProcessing_total.H2O + labourProduction_total.H2O + finalAssembly_total.H2O + transport.H2O +
                           finalFlightTests.H2O;
    production_total.NOx = materialProduction_total.NOx + materialProcessing_total.NOx + labourProduction_total.NOx + finalAssembly_total.NOx + transport.NOx +
                           finalFlightTests.NOx;
    production_total.SO2 = materialProduction_total.SO2 + materialProcessing_total.SO2 + labourProduction_total.SO2 + finalAssembly_total.SO2 + transport.SO2 +
                           finalFlightTests.SO2;
    production_total.CO = materialProduction_total.CO + materialProcessing_total.CO + labourProduction_total.CO
                          + finalAssembly_total.CO + transport.CO + finalFlightTests.CO;
    production_total.CH4 = materialProduction_total.CH4 + materialProcessing_total.CH4 + labourProduction_total.CH4 + finalAssembly_total.CH4 + transport.CH4 +
                           finalFlightTests.CH4;
    production_total.PM10 = materialProduction_total.PM10 + materialProcessing_total.PM10 + labourProduction_total.PM10 + finalAssembly_total.PM10 + transport.PM10 +
                            finalFlightTests.PM10;
    production_total.N2O = materialProduction_total.N2O + materialProcessing_total.N2O + labourProduction_total.N2O + finalAssembly_total.N2O + transport.N2O +
                           finalFlightTests.N2O;
    production_total.PFC = materialProduction_total.PFC + materialProcessing_total.PFC + labourProduction_total.PFC + finalAssembly_total.PFC + transport.PFC +
                           finalFlightTests.PFC;
    production_total.NMVOC = materialProduction_total.NMVOC + materialProcessing_total.NMVOC + labourProduction_total.NMVOC + finalAssembly_total.NMVOC + transport.NMVOC +
                             finalFlightTests.NMVOC;
    production_total.HC = materialProduction_total.HC + materialProcessing_total.HC + labourProduction_total.HC
                          + finalAssembly_total.HC + transport.HC + finalFlightTests.HC;
    production_total.CED = materialProduction_total.CED + materialProcessing_total.CED + labourProduction_total.CED
                           + finalAssembly_total.CED + transport.CED + finalFlightTests.CED;
    myRuntimeInfo->info << "  ProductionPrimary Energy: " << production_total.CED << " MJ" << std::endl;
    //For deduction at static_AC
    if (data->configuration.engineMode) {
        production_engine.CO2 = materialProduction_engine.CO2 + materialProcessing_engine.CO2 + labourProduction_engine.CO2;
        production_engine.H2O = materialProduction_engine.H2O + materialProcessing_engine.H2O + labourProduction_engine.H2O;
        production_engine.NOx = materialProduction_engine.NOx + materialProcessing_engine.NOx + labourProduction_engine.NOx;
        production_engine.SO2 = materialProduction_engine.SO2 + materialProcessing_engine.SO2 + labourProduction_engine.SO2;
        production_engine.CO = materialProduction_engine.CO + materialProcessing_engine.CO + labourProduction_engine.CO;
        production_engine.CH4 = materialProduction_engine.CH4 + materialProcessing_engine.CH4 + labourProduction_engine.CH4;
        production_engine.PM10 = materialProduction_engine.PM10 + materialProcessing_engine.PM10 + labourProduction_engine.PM10;
        production_engine.N2O = materialProduction_engine.N2O + materialProcessing_engine.N2O + labourProduction_engine.N2O;
        production_engine.PFC = materialProduction_engine.PFC + materialProcessing_engine.PFC + labourProduction_engine.PFC;
        production_engine.NMVOC = materialProduction_engine.NMVOC + materialProcessing_engine.NMVOC + labourProduction_engine.NMVOC;
        production_engine.HC = materialProduction_engine.HC + materialProcessing_engine.HC + labourProduction_engine.HC;
        production_engine.CED = materialProduction_engine.CED + materialProcessing_engine.CED + labourProduction_engine.CED;
    }
}

double Production::getNASACost(double Mass, double Quantity, double A0, double A1, double A2) {
    //Beltramo: 10% Earnings are already included in the cost calculation, S. 2-3
    //Beltramo: also system-level assembly (up to 21%) included (-> 1.1*1.21 = 1.33; for Tail, Nacelle, Thrust Reservers and Engine System only 1.21!), S. 2-13 f. S.4-4, 4-14, 4-18
    return 0.9 * A0 * pow(convertUnit(KILO, GRAM, POUND, Mass), A1) * pow(static_cast<double>(Quantity), A2);
}

/* Constuctors and destructors */
Production::~Production() {}

Production::mainComponent::mainComponent(const std::string& name)
    :
    compName(name),
    mass(0.),
    E_manuf(0.),
    RC(0.) {
}
Production::mainComponent::~mainComponent() {}

Production::material::material()
    :
    mass(0.),
    massOre(0),
    E_manuf(0.) {
}

Production::material::~material() {}

Production::manufacturingWork::manufacturingWork()
    :
    labourCost(0.),
    E_elec(0.),
    E_elec_eng(0.),
    E_heat(0.),
    E_heat_eng(0.) {
}

Production::manufacturingWork::~manufacturingWork() {
}

Production::Assembly::Assembly()
    :
    costs(0.),
    numberOfStaff(0.),
    E_elec(0.),
    E_heat(0.) {
}

Production::Assembly::~Assembly() {
}

Production::TransportLogistics::TransportLogistics(const std::string& name)
    :
    transpType(name),
    mass_total(0.),
    distance_total(0.),
    fuel(0.) {
}

Production::TransportLogistics::~TransportLogistics() {
}
