/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "missionIOData.h"
#include <unordered_map>
#include <aixml/node.h>
#include <aixml/endnode.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>
#include <standardFiles/functions.h>

using std::endl;
using std::ostream;
using std::string;
using std::vector;

MissionIOData::MissionIOData(std::shared_ptr<RuntimeIO> rtIO)
    :
    config(MissionConfig()),
    myEngine(rtIO->getEngineDataDir() + "/" +
            EndnodeReadOnly<std::string>("component_design/propulsion/specific/propulsion/engine/model").read(rtIO->acxml).value(),
            EndnodeReadOnly<double>("component_design/propulsion/specific/propulsion/engine/scale_factor").read(rtIO->acxml).value()),
    flightsPerYear(EndnodeReadOnly<int>("assessment/cost_estimation/operating_cost/direct_operating_cost/flights_per_year_study_mission").read(rtIO->acxml).value()),
    durationOperation(EndnodeReadOnly<int>("requirements_and_specifications/assessment_scenario/duration_operation").read(rtIO->acxml).value_xml()),
    numberOfEngines(0) {
        config.readAllConfigData(rtIO->moduleConfig);
        numberOfEngines = rtIO->acxml.getVector("/aircraft_exchange_file/component_design/propulsion/specific/propulsion").size();
        myEngine.set_same_engine_type_quantity(numberOfEngines);
    }


/* Output data methods*/
void MissionIOData::updateEcologicalAssessmentXml(std::shared_ptr<node> ecologicalAssessmentXML, const Mission& thisMission) {
    myRuntimeInfo->out << "  Update ecological_assessment_results.xml" << std::endl;
    std::string subPath;
    subPath = "ecological_data_file/mission_emissions/" + thisMission.typeOfMission +"_mission/emissions/LTO_cycle";
    (Endnode<double>("CO2", "LTO CO2 emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.LTO.CO2, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("H2O", "LTO H2O emissions of one " + thisMission.typeOfMission +" mission",
                         Rounding(thisMission.LTO.H2O, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("SO2", "LTO SO2 emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.LTO.SO2, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("SO4", "LTO SO4 emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.LTO.SO4, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("HC", "LTO HC emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.LTO.HC, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("CH4", "LTO CH4 emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.LTO.CH4, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("CO", "LTO CO emissions of one " + thisMission.typeOfMission +" mission",
                         Rounding(thisMission.LTO.CO, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("NOx", "LTO NOx emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.LTO.NOx, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("soot", "LTO Soot emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.LTO.soot, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("c_soot_LTO_max", "Maximum soot concentration during " + thisMission.typeOfMission +" mission LTO",
                        Rounding(thisMission.c_soot_LTO_max, 3)*pow(10, -6), "1")).update((*ecologicalAssessmentXML)[subPath]);
    subPath = "ecological_data_file/mission_emissions/" + thisMission.typeOfMission +"_mission/emissions/cruise";
    (Endnode<double>("CO2", "Cruise CO2 emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.Cruise.CO2, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("H2O", "Cruise H2O emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.Cruise.H2O, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("SO2", "Cruise SO2 emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.Cruise.SO2, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("SO4", "Cruise SO4 emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.Cruise.SO4, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("HC", "Cruise HC emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.Cruise.HC, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("CH4", "Cruise CH4 emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.Cruise.CH4, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("CO", "Cruise CO emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.Cruise.CO, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("NOx", "Cruise NOx emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.Cruise.NOx, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("soot", "Cruise Soot emissions of one " + thisMission.typeOfMission +" mission",
                        Rounding(thisMission.Cruise.soot, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    // /*ICAO Cruise*/
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/CO2") = Rounding(thisMission.ICAO_LTO.CO2, 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/H2O") = Rounding(thisMission.ICAO_LTO.H2O, 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/SO2") = Rounding(thisMission.ICAO_LTO.SO2, 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/HC") = Rounding(thisMission.ICAO_LTO.HC, 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/CO") = Rounding(thisMission.ICAO_LTO.CO, 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/NOx") = Rounding(thisMission.ICAO_LTO.NOx, 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/PM10") = Rounding(thisMission.ICAO_LTO.soot, 3);
    // /*LTO Times*/
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/Total") = Rounding(thisMission.LTOtime, 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/TaxiTime") = Rounding(thisMission.taxiTime / 60., 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/TakeOffTime") = Rounding(thisMission.takeOffTime / 60., 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/ClimbOutTime") = Rounding(thisMission.climbOutTime / 60., 3);
    // ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/ApproachTime") = Rounding(thisMission.approachTime / 60., 3);
    /* Add descriptions */
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions").addAttrib("description", "Results of mission emission calculation");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/" + thisMission.typeOfMission +"_mission").addAttrib("description",
        "Results of " + thisMission.typeOfMission +" mission calculation");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/" + thisMission.typeOfMission +"_mission/emissions").addAttrib("description",
        "Emissions of " + thisMission.typeOfMission +" mission");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/" + thisMission.typeOfMission +"_mission/emissions/LTO_cycle").addAttrib("description",
        "Emissions of landing and takeoff cycle");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/" + thisMission.typeOfMission +"_mission/emissions/cruise").addAttrib("description",
        "Emissions of cruise phase");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/design_mission").addAttrib("description", "Results of design mission calculation");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/design_mission/emissions").addAttrib("description", "Emissions of design mission ");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/design_mission/emissions/LTO_cycle").addAttrib("description",
        "Emissions of landing and takeoff cycle");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/design_mission/emissions/cruise").addAttrib("description", "Emissions of cruise phase");
}
