# Import standard modules.
import math


def estimate_masses(dict_ac_exchange, dict_mod_config, fuselage_description, runtime_output):
    """ Function to estimate fuselage and cabin design masses.

        References and calculation method:
            [1] Egbert Torenbeek, Synthesis of Subsonic Airplane Design (1982), Delft University Press

            [2] European Union Aviation Safety Agency (2024),  CS-25 Large Aeroplanes - Amendment 28

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict fuselage_description: Dict containing parameter and values of current fuselage design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict fuselage_description: Dict containing parameter and values of current fuselage design
    """
    ''' initialize local parameter '''
    lavatory_mass = 0
    over_wing_exit_flag = False
    emergency_slide_mass = 0
    fuselage_design_masses = {}
    fuselage_mass_additions = 1
    maximum_uld_type_depth = 0
    maximum_cargo_bay_width = 0
    maximum_cargo_bay_height = 0
    maximum_uld_type_empty_mass = 0
    passenger_seat_masses_left_side = 0
    passenger_seat_masses_right_side = 0
    total_number_of_full_size_trolley = 0
    passenger_seat_masses_in_the_center = 0
    fuselage_design_masses['fuselage_structural_masses'] = {'door_mass': 0}
    fuselage_design_masses['fuselage_furnishing_masses'] = {'total_furnishing_group_mass': 0}
    fuselage_design_masses['fuselage_operator_item_masses'] = {'group_masses': {'total_emergency_equipment_masses': 0},
                                                               'total_group_mass': 0,
                                                               'emergency_equipment_masses': {},
                                                               'galley_fixed_structure_mass': 0,
                                                               'wardrobe_fixed_structure_mass': 0,
                                                               'full_size_trolley_catering_mass': 0
                                                               }

    # unit conversion parameter
    convert_kg_to_lb = 2.205
    convert_inch_to_meter = 0.0254
    convert_square_inch_to_square_meter = 0.00064516
    # ToDo muss aus aircraft exchange file kommen!!!
    equivalent_air_design_dive_speed_in_meter_per_second = 320
    horizontal_tail_plane_lever_arm = 0.5 * fuselage_description['geometry_description']['fuselage_dimension']['fuselage_length'] * convert_inch_to_meter

    ''' estimate manoeuvre limit load factors (required for detailed fuselage mass estimation, coming soon) '''
    # Positive limit manoeuvre load factor according to [2] paragraph 25.337.
    # n = 2.1 + (24000 /(W + 10000)) where W is the design maximum take-off weight (lb)
    maximum_limit_manoeuvre_load = \
        round(2.1 + (24000 / (dict_ac_exchange['maximum_take_off_mass'] / convert_kg_to_lb + 10000)), 1)
    if maximum_limit_manoeuvre_load < 2.5:
        maximum_limit_manoeuvre_load = 2.5
    elif maximum_limit_manoeuvre_load > 3.8:
        maximum_limit_manoeuvre_load = 3.8

    # Ultimate manoeuvre load factor according to [1] p.453. (n_ultimate = 1.5 * maximum_limit_manoeuvre_load)
    ultimate_limit_manoeuvre_load = 1.5 * maximum_limit_manoeuvre_load

    ''' estimate fuselage structural masses in accordance to the cpacs standard '''
    fuselage_equivalent_diameter = \
        (fuselage_description['geometry_description']['fuselage_dimension']['fuselage_width'] +
         fuselage_description['geometry_description']['fuselage_dimension']['fuselage_total_height']) / 2
    fuselage_fineness_ratio = \
        round(fuselage_description['geometry_description']['fuselage_dimension']['fuselage_length']
              / fuselage_equivalent_diameter, 4)

    # Estimate fuselage volume according to [1] p.447 (B-5).
    fuselage_volume = round(math.pi / 4 * fuselage_equivalent_diameter ** 2
                            * fuselage_description['geometry_description']['fuselage_dimension']['fuselage_length']
                            * (1 - 2 / fuselage_fineness_ratio), 4)

    # Estimate wetted area according to [1] p.447 (B-6).
    fuselage_wettet_area = round(math.pi * fuselage_equivalent_diameter
                                 * fuselage_description['geometry_description']['fuselage_dimension']['fuselage_length']
                                 * (1 - 2 / fuselage_fineness_ratio) ** (2/3)
                                 * (1 + 1 / fuselage_fineness_ratio ** 2), 4)

    # Fuselage structural mass according to [1] p.282 (8-16)
    fuselage_basic_mass = \
        round(0.23 *
              math.sqrt(equivalent_air_design_dive_speed_in_meter_per_second
                        * (horizontal_tail_plane_lever_arm
                           / ((fuselage_description['geometry_description']['fuselage_dimension']['fuselage_width']
                               + fuselage_description['geometry_description']['fuselage_dimension'][
                                   'fuselage_total_height'])
                               * convert_inch_to_meter)))
                   * (fuselage_wettet_area * convert_square_inch_to_square_meter) ** 1.2, 9)

    if dict_ac_exchange['fuselage_pressurization']:
        fuselage_mass_additions += 0.08
    for key, value in dict_ac_exchange['propulsor_mounting_position'].items():
        if value == 'fuselage':
            fuselage_mass_additions += 0.04
            break
    if dict_ac_exchange['main_gear_mounting_position'] == 'fuselage_mounted':
        fuselage_mass_additions += 0.07

    fuselage_design_masses['fuselage_structural_masses']['fuselage_mass'] = \
        round(fuselage_basic_mass * fuselage_mass_additions
              * dict_mod_config['fuselage_structural_mass_technology_factor'], 9)

    # Estimate fuselage door structure mass
    exits_type_definition_list \
        = dict_ac_exchange['dict_cs_requirements']['fuselage_design_cs_requirements']['exit_definition']['exit']
    for key, value in dict_mod_config['select_specific_exit_types'].items():
        # Check if specific exit types are selected -> if false: -> break the loop and use estimated exit types
        if value is None:
            break
        else:
            for exit_type in exits_type_definition_list:
                if exit_type['name_of_exit']['value'] == value:
                    fuselage_design_masses['fuselage_structural_masses']['door_mass'] += \
                        2 * float(exit_type['mass_of_exit_type']['value'])
                    break

    ''' estimate operator item masses in accordance to the cpacs standard '''
    # Passenger seat mass per class and aircraft side
    for passenger_class in dict_ac_exchange['relevant_classes']['passenger_classes']['class_names']:
        passenger_seat_masses_left_side += \
            (fuselage_description['interior_description'][passenger_class]['number_of_seats_left_side']
             * float(dict_ac_exchange['relevant_classes']['passenger_classes'][passenger_class]['passenger_seats']
                     ['seat_left_side']['mass']['value']))

        passenger_seat_masses_right_side += \
            (fuselage_description['interior_description'][passenger_class]['number_of_seats_right_side']
             * float(dict_ac_exchange['relevant_classes']['passenger_classes'][passenger_class]['passenger_seats']
                     ['seat_left_side']['mass']['value']))

        if dict_ac_exchange['fuselage_type'] == "wide_body":
            passenger_seat_masses_in_the_center += \
                (fuselage_description['interior_description'][passenger_class]['number_of_seats_in_the_center']
                 * float(dict_ac_exchange['relevant_classes']['passenger_classes'][passenger_class]['passenger_seats']
                         ['seat_left_side']['mass']['value']))

        # Galley fixed structure mass per class.
        fuselage_design_masses['fuselage_operator_item_masses']['galley_fixed_structure_mass'] += \
            (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['galley_definition']
                   ['galley_structure_mass_per_full_size_trolley']['value'])
             * fuselage_description['interior_description'][passenger_class]['number_of_full_size_trolley'])

        # Wardrobe fixed structure mass per class.
        if dict_ac_exchange['relevant_classes']['passenger_classes'][passenger_class]['class_wardrobe']['use_wardrobe']:
            fuselage_design_masses['fuselage_operator_item_masses']['wardrobe_fixed_structure_mass'] += \
                (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['galley_definition']
                       ['galley_structure_mass_per_full_size_trolley']['value'])
                 * fuselage_description['interior_description'][passenger_class]['units_of_wardrobe'])

        # Lavatory mass per class.
        lavatory_mass += \
            (float(dict_ac_exchange['relevant_classes']['passenger_classes'][passenger_class]['class_lavatory']
                   ['lavatory']['mass']['value'])
             * fuselage_description['interior_description'][passenger_class]['number_of_class_lavatories'])

        # Estimate full size trolley catering mass incl. dry mass
        total_number_of_full_size_trolley += \
            fuselage_description['interior_description'][passenger_class]['number_of_full_size_trolley']
        fuselage_design_masses['fuselage_operator_item_masses']['full_size_trolley_catering_mass'] += \
            round(fuselage_description['interior_description'][passenger_class]['number_of_full_size_trolley']
                  * (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['trolley_definition']
                                           ['full_size_trolley']['empty_mass']['value'])
                     + float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['trolley_definition']
                                             ['full_size_trolley']['load_mass']['value'])), 9)

    # Estimate standard unit catering mass incl. dry mass
    fuselage_design_masses['fuselage_operator_item_masses']['standard_unit_catering_mass'] = \
        round(total_number_of_full_size_trolley
              * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['trolley_definition']
                                      ['standard_unit_to_full_size_trolley_ratio']['value'])
              * (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['trolley_definition']
                                       ['standard_unit']['empty_mass']['value'])
                  + (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['trolley_definition']
                                           ['standard_unit']['load_mass']['value']))), 9)

    # Estimate total passenger seat mass.
    fuselage_design_masses['fuselage_operator_item_masses']['passenger_seat_mass'] = \
        round(passenger_seat_masses_left_side + passenger_seat_masses_left_side
              + passenger_seat_masses_in_the_center, 9)

    # Estimate in flight entertainment mass
    required_number_of_flight_attendants = \
        math.ceil(dict_ac_exchange['total_number_of_passengers']
                  / int(dict_ac_exchange['dict_cs_requirements']['fuselage_design_cs_requirements']
                        ['allowed_number_of_passenger_per_flight_attendant']['value']))
    fuselage_design_masses['fuselage_operator_item_masses']['group_masses']['flight_attendant_mass'] \
        = required_number_of_flight_attendants * dict_mod_config['flight_attendant_mass']

    # Estimate in flight entertainment mass
    if dict_mod_config['use_in_flight_entertainment']:
        fuselage_design_masses['fuselage_operator_item_masses']['group_masses']['in_flight_entertainment'] = \
            round(dict_ac_exchange['total_number_of_passengers']
                  * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                                          ['in_flight_entertainment']['value']), 9)
    else:
        fuselage_design_masses['fuselage_operator_item_masses']['group_masses']['in_flight_entertainment'] = 0

    # Estimate seat belt mass
    fuselage_design_masses['fuselage_operator_item_masses']['seat_belt_mass'] = \
        (dict_ac_exchange['total_number_of_passengers']
         * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                                 ['seat_belt']['value']))

    # Estimate total fixed galley structure mass
    fuselage_design_masses['fuselage_operator_item_masses']['group_masses']['total_fixed_galley_mass'] = \
        (fuselage_design_masses['fuselage_operator_item_masses']['wardrobe_fixed_structure_mass']
         + fuselage_design_masses['fuselage_operator_item_masses']['galley_fixed_structure_mass'])

    # Estimate total seat mass
    fuselage_design_masses['fuselage_operator_item_masses']['group_masses']['total_seat_mass'] = \
        (fuselage_design_masses['fuselage_operator_item_masses']['passenger_seat_mass']
         + fuselage_design_masses['fuselage_operator_item_masses']['seat_belt_mass'])

    # Estimate total catering mass
    fuselage_design_masses['fuselage_operator_item_masses']['group_masses']['catering_mass'] = \
        (fuselage_design_masses['fuselage_operator_item_masses']['standard_unit_catering_mass']
         + fuselage_design_masses['fuselage_operator_item_masses']['full_size_trolley_catering_mass'])

    # Estimate emergency equipment masses
    # Estimate mass of emergency slides
    for key, value in dict_mod_config['select_specific_exit_types'].items():
        if value is not None:
            if value == 'Type A':
                for i in range(0, len(dict_ac_exchange['dict_accommodation'][
                    'accommodation_definitions']['emergency_slides']['slide'])):
                    if dict_ac_exchange['dict_accommodation'][
                        'accommodation_definitions']['emergency_slides']['slide'][i]['name']['value'] == 'dual_slide':       
                        emergency_slide_mass += round(float(dict_ac_exchange['dict_accommodation'][
                            'accommodation_definitions']['emergency_slides']['slide'][i]['mass']['value']), 2)
            if 'OWE' in value and not over_wing_exit_flag:
                over_wing_exit_flag = True
                for i in range(0, len(dict_ac_exchange['dict_accommodation'][
                    'accommodation_definitions']['emergency_slides']['slide'])):
                    if dict_ac_exchange['dict_accommodation'][
                        'accommodation_definitions']['emergency_slides']['slide'][i]['name']['value'] == 'OWE_slide':       
                        emergency_slide_mass += round(float(dict_ac_exchange['dict_accommodation'][
                            'accommodation_definitions']['emergency_slides']['slide'][i]['mass']['value']), 2)
            elif not 'OWE' in value and value != 'Type A':
                for i in range(0, len(dict_ac_exchange['dict_accommodation'][
                    'accommodation_definitions']['emergency_slides']['slide'])):
                    if dict_ac_exchange['dict_accommodation'][
                        'accommodation_definitions']['emergency_slides']['slide'][i]['name']['value'] == 'single_slide':       
                        emergency_slide_mass += round(float(dict_ac_exchange['dict_accommodation'][
                            'accommodation_definitions']['emergency_slides']['slide'][i]['mass']['value']), 2)
            
    fuselage_design_masses['fuselage_operator_item_masses']['emergency_equipment_masses']['emergency_slide_mass'] = \
        2 * emergency_slide_mass
    # Mass of first aid kits according to EU-OPS 1.745 -> one first aid kit per commenced 100 PAX
    fuselage_design_masses['fuselage_operator_item_masses']['emergency_equipment_masses']['first_aid_kit_mass'] = \
        (math.ceil(dict_ac_exchange['total_number_of_passengers'] / 100)
         * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                 ['first_aid_kit']['value']))

    # Mass of emergency medical kit according to EU-OPS 1.755 -> one emergency medical kit per aircraft
    fuselage_design_masses['fuselage_operator_item_masses']['emergency_equipment_masses']['emergency_medical_kit'] = \
        float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                              ['emergency_medical_kit']['value'])

    # Mass of automated external defibrillator (AED) -> for long range mission two required otherwise one
    if dict_ac_exchange['relevant_classes']['range_scenario'] == 'long_range':
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['automated_external_defibrillator'] = \
            (2 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                      ['automated_external_defibrillator']['value']))
    else:
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['automated_external_defibrillator'] = \
            float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                  ['automated_external_defibrillator']['value'])

    # Mass of flight attendant kit
    fuselage_design_masses['fuselage_operator_item_masses']['emergency_equipment_masses']['flight_attendant_kit'] = \
        (required_number_of_flight_attendants
         * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                 ['flight_attendant_kit']['value']))

    # Mass of electronic flight bag kit -> for long range mission four required otherwise two
    if dict_ac_exchange['relevant_classes']['range_scenario'] == 'long_range':
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['electronic_flight_bag_kit'] = \
            (4 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                       ['electronic_flight_bag_kit']['value']))
    else:
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['electronic_flight_bag_kit'] = \
            (2 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                       ['electronic_flight_bag_kit']['value']))

    #  Mass of PED containment bag -> for long range mission ten required otherwise five
    if dict_ac_exchange['relevant_classes']['range_scenario'] == 'long_range':
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['ped_containment_bag'] = \
            (10 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                       ['ped_containment_bag']['value']))
    else:
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['ped_containment_bag'] = \
            (5 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                       ['ped_containment_bag']['value']))

    # Mass of aircraft wheelchair, one per aircraft
    fuselage_design_masses['fuselage_operator_item_masses'][
        'emergency_equipment_masses']['wheel_chair'] = \
        float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                              ['wheel_chair']['value'])

    #  Mass of child restraint system -> for long range mission 9 required otherwise 3
    if dict_ac_exchange['relevant_classes']['range_scenario'] == 'long_range':
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['child_restraint_system'] = \
            (9 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                       ['child_restraint_system']['value']))
    else:
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['child_restraint_system'] = \
            (3 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                       ['child_restraint_system']['value']))

    #  Mass of baby bassinet system -> for long range mission 5 required otherwise 1
    if dict_ac_exchange['relevant_classes']['range_scenario'] == 'long_range':
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['baby_bassinet'] = \
            (5 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                       ['baby_bassinet']['value']))
    else:
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['baby_bassinet'] = \
            (1 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                       ['baby_bassinet']['value']))

    # Mass of aircraft crew megaphones -> one per floor level exit pair
    fuselage_design_masses['fuselage_operator_item_masses'][
        'emergency_equipment_masses']['aircraft_crew_megaphone'] = \
        math.ceil(float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                        ['aircraft_crew_megaphone']['value'])
                  * required_number_of_flight_attendants / 2)

    # Mass of aircraft crash axes -> two per aircraft
    fuselage_design_masses['fuselage_operator_item_masses'][
        'emergency_equipment_masses']['aircraft_crash_axe'] = \
        2 * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                  ['aircraft_crash_axe']['value'])

    # Mass of aircraft exit lights -> one sign per exit door plus one/two additional per exit pair
    fuselage_design_masses['fuselage_operator_item_masses'][
        'emergency_equipment_masses']['aircraft_exit_light'] = \
        (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                               ['aircraft_exit_light']['value'])
         * 3 * len(dict_mod_config['select_specific_exit_types']))
    if dict_ac_exchange['fuselage_type'] == "wide_body":
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['aircraft_exit_light'] += \
            (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                   ['aircraft_exit_light']['value'])
             * len(dict_mod_config['select_specific_exit_types']))

    # Mass of aircraft floor path marking
    # main aisle
    fuselage_design_masses['fuselage_operator_item_masses'][
        'emergency_equipment_masses']['aircraft_floor_path_marking'] = \
        2 * (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                   ['aircraft_floor_path_marking']['value'])
             * fuselage_description['geometry_description']['total_cabin_length'] * convert_inch_to_meter)

    # second main aisle for wide body aircraft
    if dict_ac_exchange['fuselage_type'] == "wide_body":
        fuselage_design_masses['fuselage_operator_item_masses'][
            'emergency_equipment_masses']['aircraft_floor_path_marking'] = \
            2 * fuselage_design_masses['fuselage_operator_item_masses'][
                'emergency_equipment_masses']['aircraft_floor_path_marking']

    # additions per exit pair
    fuselage_design_masses['fuselage_operator_item_masses'][
        'emergency_equipment_masses']['aircraft_floor_path_marking'] += \
        (2 * (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
                                   ['aircraft_floor_path_marking']['value'])
              * fuselage_description['geometry_description']['total_cabin_width'] * convert_inch_to_meter)
           * len(dict_mod_config['select_specific_exit_types']))

    # Mass of life vests -> one per passenger and flight attendant seat
    fuselage_design_masses['fuselage_operator_item_masses'][
        'emergency_equipment_masses']['life_vest'] = \
        (float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['safety_equipment_masses']
               ['life_vest']['value'])
         * (required_number_of_flight_attendants + dict_ac_exchange['total_number_of_passengers']))

    # Estimate total emergency equipment group mass
    for _, mass in \
            fuselage_design_masses['fuselage_operator_item_masses']['emergency_equipment_masses'].items():
        fuselage_design_masses['fuselage_operator_item_masses']['group_masses']['total_emergency_equipment_masses'] \
            += mass

    # Estimate total operator item group mass
    for _, mass in fuselage_design_masses['fuselage_operator_item_masses']['group_masses'].items():
        fuselage_design_masses['fuselage_operator_item_masses']['total_group_mass'] += mass

    fuselage_design_masses['fuselage_operator_item_masses']['total_group_mass'] = \
        round(fuselage_design_masses['fuselage_operator_item_masses']['total_group_mass'], 9)

    ''' estimate furnishing masses in accordance to the cpacs standard '''
    # total lavatory mass
    fuselage_design_masses['fuselage_furnishing_masses']['lavatory_mass'] = lavatory_mass

    # Estimate total flight attendant seat mass.
    fuselage_design_masses['fuselage_furnishing_masses']['flight_attendant_seat_mass'] = \
        (required_number_of_flight_attendants
         * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                 ['flight_attendant_seat']['value']))

    # Estimate floor covering mass.
    fuselage_design_masses['fuselage_furnishing_masses']['floor_covering_mass'] = \
        round(fuselage_description['geometry_description']['total_cabin_width']
              * fuselage_description['geometry_description']['total_cabin_length'] * convert_square_inch_to_square_meter
              * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                                      ['cabin_floor_covering']['value']), 9)

    # Estimate mass of wall covering for passenger cabin incl. screens, window shades, insulation and soundproofing
    fuselage_design_masses['fuselage_furnishing_masses']['wall_covering_mass'] = \
        round(float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                                    ['cabin_wall_covering']['value'])
              * (fuselage_description['geometry_description']['total_cabin_length']
              * dict_mod_config['passenger_standing_height_seat_row'] * convert_square_inch_to_square_meter) * 2, 9)

    # Estimate ceiling covering mass
    fuselage_design_masses['fuselage_furnishing_masses']['ceiling_covering_mass'] = \
        round(float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                    ['cabin_ceiling_covering']['value'])
              * (fuselage_description['geometry_description']['total_cabin_length'] * convert_inch_to_meter), 9)

    fuselage_design_masses['fuselage_furnishing_masses']['linings_mass'] = \
        (fuselage_design_masses['fuselage_furnishing_masses']['ceiling_covering_mass']
         + fuselage_design_masses['fuselage_furnishing_masses']['wall_covering_mass'])

    # Estimate thermal and acoustic insolation mass for pressurized cabin and cargo decks
    fuselage_design_masses['fuselage_furnishing_masses']['acoustic_and_thermal_insolation_mass'] = \
        round((fuselage_equivalent_diameter * convert_inch_to_meter) * math.pi
              * 2 * (fuselage_description['geometry_description']['total_cabin_length'] * convert_inch_to_meter)
              * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                                      ['acoustic_and_thermal_insolation']['value']), 9)

    # Estimate mass of overhead bins
    fuselage_design_masses['fuselage_furnishing_masses']['overhead_bin_mass'] = \
        round(2 * (fuselage_description['geometry_description']['total_cabin_length'] * convert_inch_to_meter)
              * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                                      ['overhead_bins']['value']), 9)

    # Estimate mass of passenger safety units
    fuselage_design_masses['fuselage_furnishing_masses']['passenger_safety_unit_mass'] = \
        round(dict_ac_exchange['total_number_of_passengers']
              * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                      ['passenger_safety_unit']['value']), 9)

    # Estimate mass of lightning panel
    fuselage_design_masses['fuselage_furnishing_masses']['lightning_panel_mass'] = \
        round(dict_ac_exchange['total_number_of_passengers']
              * float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                      ['lightning_panel']['value']), 9)

    # Estimate additional wide body cabin interior masses
    if dict_ac_exchange['fuselage_type'] == "wide_body":
        fuselage_design_masses['fuselage_furnishing_masses']['linings_mass'] =\
            (fuselage_design_masses['fuselage_furnishing_masses']['linings_mass']
             + fuselage_design_masses['fuselage_furnishing_masses']['ceiling_covering_mass'])
        fuselage_design_masses['fuselage_furnishing_masses']['ceiling_covering_mass'] = \
            2 * fuselage_design_masses['fuselage_furnishing_masses']['ceiling_covering_mass']
        fuselage_design_masses['fuselage_furnishing_masses']['overhead_bin_mass'] = \
            2 * fuselage_design_masses['fuselage_furnishing_masses']['overhead_bin_mass']

    # Estimate mass of oxygen system according to [1] p.291 (table 8-12)
    fuselage_design_masses['fuselage_furnishing_masses']['oxygen_system_mass'] = \
        (round(13.6 + 0.544 * dict_ac_exchange['total_number_of_passengers'], 9)
         + fuselage_design_masses['fuselage_furnishing_masses']['passenger_safety_unit_mass'])

    # Estimate cargo linings and restrains mass for wall, floor and ceiling
    # ToDo: 1/2 * cabin length durch total cargo bay length ersetzen
    for key, value in dict_ac_exchange['relevant_cargo_accommodation']['uld_type'].items():
        maximum_cargo_bay_height \
            = max(maximum_cargo_bay_height,
                  float(dict_ac_exchange['relevant_cargo_accommodation']['uld_type'][key]['height']['value']))
        maximum_cargo_bay_width \
            = max(maximum_cargo_bay_width,
                  float(dict_ac_exchange['relevant_cargo_accommodation']['uld_type'][key]['width_at_top']['value']))
        maximum_uld_type_empty_mass = \
            max(maximum_uld_type_empty_mass,
                float(dict_ac_exchange['relevant_cargo_accommodation']['uld_type'][key]['empty_mass']['value']))
        maximum_uld_type_depth = \
            max(maximum_uld_type_depth,
                float(dict_ac_exchange['relevant_cargo_accommodation']['uld_type'][key]['depth']['value']))
    # wall mass
    fuselage_design_masses['fuselage_furnishing_masses']['cargo_linings_mass'] = \
        round(float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                                    ['cabin_ceiling_covering']['value'])
              * (1/2 * fuselage_description['geometry_description']['total_cabin_length']
              * maximum_cargo_bay_height * convert_square_inch_to_square_meter) * 2, 9)

    # floor and ceiling mass
    fuselage_design_masses['fuselage_furnishing_masses']['cargo_linings_mass'] = \
        (fuselage_design_masses['fuselage_furnishing_masses']['cargo_linings_mass']
         + round(float(dict_ac_exchange['dict_accommodation']['accommodation_definitions']['cabin_interior_masses']
                                       ['cabin_ceiling_covering']['value'])
                 * (1/2 * fuselage_description['geometry_description']['total_cabin_length']
                 * maximum_cargo_bay_width * convert_square_inch_to_square_meter) * 2, 9))

    # cargo restrains mass according to [1] p.291 (table 8-12)
    fuselage_design_masses['fuselage_furnishing_masses']['cargo_linings_mass'] = \
        (fuselage_design_masses['fuselage_furnishing_masses']['cargo_linings_mass']
         + round(1.28 * (1/2 * fuselage_description['geometry_description']['total_cabin_length']
                 * convert_inch_to_meter) * (maximum_cargo_bay_width * maximum_cargo_bay_height
                                             * convert_square_inch_to_square_meter), 9))

    # Estimate cargo loadings mass
    fuselage_design_masses['fuselage_furnishing_masses']['cargo_loadings_mass'] = \
        (math.floor((1/2 * fuselage_description['geometry_description']['total_cabin_length']) / maximum_uld_type_depth)
         * maximum_uld_type_empty_mass)

    # Estimate furnishing group mass without miscellaneous mass
    for furnishing_item, mass in fuselage_design_masses['fuselage_furnishing_masses'].items():
        if furnishing_item == 'ceiling_covering_mass' \
                or furnishing_item == 'wall_covering_mass' \
                or furnishing_item == 'total_furnishing_group_mass'\
                or furnishing_item == 'passenger_safety_unit_mass':
            continue
        fuselage_design_masses['fuselage_furnishing_masses']['total_furnishing_group_mass'] \
            += mass

    # Estimate mounting equipment mass (10 percent of total furnishing group without miscellaneous mass)
    fuselage_design_masses['fuselage_furnishing_masses']['miscellaneous_mass'] =\
        round(0.1 * fuselage_design_masses['fuselage_furnishing_masses']['total_furnishing_group_mass'], 9)

    # Estimate total furnishing group mass
    fuselage_design_masses['fuselage_furnishing_masses']['total_furnishing_group_mass'] = \
        round(fuselage_design_masses['fuselage_furnishing_masses']['total_furnishing_group_mass']
              + fuselage_design_masses['fuselage_furnishing_masses']['miscellaneous_mass'], 9)

    ''' Estimate total fuselage mass '''
    fuselage_design_masses['total_fuselage_mass'] = \
        (fuselage_design_masses['fuselage_structural_masses']['door_mass']
         + fuselage_design_masses['fuselage_structural_masses']['fuselage_mass']
         + fuselage_design_masses['fuselage_furnishing_masses']['total_furnishing_group_mass']
         + fuselage_design_masses['fuselage_operator_item_masses']['total_group_mass'])

    ''' store fuselage masses and total fuselage volume and wetted area fuselage description dictionary '''
    fuselage_description['fuselage_design_masses'] = fuselage_design_masses
    fuselage_description['geometry_description']['required_number_of_flight_attendants'] \
        = required_number_of_flight_attendants
    fuselage_description['geometry_description']['fuselage_dimension']['fuselage_volume'] = fuselage_volume
    fuselage_description['geometry_description']['fuselage_dimension']['cargo_bay_height'] = maximum_cargo_bay_height
    fuselage_description['geometry_description']['fuselage_dimension']['fuselage_wettet_area'] = fuselage_wettet_area

    ''' debug print '''
    runtime_output.debug('Debug: The estimate_masses function was successfully executed!')

    return fuselage_description
