"""Module providing functions for data postprocessing."""
# Import standard modules.

# Import own modules.
from datapostprocessingmodule import method_data_postprocessing
from datapostprocessingmodule import write_key_data_to_aircraft_exchange_file
from datapostprocessingmodule import prepare_element_tree_for_module_key_parameter


def data_postprocessing(paths_and_names, routing_dict, data_dict, runtime_output):
    """Perform data postprocessing and write data to an aircraft exchange file.

    This function is responsible for data postprocessing that involves the following steps:
        (1) Data preparation: The module manager prepares a list containing all paths to the key parameters that must
        be written to the aircraft exchange file.
        (2) Write data to the aircraft exchange file: The results of the module execution are passed on to the function
        responsible for properly writing the data to the aircraft exchange file.
        (3) Method-specific data postprocessing: User-defined method-specific postprocessing is conducted as needed.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict data_dict: Dictionary containing the result of the module execution (direct operating cost estimation)
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: None
    """

    """Data preparation."""
    # Changes to this list are the sole responsibility of the module manager!
    paths_to_key_parameters_list = [
        # total landing gear parameter paths
        './component_design/landing_gear/position/x',
        './component_design/landing_gear/position/y',
        './component_design/landing_gear/position/z',
        './component_design/landing_gear/mass_properties/mass',
        './component_design/landing_gear/mass_properties/inertia/j_xx',
        './component_design/landing_gear/mass_properties/inertia/j_yy',
        './component_design/landing_gear/mass_properties/inertia/j_zz',
        './component_design/landing_gear/mass_properties/inertia/j_xy',
        './component_design/landing_gear/mass_properties/inertia/j_xz',
        './component_design/landing_gear/mass_properties/inertia/j_yx',
        './component_design/landing_gear/mass_properties/inertia/j_yz',
        './component_design/landing_gear/mass_properties/inertia/j_zx',
        './component_design/landing_gear/mass_properties/inertia/j_zy',
        './component_design/landing_gear/mass_properties/center_of_gravity/x',
        './component_design/landing_gear/mass_properties/center_of_gravity/y',
        './component_design/landing_gear/mass_properties/center_of_gravity/z',
        './component_design/landing_gear/aircraft_classification_number',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/name',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/position/x',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/position/y',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/position/z',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/mass',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/inertia/j_xx',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/inertia/j_yy',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/inertia/j_zz',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/inertia/j_xy',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/inertia/j_xz',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/inertia/j_yx',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/inertia/j_yz',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/inertia/j_zx',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/inertia/j_zy',
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/center_of_gravity/x',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/center_of_gravity/y',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/mass_properties/center_of_gravity/z',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/assembly_components/strut_diameter',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/assembly_components/strut_length',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/assembly_components/tire_description[@ID="0"]/name',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/assembly_components/tire_description[@ID="0"]/position/x',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/assembly_components/tire_description[@ID="0"]/position/y',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/assembly_components/tire_description[@ID="0"]/position/z',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/assembly_components/tire_description[@ID="0"]/tire_diameter',  # noPep8 e501
        './component_design/landing_gear/specific/geometry/landing_gear_assembly[@ID="0"]/assembly_components/tire_description[@ID="0"]/tire_section_width'  # noPep8 e501
    ]

    module_key_parameters_dict = {
        # total landing gear parameter
        'component_design': {
            'landing_gear': {
                'attributes': {
                    'description': 'Geometric description of the aircraft undercarriage.',
                    'tool_level': '0'},
                'position': {
                    'attributes': {'description': 'Position of the total undercarriage arrangement with regard to the '
                                                  'global reference point.'},
                    'x': {
                        'attributes': {'description': 'Distance in x direction with regard to the global reference '
                                                      'point. (total undercarriage arrangement)'},
                        'value': '0',
                        'unit': 'm',
                        'lower_boundary': '-100',
                        'upper_boundary': '100'},
                    'y': {
                        'attributes': {'description': 'Distance in y direction with regard to the global reference '
                                                      'point. (total undercarriage arrangement)'},
                        'value': '0',
                        'unit': 'm',
                        'lower_boundary': '-25',
                        'upper_boundary': '25'},
                    'z': {
                        'attributes': {'description': 'Distance in z direction with regard to the global reference '
                                                      'point. (total undercarriage arrangement)'},
                        'value': '0',
                        'unit': 'm',
                        'lower_boundary': '-10',
                        'upper_boundary': '10'},
                },
                'mass_properties': {
                    'attributes': {
                        'description': 'Mass properties of the total undercarriage arrangement.'},
                    'mass': {
                        'attributes': {'description': 'Mass of the total undercarriage arrangement.'},
                        'value': '0',
                        'unit': 'kg',
                        'lower_boundary': '0',
                        'upper_boundary': 'inf'},
                    'inertia': {
                        'attributes': {
                            'description': 'Inertia of the total undercarriage arrangement with regard to the total '
                                           'center of gravity.'},
                        'j_xx': {
                            'attributes': {
                                'description': 'Inertia of the total undercarriage arrangement in x.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_yy': {
                            'attributes': {
                                'description': 'Inertia of the total undercarriage arrangement in y.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_zz': {
                            'attributes': {
                                'description': 'Inertia of the total undercarriage arrangement in z.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_xy': {
                            'attributes': {
                                'description': 'Inertia of the total undercarriage arrangement in xy.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_xz': {
                            'attributes': {
                                'description': 'Inertia of the total undercarriage arrangement in xz.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_yx': {
                            'attributes': {
                                'description': 'Inertia of the total undercarriage arrangement in yx.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_yz': {
                            'attributes': {
                                'description': 'Inertia of the total undercarriage arrangement in yz.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_zx': {
                            'attributes': {
                                'description': 'Inertia of the total undercarriage arrangement in zx.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_zy': {
                            'attributes': {
                                'description': 'Inertia of the total undercarriage arrangement in zy.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                    },
                    'center_of_gravity': {
                        'attributes': {
                            'description': 'Center of gravity of the total undercarriage arrangement.'},
                        'x': {
                            'attributes': {
                                'description': 'Center of gravity in x-direction with regard to the global reference '
                                               'point. (total undercarriage arrangement)'},
                            'value': '0',
                            'unit': 'm',
                            'lower_boundary': '-100',
                            'upper_boundary': '100'},
                        'y': {
                            'attributes': {
                                'description': 'Center of gravity in y-direction with regard to the global reference '
                                               'point. (total undercarriage arrangement)'},
                            'value': '0',
                            'unit': 'm',
                            'lower_boundary': '-50',
                            'upper_boundary': '50'},
                        'z': {
                            'attributes': {
                                'description': 'Center of gravity in z-direction with regard to the global reference '
                                               'point. (total undercarriage arrangement)'},
                            'value': '0',
                            'unit': 'm',
                            'lower_boundary': '-10',
                            'upper_boundary': '10'}
                    }
                },
                'aircraft_classification_number': {
                    'attributes': {
                        'description': 'Aircraft classification number for the total undercarriage arrangement. '
                                       '(CBR = [3, 6, 10, 15])'},
                    'value': '[0, 0, 0, 0]'
                },
                'specific': {
                    'attributes': {
                        'description': 'Landing gear specific output parameter.'},
                    'geometry': {
                        'attributes': {
                            'description': 'Geometry description of the entire landing gear.'},
                        'landing_gear_assembly': {
                            'attributes': {
                                'ID': '0',
                                'description': 'Geometrical description of one entire landing gear leg.'},
                            'name': {
                                'attributes': {
                                    'description': 'Name of the landing gear leg.'},
                                'value': 'nose_gear_assembly'},
                            'position': {
                                'attributes': {
                                    'description': 'Position of one entire landing gear leg with regard to the landing'
                                                   ' gear global position.'},
                                'x': {
                                    'attributes': {
                                        'description': 'Distance in x direction with regard to the landing gear global'
                                                       ' position. (center line of the landing gear leg)'},
                                    'value': '0',
                                    'unit': 'm',
                                    'lower_boundary': '-100',
                                    'upper_boundary': '100'},
                                'y': {
                                    'attributes': {
                                        'description': 'Distance in y direction with regard to the landing gear global'
                                                       ' position. (center line of the landing gear leg)'},
                                    'value': '0',
                                    'unit': 'm',
                                    'lower_boundary': '-50',
                                    'upper_boundary': '50'},
                                'z': {
                                    'attributes': {
                                        'description': 'Distance in z direction with regard to the landing gear global '
                                                       'position. (center line of the landing gear leg)'},
                                    'value': '0',
                                    'unit': 'm',
                                    'lower_boundary': '-10',
                                    'upper_boundary': '10'}
                            },
                            'mass_properties': {
                                'attributes': {
                                    'description': 'Mass properties of one entire landing gear leg.'},
                                'mass': {
                                    'attributes': {'description': 'Mass of one entire landing gear leg.'},
                                    'value': '0',
                                    'unit': 'kg',
                                    'lower_boundary': '0',
                                    'upper_boundary': '10000'},
                                'inertia': {
                                    'attributes': {
                                        'description': 'Inertia of one entire landing gear leg with regard to his '
                                                       'center of gravity.'
                                    },
                                    'j_xx': {
                                        'attributes': {
                                            'description': 'Inertia of one entire landing gear leg in x.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_yy': {
                                        'attributes': {
                                            'description': 'Inertia of one entire landing gear leg in y.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_zz': {
                                        'attributes': {
                                            'description': 'Inertia of one entire landing gear leg in z.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_xy': {
                                        'attributes': {
                                            'description': 'Inertia of one entire landing gear leg in xy.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_xz': {
                                        'attributes': {
                                            'description': 'Inertia of one entire landing gear leg in xz.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_yx': {
                                        'attributes': {
                                            'description': 'Inertia of one entire landing gear leg in yx.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_yz': {
                                        'attributes': {
                                            'description': 'Inertia of one entire landing gear leg in yz.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_zx': {
                                        'attributes': {
                                            'description': 'Inertia of one entire landing gear leg in zx.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_zy': {
                                        'attributes': {
                                            'description': 'Inertia of one entire landing gear leg in zy.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                },
                                'center_of_gravity': {
                                    'attributes': {
                                        'description': 'Center of gravity of one entire landing gear leg.'},
                                    'x': {
                                        'attributes': {
                                            'description': 'Center of gravity in x-direction with regard to the '
                                                           'global landing gear center of gravity position. '
                                                           '(entire landing gear leg)'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '-100',
                                        'upper_boundary': '100'},
                                    'y': {
                                        'attributes': {
                                            'description': 'Center of gravity in x-direction with regard to the '
                                                           'global landing gear enter of gravity position. '
                                                           '(entire landing gear leg)'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '-50',
                                        'upper_boundary': '50'},
                                    'z': {
                                        'attributes': {
                                            'description': 'Center of gravity in x-direction with regard to the '
                                                           'global landing gear center of gravity position. '
                                                           '(entire landing gear leg)'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '-10',
                                        'upper_boundary': '10'}
                                }
                            },
                            'assembly_components': {
                                'attributes': {
                                    'description': 'Geometrical description of the assembly components of '
                                                   'one landing gear strut.'},
                                'strut_diameter': {
                                    'attributes': {
                                        'description': 'Diameter of the landing gear strut'},
                                    'value': '0',
                                    'unit': 'm',
                                    'lower_boundary': '0',
                                    'upper_boundary': '1.0'},
                                'strut_length': {
                                    'attributes': {
                                        'description': 'Length of the landing gear strut'},
                                    'value': '0',
                                    'unit': 'm',
                                    'lower_boundary': '0',
                                    'upper_boundary': '10'},
                                'tire_description': {
                                    'attributes': {
                                        'ID': '0',
                                        'description': 'Description of one tire of the wheel group.'},
                                    'name': {
                                        'attributes': {
                                            'description': 'Name of the tire type.'},
                                        'value': 'name_of_tire_type'},
                                    'position': {
                                        'attributes': {
                                            'description': 'Position of one tire of current landing gear '
                                                           'strut with regard to the landing gear strut '
                                                           'position.'},
                                        'x': {
                                            'attributes': {
                                                'description': 'Distance in x direction with regard to the '
                                                               'landing gear strut position. '
                                                               '(center line of the landing gear strut)'},
                                            'value': '0',
                                            'unit': 'm',
                                            'lower_boundary': '-10',
                                            'upper_boundary': '10'},
                                        'y': {
                                            'attributes': {
                                                'description': 'Distance in y direction with regard to the '
                                                               'landing gear strut position. '
                                                               '(center line of the landing gear strut'},
                                            'value': '0',
                                            'unit': 'm',
                                            'lower_boundary': '-10',
                                            'upper_boundary': '10'},
                                        'z': {
                                            'attributes': {
                                                'description': 'Distance in z direction with regard to the '
                                                               'landing gear strut position. '
                                                               '(center line of the landing gear strut)'},
                                            'value': '0',
                                            'unit': 'm',
                                            'lower_boundary': '-10',
                                            'upper_boundary': '10'}
                                    },
                                    'tire_diameter': {
                                        'attributes': {
                                            'description': 'Diameter of one tire of the wheel group.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '0',
                                        'upper_boundary': '2'},
                                    'tire_section_width': {
                                        'attributes': {
                                            'description': 'Width at the shoulder of a tire of the '
                                                           'wheel group.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '0',
                                        'upper_boundary': '1'}
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    paths_and_names = prepare_element_tree_for_module_key_parameter(paths_and_names, module_key_parameters_dict)

    # Run 'user_method_data_output_preparation' from 'usermethoddatapreparation.py'.
    key_output_dict, method_specific_output_dict = routing_dict['func_user_method_data_output_preparation'](data_dict)
    # Extract tool level from routing dictionary.
    if data_dict['current_tool_level'] is None:
        tool_level = routing_dict['tool_level']
    else:
        tool_level = str(data_dict['current_tool_level'])

    """Write data to aircraft exchange file."""
    # Extract root and path to aircraft exchange file.
    root_of_aircraft_exchange_tree = paths_and_names['root_of_aircraft_exchange_tree']
    path_to_aircraft_exchange_file = paths_and_names['path_to_aircraft_exchange_file']
    # Write key data to aircraft exchange file.
    write_key_data_to_aircraft_exchange_file(root_of_aircraft_exchange_tree, path_to_aircraft_exchange_file,
                                             paths_to_key_parameters_list, key_output_dict, tool_level, runtime_output)

    """Method-specific postprocessing."""
    # Run 'method_data_postprocessing' from 'datapostprocessingmodule'.
    data_dict['module_key_parameters_dict'] = module_key_parameters_dict
    data_dict['paths_to_key_parameters_list'] = paths_to_key_parameters_list
    method_data_postprocessing(paths_and_names, routing_dict, data_dict,
                               method_specific_output_dict, runtime_output)
