# Import standard modules.
import os
import sys
import xml.etree.ElementTree


def estimate_tires(dict_ac_exchange, dict_mod_config, landing_gear_description, runtime_output):
    """ Hier beschreibung hin: Tire selection in accordance to CS 25.733 und EASA ETSO tire list
        -> hier bridgestone aircraft tires aus landing gear lib

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    """
    ''' initialize local parameter '''
    nose_gear_tire_dict = {}
    main_gear_tire_dict = {}
    nose_gear_description = {}
    main_gear_description = {}
    nose_gear_arrangement = {}
    main_gear_arrangement = {}
    temp_nose_gear_tire_list = []
    temp_main_gear_tire_list = []
    number_of_nose_gear_struts = 1
    number_of_main_gear_outer_struts = 2
    number_of_main_gear_inner_struts = 1
    main_gear_tire_design_load = 0.0
    main_gear_tire_selection_flag = True
    number_of_nose_gear_axis_per_strut = 1
    number_of_main_gear_axis_per_outer_strut = 1
    number_of_main_gear_axis_per_inner_strut = 0
    number_of_main_gear_tires_per_inner_strut = 0
    # ToDo: umrechungsfaktoren in seperates python package
    meter_per_second_to_miles_per_hour = 2.2369362920544
    gravitational_acceleration = float(9.80665)
    kg_to_pounds = 2.20462262185

    ''' Read aircraft tire list as element tree. '''
    try:
        path = landing_gear_description['working_directory'] + '/landing_gear_lib/EASA_ETSO_tire_list_2022.xml'
        xml_tree = xml.etree.ElementTree.ElementTree(file=path)
        root_of_aircraft_tire_tree = xml_tree.getroot()

        # generate nose gear tire dictionary from xml input
        nose_gear_tires = root_of_aircraft_tire_tree.findall('nose_gear_tire/tire')
        i = 0
        while i < len(nose_gear_tires):
            temp_tire_dict = {}
            part_number = nose_gear_tires[i].find('part_number').text
            for tire_element in nose_gear_tires[i]:
                if not tire_element.tag == 'part_number':
                    temp_tire_dict[tire_element.tag] = tire_element.text
                nose_gear_tire_dict[part_number] = temp_tire_dict
            i += 1

        # generate main gear tire dictionary from xml input
        main_gear_tires = root_of_aircraft_tire_tree.findall('main_gear_tire/tire')
        length_of_main_tire_list = len(main_gear_tires)
        i = 0
        while i < length_of_main_tire_list:
            temp_tire_dict = {}
            part_number = main_gear_tires[i].find('part_number').text
            for tire_element in main_gear_tires[i]:
                if not tire_element.tag == 'part_number':
                    temp_tire_dict[tire_element.tag] = tire_element.text
                main_gear_tire_dict[part_number] = temp_tire_dict
            i += 1

    except OSError:
        runtime_output.critical('Error: The aircraft tire xml file could not be opened!  \n Program aborted!')
        sys.exit('Read aircraft tire xml is failed for landing gear design!')

    ''' Estimate maximum design speed for tire selection. '''
    if dict_ac_exchange['maximum_take_off_stall_speed'] is not None:
        maximum_tire_design_speed_in_miles_per_hour = round(max(dict_ac_exchange['maximum_approach_speed'],
                                                                dict_ac_exchange['maximum_take_off_stall_speed'] * 1.3)
                                                            * meter_per_second_to_miles_per_hour, 4)
    else:
        maximum_tire_design_speed_in_miles_per_hour = (dict_ac_exchange['maximum_approach_speed']
                                                       * meter_per_second_to_miles_per_hour)

    ''' Nose gear tire selection in accordance to CS 25.733 '''
    # initialize number of nose gear struts if "use_specific" is selected
    if dict_mod_config['use_specific_number_of_nose_gear_struts']:
        number_of_nose_gear_struts = dict_mod_config['specific_number_of_nose_gear_struts']

    #  initialize number of nose gear tires per strut if "use_specific" is selected
    if dict_mod_config['use_specific_number_of_nose_gear_tires_per_strut']:
        number_of_nose_gear_tires_per_strut = dict_mod_config['specific_number_of_nose_gear_tires_per_strut']
    else:
        number_of_nose_gear_tires_per_strut = dict_mod_config['minimum_number_of_nose_gear_tires_per_strut']

    # check if the number of nose gear tires per strut is 3
    #  -> if true: -> no valid arrangement -> use 4 tires per strut instead
    if number_of_nose_gear_tires_per_strut == 3:
        number_of_nose_gear_tires_per_strut += 1

    # check if a multi strut nose landing gear is selected -> if true: -> set number of nose gear axis per strut to 1
    if number_of_nose_gear_tires_per_strut > 2 and number_of_nose_gear_struts > 1:
        number_of_nose_gear_axis_per_strut = 1
    # Else if condition: Check if number of nose gear tires per strut is greater than 2
    #  -> if true: -> multi bogy arrangement, 2 axis per strut required
    elif number_of_nose_gear_tires_per_strut > 2:
        number_of_nose_gear_axis_per_strut = 2

    # Estimate single tire static load for nose gear tire in accordance to CS 25.733 (b)(1)
    nose_gear_tire_design_load = round(((landing_gear_description['landing_gear_loads']['nose_gear_maximum_static_load']
                                       / gravitational_acceleration) * kg_to_pounds)
                                       / (number_of_nose_gear_tires_per_strut * number_of_nose_gear_struts), 4)

    # Estimate single tire dynamic landing load for nose gear tire in accordance to CS 25.733 (b)(2)
    nose_gear_tire_maximum_dynamic_landing_load = \
        round(((landing_gear_description['landing_gear_loads']['nose_gear_maximum_dynamic_landing_load']
                / gravitational_acceleration) * kg_to_pounds)
              / (number_of_nose_gear_tires_per_strut * number_of_nose_gear_struts), 4)

    # Estimate single tire dynamic take-off load for nose gear tire in accordance to CS 25.733 (b)(3)
    nose_gear_tire_maximum_dynamic_take_off_load = \
        round(((landing_gear_description['landing_gear_loads']['nose_gear_maximum_dynamic_take_off_load']
                / gravitational_acceleration) * kg_to_pounds)
              / (number_of_nose_gear_tires_per_strut * number_of_nose_gear_struts), 4)

    # Elimination of all tires from the nose gear dictionary witch speed and load rating is lower than required
    to_be_removed = []
    for tire_dict_key, tire_dict in nose_gear_tire_dict.items():
        # check if the rated tire speed key in the nose gear tire dictionary but the tire rating is lower than required
        #  -> if true: -> Remove nose gear tire from dictionary
        if 'rated_speed_in_mph' in tire_dict \
                and float(tire_dict['rated_speed_in_mph']) < maximum_tire_design_speed_in_miles_per_hour:
            if tire_dict_key not in to_be_removed:
                to_be_removed.append(tire_dict_key)
        # Else if condition: check if rated tire speed key is not in the nose gear tire
        #  -> if true: -> Remove nose gear tire from dictionary
        elif 'rated_speed_in_mph' not in tire_dict:
            if tire_dict_key not in to_be_removed:
                to_be_removed.append(tire_dict_key)

        # Check if the rate tire pressure key in the nose gear tire dictionary
        #  -> if true: -> estimate unloaded and maximum burst pressures in accordance to AMC 25.734
        if 'rated_pressure_in_psi' in tire_dict:
            # unloaded operational tire pressure in psi in accordance to AMC 25.734 (3.4)
            nose_gear_tire_dict[tire_dict_key]['unloaded_operational_pressure_in_psi'] = \
                str(round(float(tire_dict['rated_pressure_in_psi']) / 1.07, 1))
            # tire maximum burst pressure in psi in accordance to AMC 25.734 (4) Model 4 (3)
            nose_gear_tire_dict[tire_dict_key]['maximum_burst_pressure_in_psi'] = \
                str(round(float(tire_dict['rated_pressure_in_psi']) / 1.07 * 1.3, 1))
        # Else condition: The rate tire pressure key is not in the nose gear tire dictionary
        #  -> Remove nose gear tire from dictionary
        else:
            if tire_dict_key not in to_be_removed:
                to_be_removed.append(tire_dict_key)

        # check if the rated tire load key is not in the nose gear tire dictionary
        #  -> if true: -> Remove nose gear tire from dictionary
        if 'rated_load_in_lbs' not in tire_dict:
            if tire_dict_key not in to_be_removed:
                to_be_removed.append(tire_dict_key)
        # Else if condition: Check if the rated tire load is lower than required
        #  -> if true: -> Remove nose gear tire from dictionary
        elif float(tire_dict['rated_load_in_lbs']) < nose_gear_tire_design_load:
            if tire_dict_key not in to_be_removed:
                to_be_removed.append(tire_dict_key)
        # Else if condition: Check if the 1.5 times rated tire load is lower than required maximum dynamic landing load
        #  -> if true: -> Remove nose gear tire from dictionary
        elif float(tire_dict['rated_load_in_lbs']) * 1.5 < nose_gear_tire_maximum_dynamic_landing_load:
            if tire_dict_key not in to_be_removed:
                to_be_removed.append(tire_dict_key)
        # Else if condition: Check if the 1.5 times rated tire load is lower than required maximum dynamic take-off load
        #  -> if true: -> Remove nose gear tire from dictionary
        elif float(tire_dict['rated_load_in_lbs']) * 1.5 < nose_gear_tire_maximum_dynamic_take_off_load:
            if tire_dict_key not in to_be_removed:
                to_be_removed.append(tire_dict_key)

    for key in to_be_removed:
        del nose_gear_tire_dict[key]

    # Select nose gear tire with minimum delta between rated and required load.
    # Check if the remaining nose gear tire dictionary is not empty -> if true: -> try to find possible nose gear tires
    try:
        if len(nose_gear_tire_dict) > 0:
            rated_nose_tire_load = float(nose_gear_tire_dict[list(nose_gear_tire_dict.keys())[0]]['rated_load_in_lbs'])
            temp_nose_gear_tire_list.append(list(nose_gear_tire_dict.keys())[0])
            i = 1
            # Loop across the nose gear tire dictionary to compare the rated loads to find minimum delta load tire
            while i < len(nose_gear_tire_dict):
                # Check if the current tire has a lower load rating than the lowest rated load of the previous tire
                #  -> if true: -> overwrite list entry and set new rated nose tire load
                if float(nose_gear_tire_dict[list(nose_gear_tire_dict.keys())[i]]['rated_load_in_lbs']) \
                        < rated_nose_tire_load:
                    temp_nose_gear_tire_list = [list(nose_gear_tire_dict.keys())[i]]
                    rated_nose_tire_load = \
                        float(nose_gear_tire_dict[list(nose_gear_tire_dict.keys())[i]]['rated_load_in_lbs'])
                # Else if condition: Check if the current tire has the same load rating than the lowest rated load of
                #  the previous tire -> if true: -> append the tire to temporary nose gear tire list
                elif float(nose_gear_tire_dict[list(nose_gear_tire_dict.keys())[i]]['rated_load_in_lbs']) \
                        == rated_nose_tire_load:
                    temp_nose_gear_tire_list.append(list(nose_gear_tire_dict.keys())[i])

                i += 1
        # Else condition: The remaining nose gear tire dictionary is empty
        #  -> Nose gear tire selection failed. Raise an error and abort program.
        else:
            raise ValueError('The nose tire selection is failed. \n '
                             '      No suitable nose landing gear tire in EASA-ETSO tire list. Program aborted!')

    # Exception handling for nose tire selection failure.
    except ValueError as e:
        runtime_output.critical('Error:', str(e))
        sys.exit('Nose tire selection failed for landing gear design!')

    # Delete all unsuitable nose landing gear tires from the nose gear tire dictionary
    to_be_removed = []
    for tire_dict_key, tire_dict in nose_gear_tire_dict.items():
        if tire_dict_key not in temp_nose_gear_tire_list:
            to_be_removed.append(tire_dict_key)

    for key in to_be_removed:
        del nose_gear_tire_dict[key]

    # save nose gear tire dictionary in the landing gear tire dictionary
    nose_gear_arrangement['nose_gear_tire_design_load'] = nose_gear_tire_design_load
    nose_gear_arrangement['number_of_nose_gear_struts'] = number_of_nose_gear_struts
    nose_gear_arrangement['number_of_nose_gear_axis_per_strut'] = number_of_nose_gear_axis_per_strut
    nose_gear_arrangement['number_of_nose_gear_tires_per_strut'] = number_of_nose_gear_tires_per_strut
    nose_gear_arrangement = {key: round(value, 4) for key, value in nose_gear_arrangement.items()}
    nose_gear_description['nose_gear_arrangement'] = nose_gear_arrangement
    nose_gear_description['nose_gear_tires'] = nose_gear_tire_dict

    # prepare output dictionary
    landing_gear_description['nose_gear_description'] = nose_gear_description

    ''' Main gear tire selection in accordance to CS 25.733 '''
    # initialize number of main gear struts if "use_specific" is selected
    if dict_mod_config['use_specific_number_of_main_gear_struts']:
        number_of_main_gear_outer_struts = 2
        number_of_main_gear_inner_struts = \
            (dict_mod_config['specific_number_of_main_gear_struts'] - number_of_main_gear_outer_struts)
        if number_of_main_gear_inner_struts == 0:
            number_of_main_gear_inner_struts += 1
        else:
            number_of_main_gear_axis_per_inner_strut = 1
            number_of_main_gear_tires_per_inner_strut = 2

    #  initialize number of main gear tires per strut if "use_specific" is selected
    if dict_mod_config['use_specific_number_of_main_gear_tires_per_strut']:
        number_of_main_gear_tires_per_outer_strut = dict_mod_config['specific_number_of_main_gear_tires_per_strut']
    else:
        number_of_main_gear_tires_per_outer_strut = dict_mod_config['minimum_number_of_main_gear_tires_per_strut']

    # check if the number of main gear tires per strut is 3
    #  -> if true: -> no valid arrangement -> use 4 tires per strut instead
    if number_of_main_gear_tires_per_outer_strut == 3:
        number_of_main_gear_tires_per_outer_strut += 1

    while main_gear_tire_selection_flag:
        # check if only one main gear tyre is mounted to one main landing gear strut
        #   -> if true: -> estimate load of one single main gear tire without additional 7 percent safety load margin
        #   according to CS 25.733 (a)(1)
        if number_of_main_gear_tires_per_outer_strut == 1:
            main_gear_tire_design_load = \
                ((landing_gear_description['landing_gear_loads']['main_gear_maximum_total_load']
                  / gravitational_acceleration) * kg_to_pounds) / (number_of_main_gear_tires_per_outer_strut
                                                                   * number_of_main_gear_outer_struts
                                                                   + number_of_main_gear_tires_per_inner_strut
                                                                   * number_of_main_gear_inner_struts)
        # Else condition: More than one tire is mounted to one main landing gear strut
        #  -> additional 7 percent safety load margin according to CS 25.733 (c)(1) is required.
        else:
            main_gear_tire_design_load = \
                ((landing_gear_description['landing_gear_loads']['main_gear_maximum_total_load']
                  / gravitational_acceleration) * kg_to_pounds) / (number_of_main_gear_tires_per_outer_strut
                                                                   * number_of_main_gear_outer_struts
                                                                   + number_of_main_gear_tires_per_inner_strut
                                                                   * number_of_main_gear_inner_struts) * 1.07

        # Elimination of all tires from the nose gear dictionary witch speed and load rating is lower than required
        to_be_removed = []
        for tire_dict_key, tire_dict in main_gear_tire_dict.items():
            # check if rated tire speed key in the main gear tire dictionary but the tire rating is lower than required
            #  -> if true: -> Remove main gear tire from dictionary
            if 'rated_speed_in_mph' in tire_dict \
                    and float(tire_dict['rated_speed_in_mph']) < maximum_tire_design_speed_in_miles_per_hour:
                if tire_dict_key not in to_be_removed:
                    to_be_removed.append(tire_dict_key)
            # Else if condition: check if rated tire speed key is not in the main gear tire dictionary
            #  -> if true: -> Remove main gear tire from dictionary
            elif 'rated_speed_in_mph' not in tire_dict:
                if tire_dict_key not in to_be_removed:
                    to_be_removed.append(tire_dict_key)

            # Check if the rate tire pressure key in the main gear tire dictionary
            #  -> if true: -> estimate unloaded and maximum burst pressures in accordance to AMC 25.734
            if 'rated_pressure_in_psi' in tire_dict:
                # unloaded operational tire pressure in psi in accordance to AMC 25.734 (3.4)
                main_gear_tire_dict[tire_dict_key]['unloaded_operational_pressure_in_psi'] = \
                    str(round(float(tire_dict['rated_pressure_in_psi']) / 1.07, 1))
                # tire maximum burst pressure in psi in accordance to AMC 25.734 (4) Model 4 (3)
                main_gear_tire_dict[tire_dict_key]['maximum_burst_pressure_in_psi'] = \
                    str(round(float(tire_dict['rated_pressure_in_psi']) / 1.07 * 1.3, 1))
            # Else condition: The rate tire pressure key is not in the main gear tire dictionary
            #  -> Remove nose gear tire from dictionary
            else:
                if tire_dict_key not in to_be_removed:
                    to_be_removed.append(tire_dict_key)

            # check if the rated tire load key is not in the main gear tire dictionary
            #  -> if true: -> Remove main gear tire from dictionary
            if 'rated_load_in_lbs' not in tire_dict:
                if tire_dict_key not in to_be_removed:
                    to_be_removed.append(tire_dict_key)
            # Else if condition: Check if the rated tire load is lower than required
            #  -> if true: -> Remove main gear tire from dictionary
            elif float(tire_dict['rated_load_in_lbs']) < main_gear_tire_design_load:
                if tire_dict_key not in to_be_removed:
                    to_be_removed.append(tire_dict_key)

        for key in to_be_removed:
            del main_gear_tire_dict[key]

        # Select main gear tire with minimum delta between rated and required load.
        # Check if the remaining main gear tire dictionary is not empty
        #  -> if true: -> try to find possible main gear tires
        try:
            if len(main_gear_tire_dict) > 0:
                rated_main_tire_load = \
                    float(main_gear_tire_dict[list(main_gear_tire_dict.keys())[0]]['rated_load_in_lbs'])
                temp_main_gear_tire_list.append(list(main_gear_tire_dict.keys())[0])
                i = 1
                # Loop across the main gear tire dictionary to compare the rated loads to find minimum delta load tire
                while i < len(main_gear_tire_dict):
                    # Check if the current tire has a lower load rating than the lowest rated load of the previous tire
                    #  -> if true: -> overwrite list entry and set new rated main tire load
                    if float(main_gear_tire_dict[list(main_gear_tire_dict.keys())[i]]['rated_load_in_lbs']) \
                            < rated_main_tire_load:
                        temp_main_gear_tire_list = [list(main_gear_tire_dict.keys())[i]]
                        rated_main_tire_load = \
                            float(main_gear_tire_dict[list(main_gear_tire_dict.keys())[i]]['rated_load_in_lbs'])
                    # Else if condition: Check if the current tire has the same load rating than the lowest rated load
                    #  of the previous tire -> if true: -> append the tire to temporary main gear tire list
                    elif float(main_gear_tire_dict[list(main_gear_tire_dict.keys())[i]]['rated_load_in_lbs']) \
                            == rated_main_tire_load:
                        temp_main_gear_tire_list.append(list(main_gear_tire_dict.keys())[i])

                    i += 1

                # Check if the final number of main gear tires is greater than 1 -> if true:
                #  -> estimate number of required axis per main landing gear strut
                if number_of_main_gear_tires_per_outer_strut > 1:
                    number_of_main_gear_axis_per_outer_strut = int(number_of_main_gear_tires_per_outer_strut / 2)
                main_gear_tire_selection_flag = False

            # Else if condition: Check if dynamic change of number of main gear tires per strut is selected and
            #  the current number of main gear tires per strut is lower than the maximum allowed number per strut
            #   -> if true: -> reinitialize main gear tire dictionary and change the number of tires per strut
            elif dict_mod_config['dynamic_change_of_number_of_main_gear_tires_per_strut'] \
                    and number_of_main_gear_tires_per_outer_strut \
                    < dict_mod_config['maximum_number_of_main_gear_tires_per_strut']:
                i = 0
                main_gear_tire_dict = {}
                while i < length_of_main_tire_list:
                    temp_tire_dict = {}
                    part_number = main_gear_tires[i].find('part_number').text
                    for tire_element in main_gear_tires[i]:
                        if not tire_element.tag == 'part_number':
                            temp_tire_dict[tire_element.tag] = tire_element.text
                        main_gear_tire_dict[part_number] = temp_tire_dict
                    i += 1
                if number_of_main_gear_tires_per_outer_strut < 4:
                    number_of_main_gear_tires_per_outer_strut *= 2
                else:
                    number_of_main_gear_tires_per_outer_strut += 2

            # Else if condition: Check if dynamic change of number of main gear struts is selected and
            #  the current number of main gear struts is lower than maximum allowed number of main gear struts
            #   -> if true: -> reinitialize main gear tire dictionary and change the number of main gear struts
            elif dict_mod_config['dynamic_change_of_number_of_main_gear_struts'] \
                    and (number_of_main_gear_outer_struts + number_of_main_gear_inner_struts)\
                    <= dict_mod_config['maximum_number_of_main_gear_struts']:
                i = 0
                main_gear_tire_dict = {}
                while i < length_of_main_tire_list:
                    temp_tire_dict = {}
                    part_number = main_gear_tires[i].find('part_number').text
                    for tire_element in main_gear_tires[i]:
                        if not tire_element.tag == 'part_number':
                            temp_tire_dict[tire_element.tag] = tire_element.text
                        main_gear_tire_dict[part_number] = temp_tire_dict
                    i += 1

                number_of_main_gear_axis_per_inner_strut += 1
                if number_of_main_gear_tires_per_inner_strut == 0:
                    number_of_main_gear_tires_per_inner_strut = 2
                elif number_of_main_gear_tires_per_inner_strut < 4:
                    number_of_main_gear_tires_per_inner_strut *= 2
                else:
                    number_of_main_gear_tires_per_inner_strut += 2

                if number_of_main_gear_tires_per_inner_strut \
                        > dict_mod_config['maximum_number_of_main_gear_tires_per_strut']:
                    number_of_main_gear_inner_struts += 1
                    number_of_main_gear_axis_per_inner_strut = 1
                    number_of_main_gear_tires_per_inner_strut = 2

            # Else condition: The remaining main gear tire dictionary is empty
            #  -> Main gear tire selection failed. Raise an error and abort program.
            else:
                raise ValueError('The main tire selection is failed. \n '
                                 '      No suitable main landing gear tire in EASA-ETSO tire list. Program aborted!')

        # Exception handling for nose tire selection failure.
        except ValueError as e:
            runtime_output.critical('Error:', str(e))
            sys.exit('Main tire selection failed for landing gear design!')

    # Delete all unsuitable nose landing gear tires from the nose gear tire dictionary
    to_be_removed = []
    for tire_dict_key, tire_dict in main_gear_tire_dict.items():
        if tire_dict_key not in temp_main_gear_tire_list:
            to_be_removed.append(tire_dict_key)

    for key in to_be_removed:
        del main_gear_tire_dict[key]

    # save nose gear tire dictionary in the landing gear tire dictionary
    main_gear_arrangement['main_gear_tire_design_load'] = main_gear_tire_design_load
    main_gear_arrangement['number_of_main_gear_struts'] = number_of_main_gear_outer_struts
    main_gear_arrangement['number_of_main_gear_outer_struts'] = number_of_main_gear_outer_struts
    main_gear_arrangement['number_of_main_gear_axis_per_outer_strut'] = number_of_main_gear_axis_per_outer_strut
    main_gear_arrangement['number_of_main_gear_tires_per_outer_strut'] = number_of_main_gear_tires_per_outer_strut
    if number_of_main_gear_tires_per_inner_strut > 0:
        main_gear_arrangement['number_of_main_gear_inner_struts'] = number_of_main_gear_inner_struts
        main_gear_arrangement['number_of_main_gear_axis_per_inner_strut'] = number_of_main_gear_axis_per_inner_strut
        main_gear_arrangement['number_of_main_gear_tires_per_inner_strut'] = number_of_main_gear_tires_per_inner_strut
        main_gear_arrangement['number_of_main_gear_struts'] = (number_of_main_gear_outer_struts
                                                               + number_of_main_gear_inner_struts)
    main_gear_arrangement = {key: round(value, 3) for key, value in main_gear_arrangement.items()}
    main_gear_description['main_gear_arrangement'] = main_gear_arrangement
    main_gear_description['main_gear_tires'] = main_gear_tire_dict

    # prepare output dictionary
    landing_gear_description['main_gear_description'] = main_gear_description

    runtime_output.debug('estimate_tires successfully executed!')

    return landing_gear_description
