/**
 * \file mission_file_data_handler.cpp
 * \author Gerrit Pauls (gerrit.pauls@tuhh.de)
 * \brief
 * \version 0.1
 * \date 2024-08-30
 *
 * @copyright Copyright (c) 2024
 *
 */

#include <aixml/endnode.h>

#include "mission_file_data_handler.h"
#include "../condition_methods/flight_conditions.h"

/* Initialization of attributes of the mission_file_data_handler object with the constructor */
mission_file_data_handler::mission_file_data_handler(const std::string& path_to_mission_xml)
    :
    mission_data{},
    mission_xml(aixml::openDocument(path_to_mission_xml)),
    /* Initialization of mission parameters */
    Range(EndnodeReadOnly<double>("/mission/range").read(mission_xml).value()),
    alternate_distance(EndnodeReadOnly<double>("/mission/alternate_distance").read(mission_xml).value()),
    desired_cruise_speed(EndnodeReadOnly<double>("/mission/desired_cruise_speed").read(mission_xml).value()),
    /* Initialization of taxiing time parameters at origin and destination airports */
    taxi_time_origin(EndnodeReadOnly<double>("/mission/taxi_time_origin").read(mission_xml).value()),
    taxi_time_destination(EndnodeReadOnly<double>("/mission/taxi_time_destination").read(mission_xml).value()),
    /* Initialization of the switch for electrically driven taxiing */
    taxiing_procedure(EndnodeReadOnly<std::string>("/mission/taxiing_procedure").read(mission_xml).value() ==
                    "electric_taxiing" ? true : false), // TODO(Gerrit): Exchange for a mode selector!
    /* Initialization of engine warm up time */
    engine_warmup_time(EndnodeReadOnly<double>("/mission/engine_warmup_time").read(mission_xml).value()),
    /* Initialization of terminal time at destination airport for go around */
    terminal_operations_time(EndnodeReadOnly<double>("/mission/terminal_operation_time").read(mission_xml).value()),
    /* Initialization of aircraft parameters */
    air_to_air_fuel_mass(mission_xml.find("/mission/air_to_air_fuel") ? // only set air to air fuel when existing
        (EndnodeReadOnly<double>("/mission/air_to_air_fuel").read(mission_xml).value()) : 0.),
    payload_mass(EndnodeReadOnly<double>("/mission/payload").read(mission_xml).value()),
    PAX(EndnodeReadOnly<int>("/mission/number_of_pax").read(mission_xml).value()),
    cargo_mass(EndnodeReadOnly<double>("/mission/cargo_mass").read(mission_xml).value()) {
    //ctor
}

/* Destructor of the flight conditions object */
mission_file_data_handler::~mission_file_data_handler() {
}


/* Third Level Functions */
void mission_file_data_handler::read_departure_steps(const double& max_operating_speed, const node& energy_provider) {
    std::string subpath("mission/departure/departure_step");
    uint16_t steps = mission_xml.getVector(subpath).size();
    mission_data_set data;
    EndnodeReadOnly<int>energy_carrier_node("energy_carrier_ID");
    int takeoff_id = energy_carrier_node.read(energy_provider.at("ground")).value();
    int climb_id   = energy_carrier_node.read(energy_provider.at("ground")).value();

    /* Initialize struct */
    data.mode = 0; // Flight mode
    data.altitude = 0.; // Final altitude [m]
    data.CAS = 0.; // End CAS [m/s]
    data.mach = 0.; // Mach number
    data.mach_switch = false;
    data.glidepath = 0.; // Glide angle [rad]
    data.rel_segment_length = 0.; // Flight route for cruise
    data.auto_altitude_steps = 0.;
    data.ground_distance = 0.;
    data.auto_FL_optimum = false;
    data.round_to_regular_FL = false;
    data.CI = 0.; // Cost index [kg/min], scaled 0 to 999 according to Sperry/Honeywell [1]
    data.CI_via_speed_switch = false; // Calculate velocity variably according to CI
    data.transfer_time = 0.;
    data.transfer_rate = 0.;
    data.rel_delta_fuel_mass = 0.;
    data.delta_payload_mass = 0.;
    subpath += "@";
    for (uint16_t i(0); i < steps; ++i) {
        data.mission_file_index = i;
        data.mode_name = EndnodeReadOnly<std::string>(subpath + num2Str(i) + "/mode").read(mission_xml).value();
        /* Bleed and offtake schedule identical for all segments */
        data.background_shaft_power = EndnodeReadOnly<double>(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/background_power_level").read(mission_xml).value(); // W
        data.background_bleed = EndnodeReadOnly<double>(subpath + num2Str(i) + "/bleed_air_takeoff_schedule/background_bleed_level").read(mission_xml).value(); // kg/s
        if (mission_xml.find(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step")) {
            uint16_t shaft_power_scheduleSteps(mission_xml.getVector(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step").size());
            for (uint16_t j(0); j < shaft_power_scheduleSteps; ++j) {
                data.shaft_power_schedule.push_back(offtake_schedule(EndnodeReadOnly<double>(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step@" +
                                                                                            num2Str(j) + "/time").read(mission_xml).value(),
                                                                  EndnodeReadOnly<double>(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step@" +
                                                                                            num2Str(j) + "/power").read(mission_xml).value())); // kg/s
            }
        }
        if (mission_xml.find(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/bleed_air_takeoff_schedule_step")) {
            uint16_t bleed_scheduleSteps(mission_xml.getVector(subpath + num2Str(i) + "/bleed_air_takeoff_schedule/takeoff_schedule_step").size());
            for (uint16_t j(0); j < bleed_scheduleSteps; ++j) {
                data.bleed_schedule.push_back(offtake_schedule(EndnodeReadOnly<double>(subpath + num2Str(i) + "/bleed_air_takeoff_schedule/takeoff_schedule_step@" +
                                                                                        num2Str(j) + "/time").read(mission_xml).value(),
                                                             EndnodeReadOnly<double>(subpath + num2Str(i) + "/bleed_air_takeoff_schedule/takeoff_schedule_step@" +
                                                                                        num2Str(j) + "/massflow").read(mission_xml).value())); // kg/s
            }
        }
        myRuntimeInfo->info << "Departure step " << i << ": " << data.mode_name << std::endl;
        /* Common parameters for all modes */
        data.rating = EndnodeReadOnly<std::string>(subpath + num2Str(i) + "/rating").read(mission_xml).value();
        data.derate = EndnodeReadOnly<double>(subpath + num2Str(i) + "/derate").read(mission_xml).value();
        data.config = EndnodeReadOnly<std::string>(subpath + num2Str(i) + "/configuration").read(mission_xml).value();
        if (mission_xml.find(subpath + num2Str(i) + "/rate_of_climb_limit")) {
            data.ROC = EndnodeReadOnly<double>(subpath + num2Str(i) + "/rate_of_climb_limit").read(mission_xml).value();
        } else if (!(mission_xml.find(subpath + num2Str(i) + "/rate_of_climb_limit")) && mission_xml.find(subpath + num2Str(i) + "/altitude")) {
            data.ROC = NAN; // Set ROC to NaN to indicate that it shall be calculated via max thrust (without ROC the new altitude could not be reached)
        } else {
            data.ROC = 0.; // No new altitude = fly with constant altitude -> ROC = 0
        }
        /* Mode specific mission parameters */
        if (data.mode_name == "takeoff") {
            data.mode = Flight_conditions::TAKEOFF;
            data.energy_carrier_id = takeoff_id;
        } else if (data.mode_name == "accelerate") {
            data.mode = Flight_conditions::ACCELERATE;
            data.mach_switch = false;
            data.energy_carrier_id = climb_id;
            data.CAS = EndnodeReadOnly<double>(subpath + num2Str(i) + "/calibrated_airspeed").read(mission_xml).value();
            if (data.CAS > max_operating_speed) {
                throwError(__FILE__, __func__, __LINE__,
                    std::format("The calibrated airspeed in the missionXML (node: {}) exceeded the maximum operating speed by {} m/s. Abort program!",
                        subpath + num2Str(i), ceil(data.CAS - max_operating_speed)));
            }
        } else if (data.mode_name == "climb") { // Mode: climb at a constant calibrated airspeed
            data.mode = Flight_conditions::CLIMBCONSTCAS;
            data.mach_switch = false;
            data.altitude = EndnodeReadOnly<double>(subpath + num2Str(i) + "/altitude").read(mission_xml).value();
            data.auto_FL_optimum = false;
            data.round_to_regular_FL = false;
            data.energy_carrier_id = climb_id;
        } else {
            continue; //no data available
        }
        /* Save mission step data */
        this->mission_data.push_back(data);
        /* Clear shaft power and bleed schedule data */
        data.shaft_power_schedule.clear();
        data.bleed_schedule.clear();
    }
}

void mission_file_data_handler::read_flight_steps(double* average_shaft_bleed, double* average_cruise_bleed, bool* auto_climb_switch, double* step_height_cruise,
                                                const double& max_operating_mach, const double& max_operating_speed, const node& energy_provider) {
    /* Read in mission data above 10.000 ft */
    std::string subpath("/mission/cruise/cruise_step");
    uint16_t steps = mission_xml.getVector(subpath).size();
    mission_data_set data;
    subpath += "@";

    /* Initialize optional values of the struct */
    data.altitude = 0.; // Final altitude [m]
    data.CAS = 0.; // End CAS [m/s]
    data.mach = 0.; // Mach number
    data.mach_switch = false;
    data.glidepath = 0.; // Glide angle [rad]
    data.rel_segment_length = 0.; // Flight route for cruise
    data.auto_altitude_steps = 0.;
    data.ground_distance = 0.;
    data.CI = 0.; // Cost index [kg/min], scaled 0 to 999 according to Sperry/Honeywell [1]
    data.transfer_time = 0.;
    data.transfer_rate = 0.;
    data.rel_delta_fuel_mass = 0.;
    data.delta_payload_mass = 0.;

    // Same energy carrier for whole cruise
    data.energy_carrier_id = EndnodeReadOnly<int>("energy_carrier_ID").read(energy_provider.at("cruise")).value();

    /* Change struct values depending on step () */
    for (uint16_t i(0); i < steps; ++i) {
        /* Common parameters for all modes */
        data.mode_name = EndnodeReadOnly<std::string>(subpath + num2Str(i) + "/mode").read(mission_xml).value();
        data.rating = EndnodeReadOnly<std::string>(subpath + num2Str(i) + "/rating").read(mission_xml).value();
        data.derate = EndnodeReadOnly<double>(subpath + num2Str(i) + "/derate").read(mission_xml).value();
        data.config = EndnodeReadOnly<std::string>(subpath + num2Str(i) + "/configuration").read(mission_xml).value();
        data.auto_FL_optimum = EndnodeReadOnly<bool>(subpath + num2Str(i) + "/auto_select_optimum_flight_level/enabled").read(mission_xml).value();
        data.CI_via_speed_switch = EndnodeReadOnly<bool>(subpath + num2Str(i) + "/flight_management_system/enabled").read(mission_xml).value();
        if (data.CI_via_speed_switch) { //is flight management system switched on?
            data.CI = EndnodeReadOnly<bool>(subpath + num2Str(i) + "/flight_management_system/cost_index").read(mission_xml).value(); //TODO(Gerrit): CI checken!
        }
        data.round_to_regular_FL = EndnodeReadOnly<bool>(subpath + num2Str(i) + "/round_to_regular_flight_level").read(mission_xml).value();
        data.mission_file_index = i;

        myRuntimeInfo->info << "Cruise step " << i << ": " << data.mode_name << std::endl;
        /* Bleed and offtake schedule identical for all segments */
        data.background_shaft_power = EndnodeReadOnly<double>(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/background_power_level").read(mission_xml).value(); // W
        data.background_bleed = EndnodeReadOnly<double>(subpath + num2Str(i) + "/bleed_air_takeoff_schedule/background_bleed_level").read(mission_xml).value(); // kg/s
        /* Calculate the average shaft power */
        *average_shaft_bleed += data.background_shaft_power; // W for tank calculation
        /* Calculate the average cruise bleed */
        *average_cruise_bleed += data.background_bleed; // kg/s for tank calculation
        if (mission_xml.find(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step")) {
            uint16_t shaft_power_scheduleSteps(mission_xml.getVector(subpath + num2Str(i) + "/shaft_power_takeoff_schedule").size());
            for (uint16_t j(0); j < shaft_power_scheduleSteps; ++j) {
                data.shaft_power_schedule.push_back(offtake_schedule(EndnodeReadOnly<double>(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step@" +
                                                                                            num2Str(j) + "/time").read(mission_xml).value(),
                                                                  EndnodeReadOnly<double>(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step@" +
                                                                                            num2Str(j) + "/power").read(mission_xml).value())); // kg/s
            }
        }
        if (mission_xml.find(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/bleed_air_takeoff_schedule_step")) {
            uint16_t bleed_scheduleSteps(mission_xml.getVector(subpath + num2Str(i) + "/bleed_air_takeoff_schedule").size());
            for (uint16_t j(0); j < bleed_scheduleSteps; ++j) {
                data.bleed_schedule.push_back(offtake_schedule(EndnodeReadOnly<double>(subpath + num2Str(i) + "/bleed_air_takeoff_schedule/takeoff_schedule_step@" +
                                                                                        num2Str(j) + "/time").read(mission_xml).value(),
                                                            EndnodeReadOnly<double>(subpath + num2Str(i) + "/bleed_air_takeoff_schedule/takeoff_schedule_step@" +
                                                                                        num2Str(j) + "/massflow").read(mission_xml).value())); // kg/s
            }
        }
        /* Read optional data which was parsed into the mission XML (if not available, values from step before are kept) */
        if (mission_xml.find(subpath + num2Str(i) + "/rate_of_climb_limit")) {
            data.ROC = EndnodeReadOnly<double>(subpath + num2Str(i) + "/rate_of_climb_limit").read(mission_xml).value();
        } else if (!(mission_xml.find(subpath + num2Str(i) + "/rate_of_climb_limit")) && mission_xml.find(subpath + num2Str(i) + "/altitude")) {
            data.ROC = NAN; // Set ROC to NaN to indicate that it shall be calculate via max thrust (without ROC the new altitude could not be reached)
        } else {
            data.ROC = 0;
        }
        if (mission_xml.find(subpath + num2Str(i) + "/calibrated_airspeed")) {
            data.CAS = EndnodeReadOnly<double>(subpath + num2Str(i) + "/calibrated_airspeed").read(mission_xml).value();
            if (data.CAS > max_operating_speed) {
                throwError(__FILE__, __func__, __LINE__,
                    std::format("The calibrated airspeed in the missionXML (node: {}) exceeded the maximum operating speed by {} m/s. Abort program!",
                        subpath + num2Str(i), ceil(data.CAS - max_operating_speed)));
            }
        }
        if (mission_xml.find(subpath + num2Str(i) + "/altitude")) {
            data.altitude = EndnodeReadOnly<double>(subpath + num2Str(i) + "/altitude").read(mission_xml).value();
        }
        if (mission_xml.find(subpath + num2Str(i) + "/relative_segment_length")) {
            data.rel_segment_length = EndnodeReadOnly<double>(subpath + num2Str(i) + "/relative_segment_length").read(mission_xml).value();
        }
        if (mission_xml.find(subpath + num2Str(i) + "/mach_number")) {
            data.mach = EndnodeReadOnly<double>(subpath + num2Str(i) + "/mach_number").read(mission_xml).value();
            if (data.mach > max_operating_mach) {
                throwError(__FILE__, __func__, __LINE__,
                    std::format("The Mach number in the missionXML (node: {}) exceeded the maximum operating Mach number by {}. Abort program!",
                        subpath + num2Str(i), data.mach - max_operating_mach));
            }
        }

        /* Identify step mode and adapt data accordingly */
        if (data.mode_name == "change_speed_to_CAS") { // Mode: accelerate, i.e. change speed to calibrated airspeed
            data.mode = Flight_conditions::ACCELERATE;
        } else if (data.mode_name == "climb_to_cruise") { // Mode: climb to cruise altitude
            data.mode = (data.CI_via_speed_switch) ? Flight_conditions::CLIMBTOCRUISECI : Flight_conditions::CLIMBTOCRUISE;
        } else if (data.mode_name == "climb_to_ceiling") { // Mode: climb to ceiling altitude
            data.mode = Flight_conditions::CLIMBTOSERVICECEILING;
            if (data.CI_via_speed_switch) { //is flight management system switched on?
                data.CI = EndnodeReadOnly<bool>(subpath + num2Str(i) + "/flight_management_system/cost_index").read(mission_xml).value();
            }
        } else if (data.mode_name == "change_flight_level" &&
                   !(EndnodeReadOnly<bool>(subpath + num2Str(i) + "/auto_select_optimum_flight_level/enabled").read(mission_xml).value())) { // Mode: change flight level
            data.mode = (data.CI_via_speed_switch) ? Flight_conditions::CLIMBVARCAS : Flight_conditions::CLIMBCONSTCAS;
            data.mach_switch = true;
        } else if (data.mode_name == "change_flight_level_constant_ROC" && // Mode: change flight level while maintaining a constant rate of climb
                  !(EndnodeReadOnly<bool>(subpath + num2Str(i) + "/auto_select_optimum_flight_level/enabled").read(mission_xml).value())) {
            data.mode = Flight_conditions::FLCHANGE;
        } else if (data.mode_name == "cruise") {
            data.mode = (data.CI_via_speed_switch) ? Flight_conditions::CRUISECI : Flight_conditions::CRUISE;
            if (data.auto_FL_optimum)
                *auto_climb_switch = data.auto_FL_optimum;
            if (accuracyCheck(data.auto_altitude_steps, 0., ACCURACY_HIGH)) {
                data.auto_altitude_steps = convertUnit(FOOT, METER, 2000.); //climb step of 2000 ft as default
            }
            *step_height_cruise = convertUnit(METER, FOOT, data.auto_altitude_steps);
        } else if (data.mode_name == "change_speed_to_Mach") { // Mode: change speed to Mach number
            data.mode = Flight_conditions::ACCELERATE;
            data.mach_switch = true;
        } else if (data.mode_name == "descend_to_approach") { // Mode: descend to approach
            data.mode = (data.CI_via_speed_switch) ? Flight_conditions::DESCENTCI : Flight_conditions::DESCENT;
        } else if (data.mode_name == "air_to_air_fuelling") { // Mode: air to air fueling (UNUSED BY NOW)
            data.mode = Flight_conditions::AIRTOAIRFUELLING;
            data.transfer_time = mission_xml.at(subpath + num2Str(i) + "/transfer_time");
            data.transfer_rate = mission_xml.at(subpath + num2Str(i) + "/transfer_rate");
            data.rel_delta_fuel_mass = mission_xml.at(subpath + num2Str(i) + "/rel_delta_fuel_mass");
            data.delta_payload_mass = mission_xml.at(subpath + num2Str(i) + "/delta_payload_mass");
        } else if (data.mode_name == "free_flight_minimum_fuel") { // Mode: free flight with minimum fuel
            data.mode = Flight_conditions::FREEFLIGHT;
            data.rel_segment_length = checkBoundaries(mission_xml.at(subpath + num2Str(i) + "/relSegmentLength"), 0., true, 1., true);
        } else {
            data.mode = 0;
            continue; //no data available
        }

         /* Save mission step data */
        this->mission_data.push_back(data);
        /* Clear shaft power and bleed schedule data */
        data.shaft_power_schedule.clear();
        data.bleed_schedule.clear();
    }
    /* Average parameters for tank calculation */
    *average_shaft_bleed /= steps; // W for tank calculation
    *average_cruise_bleed /= steps; // kg/s for tank calculation
}

void mission_file_data_handler::read_approach_steps(const double& max_operating_speed, const node& energy_provider) {
    /* Read in landing procedure */
    std::string subpath("/mission/approach/approach_step");
    uint16_t steps = mission_xml.getVector(subpath).size();
    mission_data_set data;
    subpath += "@";
    EndnodeReadOnly<int>energy_carrier_node("energy_carrier_ID");
    int descent_id = energy_carrier_node.read(energy_provider.at("descent")).value();
    int landing_id = energy_carrier_node.read(energy_provider.at("landing")).value();

    /* Initialize struct */
    data.mode = 0; // Flight mode
    data.altitude = 0.; // Final altitude [m]
    data.ROC = 0.; // Rate of climb [m/s]
    data.CAS = 0.; // End CAS [m/s]
    data.mach = 0.; // Mach number
    data.mach_switch = false;
    data.glidepath = 0.; // Glide angle [rad]
    data.rel_segment_length = 0.; // Flight route for cruise
    data.auto_altitude_steps = 0.;
    data.ground_distance = 0.;
    data.auto_FL_optimum = false;
    data.round_to_regular_FL = false;
    data.CI = 0.; // Cost index [kg/min], scaled 0 to 999 according to Sperry/Honeywell [1]
    data.CI_via_speed_switch = false; // Calculate velocity variably according to CI
    data.transfer_time = 0.;
    data.transfer_rate = 0.;
    data.rel_delta_fuel_mass = 0.;
    data.delta_payload_mass = 0.;
    for (uint16_t i(0); i < steps; ++i) {
        data.mission_file_index = i;
        data.mode_name = EndnodeReadOnly<std::string>(subpath + num2Str(i) + "/mode").read(mission_xml).value();
        /* Bleed and offtake schedule identical for all segments */
        data.background_shaft_power = EndnodeReadOnly<double>("mission/approach/approach_step@" + num2Str(i) +
                                    "/shaft_power_takeoff_schedule/background_power_level").read(mission_xml).value(); // W
        data.background_bleed = EndnodeReadOnly<double>("mission/approach/approach_step@" + num2Str(i) +
                                    "/bleed_air_takeoff_schedule/background_bleed_level").read(mission_xml).value(); // kg/s
        if (mission_xml.find(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step")) {
            uint16_t shaft_power_scheduleSteps(mission_xml.getVector(subpath + num2Str(i) + "/shaft_power_takeoff_schedule").size());
            for (uint16_t j(0); j < shaft_power_scheduleSteps; ++j) {
                data.shaft_power_schedule.push_back(offtake_schedule(EndnodeReadOnly<double>(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step@" +
                                                                                            num2Str(j) + "/time").read(mission_xml).value(),
                                                                  EndnodeReadOnly<double>(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/takeoff_schedule_step@" +
                                                                                            num2Str(j) + "/power").read(mission_xml).value())); // kg/s
            }
        }
        if (mission_xml.find(subpath + num2Str(i) + "/shaft_power_takeoff_schedule/bleed_air_takeoff_schedule_step")) {
            uint16_t bleed_scheduleSteps(mission_xml.getVector(subpath + num2Str(i) + "/bleed_air_takeoff_schedule").size());
            for (uint16_t j(0); j < bleed_scheduleSteps; ++j) {
                data.bleed_schedule.push_back(offtake_schedule(EndnodeReadOnly<double>(subpath + num2Str(i) + "/bleed_air_takeoff_schedule/takeoff_schedule_step@" +
                                                                                        num2Str(j) + "/time").read(mission_xml).value(),
                                                             EndnodeReadOnly<double>(subpath + num2Str(i) + "/bleed_air_takeoff_schedule/takeoff_schedule_step@" +
                                                                                        num2Str(j) + "/massflow").read(mission_xml).value())); // kg/s
            }
        }
        myRuntimeInfo->info << "Approach step " << i << ": " << data.mode_name << std::endl;
        /* Common parameters for all modes */
        data.rating = EndnodeReadOnly<std::string>(subpath + num2Str(i) + "/rating").read(mission_xml).value();
        data.derate = EndnodeReadOnly<double>(subpath + num2Str(i) + "/derate").read(mission_xml).value();
        data.config = EndnodeReadOnly<std::string>(subpath + num2Str(i) + "/configuration").read(mission_xml).value();
        if (data.mode_name == "descend") { // Mode: descend
            data.mode = Flight_conditions::APPROACH;
            data.altitude = EndnodeReadOnly<double>(subpath + num2Str(i) + "/altitude").read(mission_xml).value();
            data.glidepath = EndnodeReadOnly<double>(subpath + num2Str(i) + "/glide_path").read(mission_xml).value();
            data.energy_carrier_id = descent_id;
        } else if (data.mode_name == "change_speed") { // Mode: change speed
            data.mode = Flight_conditions::DECELERATE;
            data.ground_distance = EndnodeReadOnly<double>(subpath + num2Str(i) + "/ground_distance").read(mission_xml).value();
            data.glidepath = EndnodeReadOnly<double>(subpath + num2Str(i) + "/glide_path").read(mission_xml).value();
            data.CAS = EndnodeReadOnly<double>(subpath + num2Str(i) + "/calibrated_airspeed").read(mission_xml).value();
            if (data.CAS > max_operating_speed) {
                throwError(__FILE__, __func__, __LINE__,
                    std::format("The calibrated airspeed in the missionXML (node: {}) exceeded the maximum operating speed by {} m/s. Abort program!",
                        subpath + num2Str(i), ceil(data.CAS - max_operating_speed)));
            }
            data.energy_carrier_id = descent_id;
        } else if (data.mode_name == "landing") { // Mode: landing
            data.mode = Flight_conditions::LANDING;
            data.altitude = EndnodeReadOnly<double>(subpath + num2Str(i) + "/altitude").read(mission_xml).value();
            data.glidepath = EndnodeReadOnly<double>(subpath + num2Str(i) + "/glide_path").read(mission_xml).value();
            data.energy_carrier_id = landing_id;
        } else if (data.mode_name == "level_glide_slope_interception") { // Mode: level glide slope interception
            data.mode = Flight_conditions::GSINTERCEPTION;
            data.energy_carrier_id = descent_id;
        } else {
            continue; //no data available
        }
        /* Save mission step data */
        this->mission_data.push_back(data);
        /* Clear shaft power and bleed schedule data */
        data.shaft_power_schedule.clear();
        data.bleed_schedule.clear();
    }
}


/* Initialization of attributes of the offtake_schedule object with the constructor */
mission_file_data_handler::offtake_schedule::offtake_schedule(double a_time, double a_Value)
    :
    time(a_time),
    value(a_Value) {
}

/* Destructor of the offtake_schedule object */
mission_file_data_handler::offtake_schedule::~offtake_schedule() {
}
