/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "ceilingPerformance.h"
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include "../../aircraft.h"

using std::endl;

namespace low {

ceilingPerformance::ceilingPerformance(std::shared_ptr<aircraft> myAcftPt)
    :
    theAcftPt(myAcftPt) {
    // ctor
}
ceilingPerformance::~ceilingPerformance() {
    // dtor
}

void ceilingPerformance::doCeilingCalculation() {
    // mass range from MTOM -> OME
    for (int i = 0; i <= 10; i++) {
        double grossMass(theAcftPt->OME + static_cast<double>(i) / 10. * (theAcftPt->MTOM - theAcftPt->OME));
        double lastAltService(0.0);
        double lastAltMax(0.0);
        double lastAltMaxOEI(0.0);
        if (i == 0) {
            lastAltService = 15000.;
            lastAltMax = 15000.;
            lastAltMaxOEI = 15000.;
        } else {
            lastAltService = theCeiling.back().serviceCeiling;
            lastAltMax = theCeiling.back().maxCeiling;
            lastAltMaxOEI = theCeiling.back().maxCeilingOEI;
        }
        theCeiling.push_back(dataCeiling());
        theCeiling.back().Mass = grossMass;
        theCeiling.back().Mach = theAcftPt->MInitialCruise;
        theCeiling.back().serviceCeiling = iterateForCeiling(grossMass, 500. * 0.3048 / 60., theCeiling.back().Mach, lastAltService, false);
        theCeiling.back().maxCeiling = iterateForCeiling(grossMass, 0.0, theCeiling.back().Mach, lastAltMax, false);
        theCeiling.back().maxCeilingOEI = iterateForCeiling(0.925 * grossMass, 500.0 * 0.3048 / 60., theCeiling.back().Mach, lastAltMaxOEI, true);
    }
    theAcftPt->myChecks.initCruiseAlt   = checkForInitialCruiseAlt();
    theAcftPt->myChecks.initCruiseM     = checkForInitialCruiseMach();
    theAcftPt->myChecks.maxOptAlt       = checkForMaxAlt();
    theAcftPt->myChecks.OEICruiseAlt    = checkForOEICruiseAlt();
}

double ceilingPerformance::iterateForCeiling(double Mass, double ROC, double Mach, double startAlt, bool OEI) {
    double maxROC(-10.0);
    while (ROC > maxROC && startAlt >= 0.0) {
        try {
            if (OEI == true) {
                Mach = sqrt((2.*Mass * G_FORCE)
                            / (theAcftPt->atm.getDensity(startAlt) * theAcftPt->S_Wing * theAcftPt->aero.CLoptimumCruise))
                            / theAcftPt->atm.getSpeedOfSound(startAlt);
                // cout << "Mach(Ceiling): " << Mach << "Startalt = "<<startAlt<<endl;
                if (Mach > theAcftPt->MMO) {
                    Mach = theAcftPt->MMO;
                }
                maxROC = theAcftPt->ROC(Mass, startAlt, Mach, 0.0,
                                        "Clean", "MaxCont", theAcftPt->myMission.bleedOffCr,  theAcftPt->myMission.shaftOffCr, true,
                                        theAcftPt->aero);
            } else {
                maxROC = theAcftPt->ROC(Mass, startAlt, Mach, 0.0,
                                        "Clean", "MaxCont", theAcftPt->myMission.bleedOffCr,  theAcftPt->myMission.shaftOffCr, false,
                                        theAcftPt->aero);
            }
            startAlt = startAlt - 50.;
        } catch (const char * text) {
            myRuntimeInfo->info << "CL_max reached. Stop Ceiling calculation" << endl;
            break;
        }
    }
    return startAlt;
}

bool ceilingPerformance::checkForInitialCruiseAlt() {
    bool Check(false);
    // Check whether service ceiling is sufficient
    if (theCeiling.back().serviceCeiling < theAcftPt->hInitialCruise * 0.3048) {
        myRuntimeInfo->out << "Performance too low, Initial Cruise Altitude cannot be achieved with MTOM!" << endl;
        myRuntimeInfo->out << "Initial Cruise Altitude: " << theAcftPt->hInitialCruise << " ft" << endl;
        myRuntimeInfo->out << "Service Ceiling at MTOM: " << theCeiling.back().serviceCeiling / 0.3048 << " ft" << endl;
        Check = false;
    } else {
        Check = true;
    }
    return Check;
}

bool ceilingPerformance::checkForInitialCruiseMach() {
    bool Check(false);
    // Check if the flyable initial Cruise Mach number is sufficient
    theAcftPt->myEngines.calculate_N1_with_penalties(theAcftPt->myEnergyProviders.energyCarrierID_cruise, theAcftPt->hInitialCruise * 0.3048, theAcftPt->MInitialCruise,
                                                        theAcftPt->atm, 1., "MaxCont", theAcftPt->myMission.bleedOffCr, theAcftPt->myMission.shaftOffCr);
    double tempDrag(theAcftPt->aero.getCruiseDrag(theAcftPt->MInitialCruise, theAcftPt->hInitialCruise * 0.3048, 0.0, theCeiling.back().Mass, "Clean", theAcftPt->atm));
    if (tempDrag / 1000. > theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise)) {
        myRuntimeInfo->out << "Performance too low, Initital Cruise Mach number cannot be achieved with MTOM!" << endl;
        myRuntimeInfo->out << "Missing engine power: " << tempDrag / 1000. - theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise)
            / 1000. << " kN" << endl;
        Check = false;
    } else {
        Check = true;
    }
    return Check;
}

bool ceilingPerformance::checkForMaxAlt() {
    bool Check(false);
    // Check whether the maximum flight altitude can be reached.
    if (theCeiling.at(0).serviceCeiling < theAcftPt->hMaxOperating * 0.3048) {
        myRuntimeInfo->err << "Performance too low, Max. Operating Altitude cannot be achieved!" << endl;
        myRuntimeInfo->err << "Max. Operating Altitude: " << theAcftPt->hMaxOperating << " ft" << endl;
        myRuntimeInfo->err << "Service Ceiling at MTOM: " << theCeiling.at(0).serviceCeiling / 0.3048 << " ft" << endl;
        Check = false;
    } else {
        Check = true;
    }
    return Check;
}

bool ceilingPerformance::checkForOEICruiseAlt() {
    bool Check(false);
    if (theCeiling.back().maxCeilingOEI < theAcftPt->myReqs.OEICruiseAltitude * 0.3048) {
        myRuntimeInfo->out << "Performance too low, OEI max. Altitude cannot be achieved with MTOM!" << endl;
        myRuntimeInfo->out << "Req. OEI max. Altitude: " << theAcftPt->myReqs.OEICruiseAltitude << " ft" << endl;
        myRuntimeInfo->out << "Service Ceiling at MTOM: " << theCeiling.back().maxCeilingOEI / 0.3048 << " ft" << endl;
        Check = false;
    } else {
        Check = true;
    }
    return Check;
}
}  // namespace low
