/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "takeOffPerformance.h"

using std::endl;
using std::vector;

namespace low {

takeOffPerformance::takeOffPerformance(std::shared_ptr<aircraft> theAircraft)
    :
    theAcftPt(theAircraft),
    tempTODR(0.0),
    tempTime(0.0),
    BLFL(0.0),
    timeBLFL(0.0),
    runwaySlopeAngle(0.0),
    runwayFriction(0.0),
    brakingFriction(0.0),
    windSpeed(0.0) {
    // ctor
    scdSegmentGradient = 0.0;
    finalSegmentGradient = 0.0;

}
takeOffPerformance::~takeOffPerformance() {
    // dtor
}

void takeOffPerformance::doTakeOffPerformance(double theRunwaySlope, double theGroundFriction, double theGroundFrictionBraking, double theHeadWind) {
    runwaySlopeAngle = theRunwaySlope;
    runwayFriction = theGroundFriction;
    brakingFriction = theGroundFrictionBraking;
    windSpeed = theHeadWind;
    myRuntimeInfo->out << "      ... Calculation of TOFL_AEO" << endl;
    /* Calculation of TOFL performance as a function of mass and height */
    for (double elevation(0.0); elevation <= 12000.; elevation += 2000.) {
        myTOFLField.push_back(dataElevationTOFL());
        myTOFLField.back().elevation = convertUnit(FOOT, METER, elevation);
        for (int i(0); i <= 10; i++) {
            myTOFLField.back().myTOFL.push_back(dataTOFL());
            myTOFLField.back().myTOFL.back().grossMass = theAcftPt->OME + static_cast<double>(i) / 10. * (theAcftPt->MTOM - theAcftPt->OME);
            calcTODR_AEO(myTOFLField.back().myTOFL.back().grossMass, myTOFLField.back().elevation);
            myTOFLField.back().myTOFL.back().TODR_AEO = tempTODR;
            myTOFLField.back().myTOFL.back().Time = tempTime;
        }
    }
    /* Checking whether the starting distance requirement can be maintained */
    theAcftPt->myChecks.TOFL                        = checkForTOFLReq();
    theAcftPt->myChecks.climbGradientSecondSegment  = checkForSecondSegment();
    theAcftPt->myChecks.climbGradientFinalSegment   = checkForFinalSegment();
}

void takeOffPerformance::calcTODR_AEO(double acMass, double Altitude) {
    tempTODR = 0.0;
    tempTime = 0.0;
    calcGroundRun(theAcftPt->MTOM, 0.0, 0., theAcftPt->vLOF(acMass, Altitude) - windSpeed, &tempTODR, &tempTime, false); // Ground run until V_LOF
    calcDistanceToScreenHeight(acMass, Altitude, &tempTODR, &tempTime, false);  // Climb until screen height: 35 ft
    tempTODR = 1.15 * tempTODR;
}

void takeOffPerformance::calcBLFL(double theRunwaySlope, double theGroundFriction, double theGroundFrictionBraking, double theHeadWind) {
    runwaySlopeAngle = theRunwaySlope;
    runwayFriction = theGroundFriction;
    brakingFriction = theGroundFrictionBraking;
    windSpeed = theHeadWind;
    // BLFL for MTOM and at altitude 0m
    myBLFL.grossMass = theAcftPt->MTOM;
    myBLFL.TODR_AEO = myTOFLField.at(0).myTOFL.back().TODR_AEO;
    /* Iteration over the speeds */
    int stepSize(40);
    for (int i = 0; i <= stepSize; i++) {
        myBLFL.vEF.push_back(theAcftPt->v1gS_TO(myBLFL.grossMass, 0.0) + static_cast<double>(i) /
                             static_cast<double>(stepSize) * (theAcftPt->v2(myBLFL.grossMass, 0.0) - theAcftPt->v1gS_TO(myBLFL.grossMass, 0.0)));
        myBLFL.TODR_OEI.push_back(0.);
        myBLFL.ASDR_OEI.push_back(0.);
        double tempTimeBLFL(0.0);
        calcGroundRun(myBLFL.grossMass, 0.0, 0., myBLFL.vEF.back() - windSpeed, &myBLFL.TODR_OEI.back(), &tempTimeBLFL, false);
        myBLFL.ASDR_OEI.back() = myBLFL.TODR_OEI.back();
        // TODR_OEI
        calcGroundRun(myBLFL.grossMass, 0.0, myBLFL.vEF.back() - windSpeed, theAcftPt->vLOF(myBLFL.grossMass, 0.0) - windSpeed, &myBLFL.TODR_OEI.back(), &tempTimeBLFL, true);
        calcDistanceToScreenHeight(myBLFL.grossMass, 0.0, &myBLFL.TODR_OEI.back(), &tempTimeBLFL, true);
        // ASDR_OEI
        double theAcceleration = calcAccelerationGroundRoll(myBLFL.grossMass, 0.0, myBLFL.vEF.back() - windSpeed, true);
        double theV = myBLFL.vEF.back() + 3.0 * theAcceleration;
        myBLFL.ASDR_OEI.back() += (0.5 * (myBLFL.vEF.back() + theV) - windSpeed) * 3.0;
        tempTimeBLFL += 3.0;
        calcGroundRun(myBLFL.grossMass, 0.0, theV - windSpeed, 0.0, &myBLFL.ASDR_OEI.back(), &tempTimeBLFL, true);
    }
    /* Determining the intersection point -> v1 and BLFL */
    unsigned int criticalPoint(0);
    if (myBLFL.TODR_OEI.at(criticalPoint) < myBLFL.ASDR_OEI.at(criticalPoint)) {
        myRuntimeInfo->out << "Required takeoff distance OEI with vLOF = vS1g (" << myBLFL.TODR_OEI.at(criticalPoint) << " m) is less than required accelerate-stop distance ("
                           << myBLFL.ASDR_OEI.at(criticalPoint) << " m) --> No BLFL could be found!" << endl;
    } else {
        while (myBLFL.TODR_OEI.at(criticalPoint) > myBLFL.ASDR_OEI.at(criticalPoint)) {
            criticalPoint++;
            if (criticalPoint >= myBLFL.TODR_OEI.size()) {
                criticalPoint--;
                myRuntimeInfo->out << "Required takeoff distance OEI with vLOF = v2 (" << myBLFL.TODR_OEI.at(criticalPoint)
                                   << " m) is larger than required accelerate-stop distance (" << myBLFL.ASDR_OEI.at(criticalPoint) << " m) --> No BLFL could be found!" << endl;
                break;
            }
        }
        double m0((myBLFL.TODR_OEI.at(criticalPoint) - myBLFL.TODR_OEI.at(criticalPoint - 1)) / (myBLFL.vEF.at(criticalPoint) - myBLFL.vEF.at(criticalPoint - 1)));
        double m1((myBLFL.ASDR_OEI.at(criticalPoint) - myBLFL.ASDR_OEI.at(criticalPoint - 1)) / (myBLFL.vEF.at(criticalPoint) - myBLFL.vEF.at(criticalPoint - 1)));
        double a0(myBLFL.TODR_OEI.at(criticalPoint) - m0 * myBLFL.vEF.at(criticalPoint));
        double a1(myBLFL.ASDR_OEI.at(criticalPoint) - m1 * myBLFL.vEF.at(criticalPoint));
        myBLFL.v1 = (a1 - a0) / (m0 - m1);
        myBLFL.BLFL = m0 * myBLFL.v1 + a0;
    }
}

void takeOffPerformance::calcGroundRun(double mass, double alt, double V0, double V1, double *distance, double *time, bool OEI) {
    int stepSize(50.);
    double deltaV = (V1 - V0) / stepSize;
    for (int i = 0; i <= stepSize; i++) {
        double vGround = V0 + static_cast<double>(i) * deltaV;
        double vAir(0.0);
        if (vGround >= windSpeed) {
            vAir = windSpeed + vGround;
        }
        double acceleration(0.0);
        if (V1 >= V0) {
            acceleration = calcAccelerationGroundRoll(mass, alt, vAir, OEI);
        } else {
            acceleration = calcDeAccelerationGroundRoll(mass, alt, vAir, OEI);
        }
        /* trapezoidal rule */
        if (i == 0 || i == stepSize) {
            *distance += 0.5 * vGround / acceleration * deltaV;
        } else {
            *distance += vGround / acceleration * deltaV;
            *time += (vGround / acceleration * deltaV) / vGround;
        }
    }
}

double takeOffPerformance::calcAccelerationGroundRoll(double mass, double alt, double Velocity, bool OEI) {
    double Mach = Velocity / theAcftPt->atm.getSpeedOfSound(alt);
    theAcftPt->myEngines.calculate_N1_with_penalties(theAcftPt->myEnergyProviders.energyCarrierID_takeoff, alt, Mach, theAcftPt->atm, 1.0, "TO", theAcftPt->myMission.bleedOffTO,
        theAcftPt->myMission.shaftOffTO);
    double tempDrag(0.0);
    double tempLift(0.0);
    double tempThrust(theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_takeoff));
    tempDrag = theAcftPt->aero.getCLDrag(Mach, alt, theAcftPt->aero.CLgroundRoll, "TakeOff", theAcftPt->atm);
    tempLift = theAcftPt->aero.CLgroundRoll * theAcftPt->S_Wing
               * 0.5 * theAcftPt->atm.getDensity(alt) * pow(Velocity, 2);
    if (OEI == true) {
        tempThrust = tempThrust * static_cast<double>(theAcftPt->numberEngines - 1) / static_cast<double>(theAcftPt->numberEngines);
        tempDrag = tempDrag + theAcftPt->calcWindMillingDrag(Mach, alt);
    }
    return 1.0 / mass * (convertUnit(KILO, NEWTON, NEWTON, tempThrust)
                           - mass * G_FORCE * (runwayFriction + sin(runwaySlopeAngle))
                           - (tempDrag - runwayFriction * tempLift));
}

double takeOffPerformance::calcDeAccelerationGroundRoll(double mass, double alt, double Velocity, bool OEI) {
    double Mach = Velocity / theAcftPt->atm.getSpeedOfSound(alt);
    double tempThrust(0.0);
    double tempLift(0.0);
    double tempDrag(0.0);
    tempDrag = theAcftPt->aero.getCLDrag(Mach, alt, theAcftPt->aero.CLgroundRoll, "TakeOff", theAcftPt->atm);
    return 1.0 / mass * (convertUnit(KILO, NEWTON, NEWTON, tempThrust)
                           - mass * G_FORCE * (brakingFriction + sin(runwaySlopeAngle))
                           - (tempDrag - brakingFriction * tempLift));
}

void takeOffPerformance::calcDistanceToScreenHeight(double mass, double alt, double *distance, double *time, bool OEI) {
    for (int i = 0; i <= 35; i++) {
        double tempAlt = alt + convertUnit(FOOT, METER, static_cast<double>(i)); //stepsize: 1 ft
        double tempTAS = theAcftPt->vLOF(mass, alt) + static_cast<double>(i) / 35. * (theAcftPt->v2(mass, alt + convertUnit(FOOT, METER, 35.)) - theAcftPt->vLOF(mass, alt));
        double tempMach = tempTAS / theAcftPt->atm.getSpeedOfSound(tempAlt);
        double fACC(0.0);
        double tempROC(0.0);
        if (OEI == true) {
            fACC = (tempTAS / G_FORCE) * ((theAcftPt->v2(mass, alt + convertUnit(FOOT, METER, 35.)) - theAcftPt->vLOF(mass, alt)) / convertUnit(FOOT, METER, 35.));
            tempROC = theAcftPt->ROC(mass, tempAlt, tempMach, fACC,
                                     "TakeOff", "TO", theAcftPt->myMission.bleedOffTO,  theAcftPt->myMission.shaftOffTO, true, theAcftPt->aero);
        } else {
            fACC = (tempTAS / G_FORCE) * ((theAcftPt->v2(mass, alt + convertUnit(FOOT, METER, 35.)) + convertUnit(KNOTS, METERPERSECOND, 10.) -
                                           theAcftPt->vLOF(mass, alt)) / convertUnit(FOOT, METER, 35.));
            tempROC = theAcftPt->ROC(mass, tempAlt, tempMach, fACC,
                                     "TakeOff", "TO", theAcftPt->myMission.bleedOffTO,  theAcftPt->myMission.shaftOffTO, false, theAcftPt->aero);
        }
        if (tempROC <= 0.) {
            myRuntimeInfo->err << "No climbing ability to reach Screen Height!" << endl;
            myRuntimeInfo->err << "ROC: " << tempROC << endl;
            myRuntimeInfo->err << "Abort!" << endl;
            exit(1);
        }
        *distance += convertUnit(FOOT, METER, 1.) / tempROC * tempTAS;
        *time += convertUnit(FOOT, METER, 1.) / tempROC;
    }
}

bool takeOffPerformance::checkForTOFLReq() {
    bool Check(false);
    if (myTOFLField.at(0).myTOFL.back().TODR_AEO > theAcftPt->myReqs.TOFL) {
        myRuntimeInfo->out << "The requirement regarding TOFL cannot be maintained.!" << endl;
        myRuntimeInfo->out << "TOFL_req        = " << theAcftPt->myReqs.TOFL << " m" << endl;
        myRuntimeInfo->out << "TOFL_calculated = " << myTOFLField.at(0).myTOFL.back().TODR_AEO << " m" << endl;
        Check = false;
    } else {
        Check = true;
    }
    return Check;
}

bool takeOffPerformance::checkForSecondSegment() {
    bool Check(false);
    vector<double> tempAlt;
    tempAlt.push_back(convertUnit(FOOT, METER, 35.));
    tempAlt.push_back(convertUnit(FOOT, METER, 400.));
    double tempClimbGradient(99999.0);
    for (unsigned int i(0); i <= tempAlt.size() - 1; i++) {
        double tempTAS = theAcftPt->v2(theAcftPt->MTOM, tempAlt.at(i));
        double tempMach = tempTAS / theAcftPt->atm.getSpeedOfSound(tempAlt.at(i));
        double tempROC = theAcftPt->ROC(theAcftPt->MTOM, tempAlt.at(i), tempMach, 0.0,
                                        "Climb", "TO", theAcftPt->myMission.bleedOffTO,  theAcftPt->myMission.shaftOffTO, true, theAcftPt->aero);
        if (tempClimbGradient > tempROC / tempTAS * 100.) {
            tempClimbGradient = tempROC / tempTAS * 100.;
        }
    }
    if (tempClimbGradient < theAcftPt->myReqs.climbGradientSecondTOSegment) {
        myRuntimeInfo->out << "The requirement regarding the climb gradient in the 2nd segment cannot be met.!" << endl;
        myRuntimeInfo->out << "Req. : " << theAcftPt->myReqs.climbGradientSecondTOSegment << " %" << endl;
        myRuntimeInfo->out << "Calc : " << tempClimbGradient << " %" << endl;
        Check = false;
    } else {
        Check = true;
    }
    scdSegmentGradient = tempClimbGradient;
    return Check;
}

bool takeOffPerformance::checkForFinalSegment() {
    bool Check(false);
    vector<double> tempAlt;
    tempAlt.push_back(convertUnit(FOOT, METER, 400.));
    tempAlt.push_back(convertUnit(FOOT, METER, 1500.));
    double tempClimbGradient(99999.0);
    for (unsigned int i(0); i <= tempAlt.size() - 1; i++) {
        double tempTAS(0.);
        if (i == 0) {
            tempTAS = theAcftPt->v2(theAcftPt->MTOM, tempAlt.at(i));
        } else {
            tempTAS = theAcftPt->vFTO(theAcftPt->MTOM, tempAlt.at(i));
        }
        // double fACC = ( 0.5*(theAcftPt->v2(theAcftPt->MTOM, tempAlt.at(i))+theAcftPt->vFTO(theAcftPt->MTOM, tempAlt.at(i)))/9.81)
        //               * fabs(theAcftPt->v2(theAcftPt->MTOM, tempAlt.at(i))-theAcftPt->vFTO(theAcftPt->MTOM, tempAlt.at(i)))/(1100.*0.3048);
        double tempMach = tempTAS / theAcftPt->atm.getSpeedOfSound(tempAlt.at(i));
        double tempROC = theAcftPt->ROC(theAcftPt->MTOM, tempAlt.at(i), tempMach, 0.0,
                                        "Climb", "TO", theAcftPt->myMission.bleedOffTO,  theAcftPt->myMission.shaftOffTO, true, theAcftPt->aero);
        if (tempClimbGradient > tempROC / tempTAS * 100.) {
            tempClimbGradient = 100. * tempROC / tempTAS;
        }
    }
    if (tempClimbGradient < theAcftPt->myReqs.climbGradientFinalTOSegment) {
        myRuntimeInfo->out << "The requirement regarding the climb gradient in the final segment T-O cannot be maintained.!" << endl;
        myRuntimeInfo->out << "Req. : " << theAcftPt->myReqs.climbGradientFinalTOSegment << " %" << endl;
        myRuntimeInfo->out << "Calc : " << tempClimbGradient << " %" << endl;
        Check = false;
    } else {
        Check = true;
    }
    finalSegmentGradient = tempClimbGradient;
    return Check;
}
}  // namespace low
