/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef performance_assessment_SRC_TAW_DEFAULTMETHODS_TAKEOFFPERFORMANCE_LOW_FIDELITY_TAKEOFFPERFORMANCE_H_
#define performance_assessment_SRC_TAW_DEFAULTMETHODS_TAKEOFFPERFORMANCE_LOW_FIDELITY_TAKEOFFPERFORMANCE_H_

#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/unitConversion.h>

#include <vector>
#include <memory>

#include "../abstractTakeOffPerformance.h"
#include "../../aircraft.h"

class aircraft;

using std::vector;

namespace low {

/**\class takeOffPerformance global class contains informations of the aircraft and the runway at take-off phase
*
*/
class takeOffPerformance : public abstractTakeOffPerformance {
 public:
    // /**\class dataTOFL class contains informations about the take-off field length
    // *
    // */
    // class dataTOFL {
    //  public:
    //     double TODR_AEO; /**< Take-off distance required with all engines operating (m) */
    //     double Time; /**< Time (s) */
    //     double grossMass; /**< Aircraft gross mass (kg) */

    //     dataTOFL() {
    //         grossMass = 0.0;
    //         TODR_AEO = 0.0;     // Take-off Distance Required, All Engines Operating
    //         Time = 0.0;
    //     }
    // };

    // /**\class dataElevationTOFL class contains airfield elevation data
    // *
    // */
    // class dataElevationTOFL {
    //  public:
    //     double elevation; /**< Elevation of the airfield (m) */
    //     vector<dataTOFL> myTOFL; /**< List of take-off field length data (Aircraft mass, Take-off distance, Time) */

    //     dataElevationTOFL() {
    //         elevation  = 0.0;
    //     }
    // };
    // vector<dataElevationTOFL> myTOFLField; /**< List of airfield elevation data */


    /**\class dataBLFL class contains data for the balanced field length
    *
    */
    class dataBLFL {
     public:
        double grossMass; /**< Aircraft mass (kg) */
        double v1; /**< Decision speed (m/s) */
        double BLFL; /**< Balanced field length (m) */
        double TODR_AEO; /**< Take-off distance required with all engines operating (m) */
        vector<double> vEF; /**< Engine failure speed (m/s) */
        vector<double> TODR_OEI; /**< Take-off distance required with one engine operative (m) */
        vector<double> ASDR_OEI; /**< Accelerate stop distance required with one engine inoperative (m) */

        dataBLFL() {
            grossMass = 0.0;
            v1 = 0.;
            BLFL = 0.0;
            TODR_AEO = 0.0;
        }
    };
    // dataBLFL myBLFL;

    // double scdSegmentGradient; /**< Climb or descend gradient of the second segment (%) */
    // double finalSegmentGradient; /**< Descend gradient of the final segment (%) */

    /**\brief Function to compute the take-off performance
    *\param theRunwaySlope Slope of the runway (rad)
    *\param theGroundFriction Friction coefficient of the ground (-)
    *\param theGroundFrictionBraking Friction coefficient at full braking (-)
    *\param theHeadWind Front wind speed (m/s)
    */
    void doTakeOffPerformance(double theRunwaySlope, double theGroundFriction, double theGroundFrictionBraking, double theHeadWind) override;

    /**\brief Function to compute the balanced field length
    *\param theRunwaySlope Slope of the runway (rad)
    *\param theGroundFriction Friction coefficient of the ground (-)
    *\param theGroundFrictionBraking Friction coefficient at full braking (-)
    *\param theHeadWind Front wind speed (m/s)
    */
    void calcBLFL(double theRunwaySlope, double theGroundFriction, double theGroundFrictionBraking, double theHeadWind) override;

    /**\brief Constructor takeOffPerformance
    *\param theAircraft Address to the aircraft object
    */
    explicit takeOffPerformance(std::shared_ptr<aircraft> theAircraft);

    /**\brief Destructor takeOffPerformance
    *
    */
    virtual ~takeOffPerformance();

 private:
    std::shared_ptr<aircraft> theAcftPt; /**< Pointer to the aircraft object */

    double tempTODR; /**< Take-off distance required (m) */
    double tempTime; /**< Time required for take-off (s) */

    double BLFL; /**< Balanced field length (m) */
    double timeBLFL; /**< Time required for the balanced field length (s) */

    double runwaySlopeAngle; /**< Runway slope (rad) */
    double runwayFriction; /**< Friction of the runway (-) */
    double brakingFriction; /**< Friction of the aircraft brakes (-) */
    double windSpeed; /**< Wind speed (m/s) */

    /**\brief Function to compute the take-off distance required with all engines operative
    *\param acMass Aircraft mass (kg)
    *\param Altitude Aircraft altitude (m)
    */
    void calcTODR_AEO(double acMass, double Altitude);

    /**\brief Function to compute the balanced field length
    *
    */
    void calcBLFL();

    /**\brief Function to calculate the aircraft performance on ground
    *\param mass Aircraft mass (kg)
    *\param alt Aircraft altitude (m)
    *\param V0 Aircraft speed at the beginning (m/s)
    *\param V1 Aircraft speed at the end of the acceleration (m/s)
    *\param distance Take-off distance (m)
    *\param time Take-off time (s)
    *\param OEI Switch if one engine is inoperative
    */
    void calcGroundRun(double mass, double alt, double V0, double V1, double *distance, double *time, bool OEI);

    /**\brief Function to calculate the aircraft acceleration on ground
    *\param mass Aircraft mass (kg)
    *\param alt Aircraft altitude (m)
    *\param Velocity Aircraft speed (m/s)
    *\param OEI Swicht if one engine is inoperative
    *\return Acceleration of the aircraft on ground (m/s^2)
    */
    double calcAccelerationGroundRoll(double mass, double alt, double Velocity, bool OEI);

    /**\brief Function to calculate the aircraft deceleration on ground
    *\param mass Aircraft mass (kg)
    *\param alt Aircraft altitude (m)
    *\param Velocity Aircraft speed (m/s)
    *\param OEI Switch if one engine is inoperative
    *\return Deceleration of the aircraft on ground (m/s^2)
    */
    double calcDeAccelerationGroundRoll(double mass, double alt, double Velocity, bool OEI);

    /**\brief Function to calculate the distance needed to reach the screen height of 35 ft
    *\param mass Aircraft mass (kg)
    *\param alt Airfield (aircraft) altitude (m)
    *\param distance Take-off distance (m)
    *\param time Take-off time (s)
    *\param OEI Switch if one engine is inoperative
    */
    void calcDistanceToScreenHeight(double mass, double alt, double *distance, double *time, bool OEI);

    /**\brief Function to verify if the take-off field length meets the requirements
    *\return True or false
    */
    bool checkForTOFLReq();

    /**\brief Function to verify if the climb rate for the second climb segment can be achieved
    *\return True or false
    */
    bool checkForSecondSegment();

    /**\brief Function to verify if the climb rate for the final climb segment can be achieved
    *\return True or false
    */
    bool checkForFinalSegment();
};
}  // namespace low
#endif  // performance_assessment_SRC_TAW_DEFAULTMETHODS_TAKEOFFPERFORMANCE_LOW_FIDELITY_TAKEOFFPERFORMANCE_H_
