/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

/* The following statements import necessary header files */
#include "conventionalATA28.h"                                       // Import of header file from src/aircraftSystems folder for conventionalATA28
#include <runtimeInfo/runtimeInfo.h>                                 // Import of header file from external runtime library for using runtimeInfo

/* Constructor for conventionalATA28 */
conventionalATA28::conventionalATA28(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO) {
    ATA = "ATA-28";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

/* Destructor for conventionalATA28 */
conventionalATA28::~conventionalATA28() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA28::readParameters(const node& config) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(config);
    //no system specific config parameters for ATA28
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA28::updateXML() {
    updateMass("fuel_system", Mass, ATA);
}

/* Calculation of the system power */
void conventionalATA28::calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
        const std::vector<powerProfile>& allPreviousSystemsHydraulic,
        const std::vector<powerProfile>& allPreviousSystemsElectric,
        const powerProfile& allPreviousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(allPreviousSystemsBleed);
    UNUSED(allPreviousSystemsHydraulic);
    UNUSED(allPreviousSystemsElectric);
    UNUSED(allPreviousSystemsHeat);
    /* Calculate Systems Power */
    initializePowerProfile();
    if (Operating == true) {
        getPower();
    }
    calculateMaxMissionPower();
}

/* Calculation of the skin mass */
void conventionalATA28::calculateSinkMass(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    /* Calculate Systems Mass */
    Mass = getMass();
    refPoint = getRefPoint();
}

/* Determination of the power */
void conventionalATA28::getPower() {
    /* Factor calibrated according to [Bue14] p.66 (MTOM 77000t at 11.9kW -> 0.154547)*/
    for (size_t n(0); n < Electric.size(); n++) {
        // NOTE (aigner#1#): Temporary equal distribution on all electrical circuits
        Electric.at(n).designPower = 0.154547 * data_->data.Mass.MTOM / Electric.size();
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        bleedAir.missionPower.at(i).baseLoad = 0.0;
        heatLoad.missionPower.at(i).baseLoad = 0.0;
        for (size_t n(0); n < Electric.size(); n++) {
            Electric.at(n).missionPower.at(i).baseLoad = Electric.at(n).designPower;
        }
        for (size_t n(0); n < Hydraulic.size(); n++) {
            Hydraulic.at(n).missionPower.at(i).baseLoad = 0.0;
        }
    }
}

/* Determination of the mass */
double conventionalATA28::getMass() {
    std::vector<double> m_ATA28;
    //Method from [Ste10], p. 27 eq. 4-15, with 0.7 as technology factor.
    m_ATA28.push_back(0.7 * 0.9184 * pow((data_->data.Mass.maxUsableFuel / data_->data.fuelDensity), 0.6));
    /* [Ray18]: m_ATA28 = 2.405 * V_useableFuel^0.606 * 0.5 * 1 * n_tanks^0.5 */
    double V_useableFuel = data_->data.Mass.maxUsableFuel / data_->data.fuelDensity * 0.264172037; //kg -> l -> gal
    m_ATA28.push_back(0.5 * 2.405 * pow(V_useableFuel, 0.606) * pow(3., 0.5));
    m_ATA28.back() = m_ATA28.back() * 0.4535924; //lb -> kg
    double massATA28(0.);
    for (size_t i = 0; i < m_ATA28.size(); i++) {
        massATA28 += m_ATA28.at(i);
    }
    massATA28 = massATA28 / static_cast<double>(m_ATA28.size());
    return  massATA28;
}

/* Determination of the reference point */
Vec3 conventionalATA28::getRefPoint() {
    /* Assumption: center of gravity is in the CG of the wing*/
    return Vec3(data_->data.CofG.Wing);
}
