/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "conventionalATA70.h"

#include <unitConversion/constants.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>

conventionalATA70::conventionalATA70(const std::string& systemName, bool systemIsOperating,
                                     std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO) {
    /* Define ATA name */
    ATA = "ATA-70";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

conventionalATA70::~conventionalATA70() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA70::readParameters(const node& config) {
    percentageBleedAir          = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/percentage_bleed_air/value");
    data_->configuration.checkForValidity(percentageBleedAir, 0.0, 1.0, "Propulsion.percentageBleedAir");
    percentageElectricPower     = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/percentage_electrical_power/value");
    data_->configuration.checkForValidity(percentageElectricPower, 0.0, 1.0, "Propulsion.percentageElectricPower");
    percentageHydraulicPower    = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/percentage_hydraulic_power/value");
    data_->configuration.checkForValidity(percentageHydraulicPower, 0.0, 1.0, "Propulsion.percentageHydraulicPower");
    efficiencyFactorBleedAir    = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/efficiency_factor_bleed_air/value");
    data_->configuration.checkForValidity(efficiencyFactorBleedAir, 0.0, 1.0, "Propulsion.efficiencyFactorBleedAir");
//    efficiencyFactorHydraulic   = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/efficiencyFactorHydraulic");
//    data_->configuration.checkForValidity(efficiencyFactorHydraulic, 0.0, 1.0, "Propulsion.efficiencyFactorHydraulic" );
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA70::updateXML() {
    // maximum power demand hydraulic + electric -> engine shaft offtakes
    std::string subPath = "aircraft_exchange_file/component_design/systems/specific";
    if (!(rtIO_.acxml.find(subPath))) {
        rtIO_.acxml[subPath];
    }
    Endnode<double> maxPower = data_->data.create_endnode(-1.0 * (getElectricPower() + getHydraulicPower()), "W", subPath + "/maximium_power_demand",
                                                      "maximum power (hydraulic + electric) demand");
    maxPower.set_boundaries(0, INFINITY);
    maxPower.update(rtIO_.acxml);
    // maximum bleed air offtake from engine
    Endnode<double> maxBleed = data_->data.create_endnode(-1.0 * bleedAir.designPower, "kg/s", subPath + "/maximium_bleed_air_demand",
                                                      "maximum bleed air demand");
    maxBleed.set_boundaries(0, INFINITY);
    maxBleed.update(rtIO_.acxml);
}

void conventionalATA70::calculateSourcePower(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    initializePowerProfile();
    getPower(sourceBleed, sourceElectric, sourceHydraulic);
    calculateMaxMissionPower();
}

void conventionalATA70::calculateSourceMass(const powerProfile& sourceBleed,
                                            const std::vector<powerProfile>& sourceElectric,
                                            const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceHydraulic);
    UNUSED(sourceElectric);
}

void conventionalATA70::getPower(powerProfile theBleed, std::vector<powerProfile> theElectric, std::vector<powerProfile> theHydraulic) {
    // Negating the total power load for energy balance
    powerNegation(&theBleed);
    for (size_t n(0); n < theElectric.size(); n++) {
        powerNegation(&theElectric.at(n));
    }
    /*Loop over hydraulic circuits*/
    for (size_t n(0); n < theHydraulic.size(); n++) {
        powerNegation(&theHydraulic.at(n));
    }
    bleedAir.designPower  = (percentageBleedAir / efficiencyFactorBleedAir) * theBleed.designPower;
    for (size_t n(0); n < theElectric.size(); n++) {
        for (size_t m(0); m < data_->configuration.Electric.circuits.at(n).generators.size(); m++) {
            double genOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(m).operationFactor;
            double genEfficiency = data_->configuration.Electric.circuits.at(n).generators.at(m).efficiency;
            Electric.at(n).designPower  += genOperationFactor * (percentageElectricPower / genEfficiency) * theElectric.at(n).designPower;
        }
    }
    /*Loop over hydraulic circuits*/
    for (size_t n(0); n < theHydraulic.size(); n++) {
        for (size_t m(0); m < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); m++) {
            double pumpOperationFactor = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).operationFactor;
            double pumpEfficiency = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).efficiency;
            Hydraulic.at(n).designPower  += pumpOperationFactor * (percentageHydraulicPower / pumpEfficiency) * theHydraulic.at(n).designPower;
        }
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        // Bleed AiR
        bleedAir.missionPower.at(i).baseLoad = (percentageBleedAir / efficiencyFactorBleedAir) * theBleed.missionPower.at(i).baseLoad;
        for (size_t j = 0; j < theBleed.missionPower.at(i).peakLoad.size(); j++) {
            bleedAir.missionPower.at(i).peakLoad.push_back((percentageBleedAir / efficiencyFactorBleedAir) * theBleed.missionPower.at(i).peakLoad.at(j));
            bleedAir.missionPower.at(i).timePeak.push_back(theBleed.missionPower.at(i).timePeak.at(j));
        }
    }
    for (size_t n(0); n < theElectric.size(); n++) {
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            // Electric
            for (size_t m(0); m < data_->configuration.Electric.circuits.at(n).generators.size(); m++) {
                double genOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(m).operationFactor;
                double genEfficiency = data_->configuration.Electric.circuits.at(n).generators.at(m).efficiency;
                Electric.at(n).missionPower.at(i).baseLoad += genOperationFactor * (percentageElectricPower / genEfficiency) * theElectric.at(n).missionPower.at(i).baseLoad;
            }
            for (size_t j = 0; j < theElectric.at(n).missionPower.at(i).peakLoad.size(); j++) {
                bool existingStep{false};
                for (size_t k = 0; k < Electric.at(n).missionPower.at(i).peakLoad.size(); k++) {
                    if (accuracyCheck(Electric.at(n).missionPower.at(i).timePeak.at(k), theElectric.at(n).missionPower.at(i).timePeak.at(k), ACCURACY_HIGH)) {
                        existingStep = true;
                        for (size_t m(0); m < data_->configuration.Electric.circuits.at(n).generators.size(); m++) {
                            double genOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(m).operationFactor;
                            double genEfficiency = data_->configuration.Electric.circuits.at(n).generators.at(m).efficiency;
                            Electric.at(n).missionPower.at(i).peakLoad.at(k) += genOperationFactor * (percentageElectricPower / genEfficiency)
                                    * theElectric.at(n).missionPower.at(i).peakLoad.at(j);
                        }
                        break;
                    } else {
                        existingStep = false;
                    }
                }
                if (existingStep == false) {
                    Electric.at(n).missionPower.at(i).peakLoad.push_back(0.);
                    Electric.at(n).missionPower.at(i).timePeak.push_back(theElectric.at(n).missionPower.at(i).timePeak.at(j));
                    for (size_t m(0); m < data_->configuration.Electric.circuits.at(n).generators.size(); m++) {
                        double genOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(m).operationFactor;
                        double genEfficiency = data_->configuration.Electric.circuits.at(n).generators.at(m).efficiency;
                        Electric.at(n).missionPower.at(i).peakLoad.back() += genOperationFactor * (percentageElectricPower / genEfficiency)
                                * theElectric.at(n).missionPower.at(i).peakLoad.at(j);
                    }
                }
            }
        }
    }
    for (size_t n(0); n < theHydraulic.size(); n++) {
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            // Hydraulic
            for (size_t m(0); m < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); m++) {
                double pumpOperationFactor = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).operationFactor;
                double pumpEfficiency = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).efficiency;
                Hydraulic.at(n).missionPower.at(i).baseLoad += pumpOperationFactor * (percentageHydraulicPower / pumpEfficiency) * theHydraulic.at(n).missionPower.at(i).baseLoad;
            }
            for (size_t j = 0; j < theHydraulic.at(n).missionPower.at(i).peakLoad.size(); j++) {
                bool existingStep{false};
                for (size_t k = 0; k < Hydraulic.at(n).missionPower.at(i).peakLoad.size(); k++) {
                    if (accuracyCheck(Hydraulic.at(n).missionPower.at(i).timePeak.at(k), theHydraulic.at(n).missionPower.at(i).timePeak.at(k), ACCURACY_HIGH)) {
                        existingStep = true;
                        for (size_t m(0); m < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); m++) {
                            double pumpOperationFactor = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).operationFactor;
                            double pumpEfficiency = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).efficiency;
                            Hydraulic.at(n).missionPower.at(i).peakLoad.at(k) += pumpOperationFactor * (percentageHydraulicPower / pumpEfficiency)
                                    * theHydraulic.at(n).missionPower.at(i).peakLoad.at(j);
                        }
                        break;
                    } else {
                        existingStep = false;
                    }
                }
                if (existingStep == false) {
                    Hydraulic.at(n).missionPower.at(i).peakLoad.push_back(0.);
                    Hydraulic.at(n).missionPower.at(i).timePeak.push_back(theHydraulic.at(n).missionPower.at(i).timePeak.at(j));
                    for (size_t m(0); m < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); m++) {
                        double pumpOperationFactor = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).operationFactor;
                        double pumpEfficiency = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).efficiency;
                        Hydraulic.at(n).missionPower.at(i).peakLoad.back() += pumpOperationFactor * (percentageHydraulicPower / pumpEfficiency)
                                * theHydraulic.at(n).missionPower.at(i).peakLoad.at(j);
                    }
                }
            }
        }
    }
    //Clean up peaks
    for (size_t n(0); n < Hydraulic.size(); n++) {
        cleanPowerPeaks(&Hydraulic.at(n));
    }
    for (size_t n(0); n < Electric.size(); n++) {
        cleanPowerPeaks(&Electric.at(n));
    }
    cleanPowerPeaks(&bleedAir);
}

double conventionalATA70::getMass() {
    return  static_cast<double>(0.0);
}

Vec3 conventionalATA70::getRefPoint(Vec3 forwardCenterGravity, Vec3 aftCenterGravity) {
    /* Assumption: engine system is exactly in the aircraft center of gravity  **/
    return Vec3(0.5 * (forwardCenterGravity + aftCenterGravity));
}

void conventionalATA70::powerNegation(powerProfile *Power) {
    Power->maxMissionPower = (-1.0) * Power->maxMissionPower;
    Power->designPower = (-1.0) * Power->designPower;
    for (size_t i = 0; i < Power->missionPower.size(); i++) {
        Power->missionPower.at(i).baseLoad = (-1.0) * Power->missionPower.at(i).baseLoad;
        for (size_t j = 0; j < Power->missionPower.at(i).peakLoad.size(); j++) {
            Power->missionPower.at(i).peakLoad.at(j) = (-1.0) * Power->missionPower.at(i).peakLoad.at(j);
        }
    }
}

