/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_HIGHLIFTSYSTEM_HIGHLIFTSYSTEM_H_
#define SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_HIGHLIFTSYSTEM_HIGHLIFTSYSTEM_H_

#include <string>
#include <vector>
#include <memory>
#include "ActuationAndSupport.h"
#include "PowerControlUnit.h"
#include "theGearboxes.h"
#include "theTorqueLimiters.h"
#include "TransmissionShaft.h"
#include "wingTipBrakes.h"
#include "../conventionalATA27.h"

class liftingSurface;
class systemsIOData;

/** \brief contains all properties of the entire high lift system
 */
class highLiftSystem {
 public:
    /** \brief contains all properties of one single high lift device
     */
    class highLiftDevice {
     public:
        double Sref;/**<[m^2] reference area of the high lift device  */
        double l_m_Device;/**<[m] device length  */
        //double actuationMass;
        double actuationSupportMass;/**<[kg] mass of the actuation and support system  */
        std::vector<ActuationAndSupport> ActuationSupport;/**< ActuationAndSupport object  */

        highLiftDevice() {
            Sref = 0.0;
            l_m_Device = 0.0;
            //deviceMass = 0.0;
            actuationSupportMass = 0.0;
        }
    };

    /** \brief contains all properties of one indidual high lift system (trailing edge or leading edge)
     */
    class HLSystem {
     public:
        std::string Edge;/**< edge the high lift system is mounted on (trailing or leading edge) */
        std::string Type;/**< type of slat or flap */
        double systemSref;/**< [m^2] entire panel reference area */
        double systemMass;/**< [kg] mass of high lift system*/
        double requiredPowerActuators;/** [W] required power of all actuators of this HL system*/
        double requiredMomentActuatorsMax;/**< [Nm] maximum required moment from the actuator*/
        double requiredPowerPCU; /**< [W] required PCU power for one extension*/
        double requriedMomentPCUMax;/**< [Nm] maximum required PCU moment */
        double maximumPowerPCU;/**< [W] maximum PCU power*/

        std::vector<highLiftDevice> Device;/**< vector containing the high lift devices */
        std::vector<PowerControlUnit> PCU;/**< vector containing the power control units */
        std::vector<TransmissionShaft> TorqueShaft;/**< vector containing the torque shafts */
        std::vector<theGearboxes> Gearboxes;/**< vector containing the gear boxes */
        std::vector<theTorqueLimiters> TorqueLimiters;/**< vector containing the torque limiters */
        wingTipBrakes WTB;/**< wing tip brake objet */

        HLSystem()
        :
        Type("I am a ..."),
        systemSref(0.0),
        systemMass(0.0),
        requiredPowerActuators(0.0),
        requiredMomentActuatorsMax(0.0),
        requiredPowerPCU(0.0),
        requriedMomentPCUMax(0.0),
        maximumPowerPCU(0.0) {}
    };

    std::vector<HLSystem> Systems;/**< vector containing the high lift systems */
    bool theControlDeviceWarning;

    std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> LEDevices; /** vector for all leading edge devices */
    std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> TEDevices; /** vector for all trailing edge devices */

    /**
     * \brief function sorts control devices into leading and trailing edge
     * 
     * \return void
    */
    void sortControlDevices();

    /** \brief function gets all leading edge high lift devices
     *
     * \return void
     *
     */
    void getLEDevices();

    /** \brief function gets all trailing edge high lift devices
     *
     * \return void
     *
     */
    void getTEDevices();

    /** \brief function gets the total system mass of the high lift system
     *
     * \param SystemID int: identifier of the system
     * \return double Mass: [kg] mass of the HL-system
     *
     */
    double getSystemMass(int SystemID);

    /** \brief calculates the required power of all actuators
     *
     * \param SystemID int
     * \return double Power: [W]
     *
     */
    double getRequiredPowerActuators(int SystemID);

    /** \brief calculates the required moment of all actuators
     *
     * \param SystemID int
     * \return double Moment: [NM]
     *
     */
    double getRequiredMomentActuatorsMax(int SystemID);
    /** \brief calculates the required power of the power control unit
     *
     * \param SystemID int
     * \return double Power: [W]
     *
     */
    double getRequiredPowerPCU(int SystemID);

    /** \brief calculates the maximum required moment of the power control unit
     *
     * \param SystemID int
     * \return double Moment: [Nm]
     *
     */
    double getRequriedMomentPCUMax(int SystemID);

    /** \brief calculates the maximum power of the power control unit
     *
     * \param SystemID int
     * \return double Power: [w]
     *
     */
    double getMaximumPowerPCU(int SystemID);


    /** \brief Test output function
     *
     * \param SystemID int
     * \return void
     *
     */
    void testOutput(int SystemID);


    /** \brief constructor of the high lift system
     *
     * \param Wing liftingSurface
     * \param mySettings const systemsDesignSettings&
     * \param airDensity double: [kg/m^3]
     * \param airSpeed double: [m/s]
     * \param theMTOM double: [kg]
     *
     */
    highLiftSystem(const systemsIOData::Wing& Wing, bool controlDeviceWarning, const std::vector<conventionalATA27::flightControlSystem::HighLiftSystem>& highLift,
                   double airDensity, double airSpeed, double theMTOM);
    virtual ~highLiftSystem();

 private:
    const systemsIOData::Wing *theWing;
};

#endif // SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_HIGHLIFTSYSTEM_HIGHLIFTSYSTEM_H_
