/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_WINGANTIICING_WINGANTIICING_H_
#define SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_WINGANTIICING_WINGANTIICING_H_

#include <vector>
#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/processing/measure.h>

class atmosphere;
class liftingSurface;

/** \brief contains all properties of the wing anti-icing
 */
class wingAntiIcing {
 public:
    double totalExternalHeatFlux;/**<[kW] total external heat flux*/

    class wingSegment {
     public:
        double segSpan;/**<[m] segment span*/
        double yPos;/**<[m] y-position along the wing*/
        double segThickness;/**<[ft] segment thickness*/
        double segFrontSpar;/**<[ft] front spar position along the chord in this segment*/
        double segIcingSurface;/**<[ft^2] icing surface of the segment*/

        wingSegment() {
            segSpan         = 0.0;
            yPos            = 0.0;
            segThickness    = 0.0;
            segFrontSpar    = 0.0;
            segIcingSurface = 0.0;
        }
    };
    std::vector<wingSegment> theWingSegments; /**< vector containing the wing segment */

    /**
     * \brief sets the power consumption based on icing conditions
     * 
     * \param Mach mach number [-]
     * \param Alt altitude [m]
     * \param theISA ISA atmosphere
     * \param liquidWaterContent Water quantity per wing segment, from [SAE04] page 12 Graph [lb/h]
    */
    void setPowerConsumption(double Mach, double Alt, atmosphere *theISA,
                             double liquidWaterContent);
    /**
     * \brief calculates bleed air consumption
     * 
     * \param bleedAirTemperature bleed air temperature [?]
     * \param efficiencyBleed bleed air efficiency [-]
     * 
     * \return double bleed air consumption [kg/s]
    */
    double getBleedAirConsumption(double bleedAirTemp, double efficiencyBleed);
    /**
     * \brief calculates electric energy consumption
     * 
     * \param efficiencyElectroThermic electro thermic efficiency [-]
     * 
     * \return electric consumption [?]
    */
    double getElectricConsumption(double efficiencyElectroThermic);

    /** \brief Constructor
     *
     * \param theWing systemsIOData::Wing
     * \param skinThickness double: [m]
     * \param relSpanAntiIcing double
     * \param theHeatConductivity double
     *
     */
    wingAntiIcing(const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing, double skinThickness,
                  double wingRelSpanAntiIcingStart, const double& wingRelSpanAntiIcingEnd, double theHeatConductivity,
                  double frontSparPos);
    virtual ~wingAntiIcing();

 private:
    /* Wing properties */
    double theSkinThickness;/**<Wing skin thickness*/
    double heatConductivity;/**<heat conductivity of the wing skin*/
    int nrWingSegments;/**<Number of wing segments*/

    /* Heat flows */
    std::vector<double> externalHeatFlux; /**<[Btu/h] Heat flow from surface to air/water */
    std::vector<double> TinnerSkin;/**< Temperature at the inner skin of the nose */

    /**
     * \brief determines surfing area that needs anti icing
     * 
     * \param theYPOS spanwise position of the segment 
     * \param theSegWidth width of the wing segment
     * \param frontSpar front spar position along chord
     * \param Wing the Wing
     * 
     * \return surface area
    */
    double getIcingSurface(double theYPOS, double theSegWidth, double frontSpar,
                           const geom2::MultisectionSurface<geom2::AirfoilSection>& Wing);

    //double getLiquidWaterContent( double Temperature, double myDropDia );
    /**
     * \brief calculates the water catch based on the profile thickness, air speed and water content
     * 
     * \param theThickness thickness of the wing profile [m]
     * \param theTAS true air speed [m/s]
     * \param theWaterContent water content of the air [?]
     * 
     * \return water catch [?]
    */
    double getWaterCatch(double theThickness, double theTAS, double theWaterContent);
    /**
     * \brief calculates the external heat transfer coefficient
     * 
     * \param theDensity air density [kg/m^3]
     * \param theTAS true air speed [m/S?]
     * \param Lx ?
     * \param airT air temperature [K]
     * \param skinT skin temperature [K]
    */
    double getExternalHeatTransferCoeff(double theDensity, double theTAS, double Lx, double airT, double skinT);
    /**
     * \brief calculates vapour pressure at skin
     * 
     * \param theWaterCatch water catch
     * \param theHeatTransCoefSkin heat transfer coefficient of the skin
     * \param theSurface surface area [m^2]
     * \param thePressure pressure [P]
     * 
     * \return vapour pressure
    */
    double getVapourPressureSkin(double theWaterCatch, double theHeatTransCoefSkin, double theSurface, double thePressure);
    /**
     * \brief calculates the skin temperature based on the vapour pressure
     * 
     * \param theVapourPressureSkin vapour pressure at skin
     * 
     * \return skin temperature [K]
    */
    double getskinTFromVapourPressure(double theVapourPressureSkin);
    /**
     * \brief calculates enthalpy of vapourization
     * 
     * \param theSkinTemp skin temperature [K]
     * 
     * \return enthalpy
    */
    double getEnthalpyVapourization(double theSkinTemp);

    /**
     * \brief calculates heat flux from the skin to the air
     * 
     * \param theWaterCatch water catch
     * \param theHeatTransferCoefficient heat transfer coefficient
     * \param theIcingSurface icing surface [m^2]
     * \param theTskin skin temperature [K]
     * \param theTair air temperature [K]
     * \param theMach Mach number [-]
     * \param dynamicViscosity dynamic viscosity
     * \param theEnthalpyVapourization enthalpy of vapourization
     * 
     * \return heat flux from skin to air
    */
    double getHeatFluxSkinToAir(double theWaterCatch, double theHeatTransferCoefficient, double theIcingSurafe,
                                double theTskin, double theTair, double theMach, double dynamicViscosity, double theEnthalpyVapourization);

    /**
     * \brief calculates the heat capacity of air
     * 
     * \param theAirTemp air temperature [K]
     * 
     * \return heat capacity of the air
    */
    double getHeatCapacityOfAir(double theAirTemp);
    /**
     * \brief calculates the heat conductivity of air
     * 
     * \param theAirTemp air temperature [K]
     * 
     * \return heat conductivity of the air
    */
    double getHeatConductivityOfAir(double theAirTemp);

    /**
     * \brief calculates the mass flow of the bleed air
     * 
     * \param theInnerSkinTemp inner skin temperature [K]
     * \param theBleedAirTemp bleed air temperature [K]
     * \param effiencyFactor efficiency factor [-]
     * \param theHeatFlux heat flux skin to air
     * 
     * \return mass flow [kg/s]
    */
    double getMassFlow(double theInnerSkinTemp, double theBleedAirTemp, double effiencyFactor, double theHeatFlux);

    /**
     * \brief linear interpolation between two points
     * 
     * \param X0 x value of first point
     * \param X1 x value of second point
     * \param Y0 y value of first point
     * \param Y1 y value of second point
     * \param theX x value desired point
     * 
     * \return y value of the desired point
    */
    double Interpolation(double X0, double X1, double Y0, double Y1, double theX);

    /**
     * \brief function to determine the maximum thickness of an airfoil at the required segment (section of the wing at a given position)
     * To get the thickness at a specific span position, the maximum thickness of the airfoil is linearly interpolated between the sections
     * before and after the given y-position.
     * \param theWing the Wing geometry
     * \param yPos the y position / span position where the thickness should be determined 
    */
    double getMaxSegThickness(const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing, double yPos);
};

#endif // SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_WINGANTIICING_WINGANTIICING_H_
