/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef SYSTEMSDESIGN_SRC_POWERPROFILE_H_
#define SYSTEMSDESIGN_SRC_POWERPROFILE_H_

#include <vector>                            // Import c++ standard library for using vector

/** \brief Class contains the base load, peak load and duration of a peak load for one mission step*/
class missionLoads {
 public:
    double baseLoad;                 /**< [W] Base load in the individual mission phases **/
    std::vector<double> peakLoad;    /**< [W] Temporary peaks in the individual mission sections **/
    std::vector<double> timePeak;    /**< [sek] Duration of the respective peak **/

    missionLoads();
    virtual ~missionLoads();
};

/** \brief contains the design power, the maximum mission power and the power at each mission step*/
class powerProfile {
 public:
    double designPower;                     /** Energy consumption for the design case **/
    double maxMissionPower;                 /**< [W] maximum energy consumption over the mission **/
    std::vector<missionLoads> missionPower; /**< [W] Mission Dependent Loads **/

    /** \brief calculate maximum power of the system
     * \return double: max. power [W]
     */
    double getMaxPower() const;

    powerProfile();
    virtual ~powerProfile();
};


/** \brief accumulates the individual power profiles of a powerProfile vector to a total powerProfile,
 * \detailed Example: 3 hydraulic circuit profiles with one powerProfile each are combined to form a total hydraulic powerProfile
 * \param profileVector vector<powerProfile>
 * \return powerProfile
 *
 */
powerProfile getPowerProfileSum(std::vector<powerProfile> profileVector);


/** \brief Adds peaks with the same duration to a single peak., Deletes then superfluous peaks, where the peakLoad = 0.
 *
 * \param uncleanedPowerProfile powerProfile*
 * \return void
 *
 */
void cleanPowerPeaks(powerProfile *uncleanedPowerProfile);


#endif // SYSTEMSDESIGN_SRC_POWERPROFILE_H_
