""" Module containing functions for kerosene tank calculation. """
# Import standard libraries.
import sys
import math
import numpy as np

# Import own libraries.
from pyunitconversion import constants


def calculate_tank_span(span_index_inner_surface, span_index_outer_surface):
    """Tank span in m.

    Span index equals position in mm.

    :param int span_index_inner_surface: Index of span position of inner surface
    :param int span_index_outer_surface: Index of span position of outer surface
    :return float: Tank span in m
    """
    tank_span = (span_index_outer_surface - span_index_inner_surface)/100

    return tank_span


def calculate_obelisk_volume(method, dict_wing_geometry, inner_surface_str, outer_surface_str):
    """Calculate geometrical volume of obelisk.

    This function calculates the obelisk volume according to the selected method (Torenbeek [1] or Simpson).
    
    [1] E. Torenbeek "Synthesis of Subsonic Airplane Design" (1982), Fig. B-4 [1]

    :param str method: Calculation method name
    :param dict dict_wing_geometry: Dictionary containing wing geometry parameters
    :param str inner_surface_str: Name of inner obelisk surface
    :param str outer_surface_str: Name of outer obelisk surface
    :return float: Volume of obelisk in m^3
    """
    interpolated_values = dict_wing_geometry['interpolated_values']

    inner_span_index = abs(dict_wing_geometry['sections'][inner_surface_str]['span_index'])
    outer_span_index = abs(dict_wing_geometry['sections'][outer_surface_str]['span_index'])
    tank_span = (outer_span_index - inner_span_index)/100
    maximum_tank_height_inner = interpolated_values['tank_heights'][inner_span_index]
    maximum_tank_height_outer = interpolated_values['tank_heights'][outer_span_index]
    maximum_tank_length_inner = interpolated_values['tank_lengths'][inner_span_index]
    maximum_tank_length_outer = interpolated_values['tank_lengths'][outer_span_index]
    area_inner = maximum_tank_height_inner*maximum_tank_length_inner
    area_outer = maximum_tank_height_outer*maximum_tank_length_outer

    match method:
        case 'Torenbeek':
        # Calculate volume of obelisk according to Torenbeek.
            obelisk_volume = (tank_span/3*(area_inner + area_outer
                                           + ((maximum_tank_height_inner*maximum_tank_length_outer)
                                              + (maximum_tank_length_inner*maximum_tank_height_outer))/2))
        case 'Simpson':
            # Calculate volume of obelisk according to Simpson's rule.
            middle_area = (area_inner + area_outer)/2
            obelisk_volume = tank_span / 6.0 * (area_inner + 4.0 * middle_area + area_outer)
    
    return obelisk_volume


def calculate_kerosene_tanks(dict_ac_data, dict_tank_design, runtime_output):
    """Calculate kerosene tanks of tube-and-wing aircraft.
    
    This function adds the following sub dictionaries to the 'dict_tank_design' dictionary:
    ...

    :param dict dict_ac_data: Dictionary containing data from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    # Extract data and initialize variables.
    total_tank_energy = 0
    wing_tank_energy_without_wing_center_tank = 0
    kerosene_volumetric_energy_density = constants.KEROSENE_VOLUMETRIC_ENERGY_DENSITY

    """Check if initial sizing loop and estimate energy demand if necessary."""
    initial_sizing_loop = all(value is None for value in dict_ac_data['mission_energy_amount'].values())
    if initial_sizing_loop:
        runtime_output.print('Initial sizing loop! Energy demand is estimated (3.35 liters per PAX per 100 km).')
        energy_demand = 3.35  # in liter per passenger
        mission_fuel_amount = dict_ac_data['number_of_passengers']*dict_ac_data['range']/1000*energy_demand/100
        mission_energy_amount = (mission_fuel_amount/1000*kerosene_volumetric_energy_density)
    else:
        mission_energy_amount = dict_ac_data['mission_energy_amount']['mission_energy_amount_ID0']

    """ Calculate wing tank volume. """
    # Wing tank volume must be calculated for all possible tank configurations.
    dict_tank_design = calculate_wing_tanks(dict_ac_data, dict_tank_design, runtime_output, mission_energy_amount)
    total_tank_energy += dict_tank_design['tmp_energies']['wing_tank_with_center_tank']
    wing_tank_energy_without_wing_center_tank += dict_tank_design['tmp_energies']['wing_tank_without_center_tank']
    # Check if wing tanks can store mission energy amount.
    if wing_tank_energy_without_wing_center_tank >= mission_energy_amount:
        runtime_output.print('Energy check: Wing center tank created but unnecessary to store required energy '
                             + 'amount.')
        runtime_output.print('Energy check: Energy demand covered.')
        # Set 'energy_required' to 0.
        wing_tank_entity_list = [k for k, v in dict_tank_design.items()
                        if k.startswith('tank_') and k[5:].isdigit() and v['tank_location'] == 'wing']
        wing_center_tank_entity = [k for k in wing_tank_entity_list
                                   if dict_tank_design[k]['tank_position'].endswith('center')][0]
        dict_tank_design[wing_center_tank_entity]['energy_required_for_mission'] = False
        # Set energy demand flag.
        energy_demand_covered = True
    elif total_tank_energy >= mission_energy_amount:
        runtime_output.print('Energy check: Wing center tank necessary to store required energy amount.')
        runtime_output.print('Energy check: Energy demand covered.')
        energy_demand_covered = True
    else:
        energy_demand_covered = False

    # TEST
    # If energy amount not covered by wing tanks: Calculate other fuel tanks (depending on tank configuration).
    # Calculate additional tank volumes according to defined tank configuration.
    match dict_tank_design['tank_configuration']:
        case 'wing_with_trim_tank':
            if not energy_demand_covered:
                runtime_output.print('Energy check: Demand not covered yet. Continue with design of trim tank.')
            dict_tank_design = calculate_trim_tank(dict_ac_data, dict_tank_design, runtime_output,
                                                   energy_demand_covered)
            total_tank_energy += dict_tank_design['tmp_energies']['trim_tank']
            if energy_demand_covered:
                runtime_output.print('Trim tank is generated but unnecessary to store required energy amount.')
            else:
                if total_tank_energy >= mission_energy_amount:
                    energy_demand_covered = True
                    runtime_output.print('Energy check: Energy demand covered.')
        case 'wing_with_additional_center_tank':
            if not energy_demand_covered:
                runtime_output.print('Energy check: Demand not covered yet. Continue with design of additional '
                                     + 'center tank.')
            dict_tank_design = calculate_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output,
                                                                energy_demand_covered)
            total_tank_energy += (dict_tank_design['tmp_energies']['additional_center_tank'])
            if energy_demand_covered:
                runtime_output.print('Additional center tank is generated but unnecessary to store required energy '
                                     + 'amount.')
            else:
                if total_tank_energy >= mission_energy_amount:
                    energy_demand_covered = True
                    runtime_output.print('Energy check: Energy demand covered.')
                    
        case 'wing_with_additional_center_and_trim_tank':
            additional_center_tank_entity = [k for k, v in dict_tank_design.items()
                                             if k.startswith('tank_') and k[5:].isdigit()
                                             and v['tank_location'] == 'fuselage'
                                             and v['tank_position'] == 'center']
            trim_tank_entity = [k for k, v in dict_tank_design.items() if k.startswith('tank_') and k[5:].isdigit()
                                and v['tank_location'] == 'horizontal_stabilizer' and v['tank_position'] == 'total']
            first_act_then_trim = (int(additional_center_tank_entity[-1][-1]) < int(trim_tank_entity[-1][-1]))
            # first_act_then_trim = (dict_tank_design['tank_5']['tank_location'] == 'fuselage')
            first_trim_then_act = not first_act_then_trim
            if first_act_then_trim and not first_trim_then_act:
                if not energy_demand_covered:
                    runtime_output.print('Energy check: Demand not covered yet. Continue with design of additional '
                                        + 'center tank.')
                dict_tank_design = calculate_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output,
                                                                    energy_demand_covered)
                total_tank_energy += dict_tank_design['tmp_energies']['additional_center_tank']
                if energy_demand_covered:
                    runtime_output.print('Additional center tank is generated but unnecessary to store required energy'
                                         + ' amount.')
                else:
                    if total_tank_energy >= mission_energy_amount:
                        energy_demand_covered = True
                        runtime_output.print('Energy check: Energy demand covered.')
                    else:
                        runtime_output.print('Energy check: Demand not covered yet. Continue with design of trim '
                                             + 'tank.')
                dict_tank_design = calculate_trim_tank(dict_ac_data, dict_tank_design, runtime_output,
                                                       energy_demand_covered)
                total_tank_energy += dict_tank_design['tmp_energies']['trim_tank']
                if energy_demand_covered:
                    runtime_output.print('Trim tank is generated but unnecessary to store required energy '
                                     + 'amount.')
                else:
                    if total_tank_energy >= mission_energy_amount:
                        energy_demand_covered = True
                        runtime_output.print('Energy check: Energy demand covered.')
            if first_trim_then_act and not first_act_then_trim:
                if not energy_demand_covered:
                    runtime_output.print('Energy check: Demand not covered yet. Continue with design of trim tank.')
                dict_tank_design = calculate_trim_tank(dict_ac_data, dict_tank_design, runtime_output,
                                                       energy_demand_covered)
                total_tank_energy += dict_tank_design['tmp_energies']['trim_tank']
                if energy_demand_covered:
                    runtime_output.print('Trim tank is generated but unnecessary to store required energy '
                                     + 'amount.')
                else:
                    if total_tank_energy >= mission_energy_amount:
                        energy_demand_covered = True
                        runtime_output.print('Energy check: Energy demand covered.')
                    else:
                        runtime_output.print('Energy check: Demand not covered yet. Continue with design of '
                                             + 'additional center tank.')
                dict_tank_design = calculate_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output, 
                                                                    energy_demand_covered)
                total_tank_energy += dict_tank_design['tmp_energies']['additional_center_tank']
                if energy_demand_covered:
                    runtime_output.print('Additional center tank is generated but unnecessary to store required energy '
                                     + 'amount.')
                else:
                    if total_tank_energy >= mission_energy_amount:
                        energy_demand_covered = True
                        runtime_output.print('Energy check: Energy demand covered.')

    # Final check if tanks can store mission energy amount.
    if energy_demand_covered:
        runtime_output.print('Tank design successful.')
    else:
        energy_missing = mission_energy_amount - total_tank_energy
        volume_missing = energy_missing/kerosene_volumetric_energy_density
        runtime_output.print('Attention: Necessary amount of fuel cannot be stored in selected tanks! '
                                + str(int(round(energy_missing,0))) + 'J (' + str(round(volume_missing,2)) +
                                'L) missing.')
    
    # TEST END

    # """ Print information or calculate other fuel tanks. """
    # if energy_demand_covered:
    #     # Energy amount that can be stored in wing tanks already covers mission energy amount. All other tanks are
    #     # generated but remain empty ('volume_available' and 'energy_available' are 0.0).
    #     match dict_tank_design['tank_configuration']:
    #         case 'wing_with_trim_tank':
    #             runtime_output.print('Trim tank is generated but unnecessary and remains empty.')
    #         case 'wing_with_additional_center_tank':
    #             runtime_output.print('Additional center tank is generated but unnecessary and remains empty.')
    #         case 'wing_with_additional_center_and_trim_tank':
    #             runtime_output.print('Trim and additional center tank are generated but unnecessary and '
    #                                     + 'remain empty.')
    # else:
    #     # If energy amount not covered by wing tanks: Calculate other fuel tanks (depending on tank configuration).
    #     # Calculate additional tank volumes according to defined tank configuration.
    #     match dict_tank_design['tank_configuration']:
    #         case 'wing_with_trim_tank':
    #             dict_tank_design = calculate_trim_tank(dict_ac_data, dict_tank_design, runtime_output)
    #             total_tank_energy += dict_tank_design['tmp_energies']['trim_tank']
    #             if total_tank_energy >= mission_energy_amount:
    #                 energy_demand_covered = True
    #         case 'wing_with_additional_center_tank':
    #             dict_tank_design = calculate_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output)
    #             total_tank_energy += (
    #                 dict_tank_design['tmp_energies']['additional_center_tank'])
    #             if total_tank_energy >= mission_energy_amount:
    #                 energy_demand_covered = True
    #         case 'wing_with_additional_center_and_trim_tank':
    #             first_act_then_trim = (dict_tank_design['tank_5']['tank_location'] == 'fuselage')
    #             first_trim_then_act = (dict_tank_design['tank_5']['tank_location'] == 'horizontal_stabilizer')
    #             if first_act_then_trim and not first_trim_then_act:
    #                 dict_tank_design = calculate_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output)
    #                 total_tank_energy += dict_tank_design['tmp_energies']['additional_center_tank']
    #                 if total_tank_energy >= mission_energy_amount:
    #                     energy_demand_covered = True
    #                     runtime_output.print('Energy check: Energy demand covered.')
    #                     runtime_output.print('Trim tank is generated but unnecessary and remains empty.')
    #                 else:
    #                     runtime_output.print('Energy check: Demand not covered yet. Continue with design of trim '
    #                                          + 'tank.')
    #                     dict_tank_design = calculate_trim_tank(dict_ac_data, dict_tank_design, runtime_output)
    #                     additional_tank_energy += dict_tank_design['tmp_energies']['trim_tank']
    #                     total_tank_energy += additional_tank_energy
    #                     if total_tank_energy >= mission_energy_amount:
    #                         energy_demand_covered = True
    #             if first_trim_then_act and not first_act_then_trim:
    #                 dict_tank_design = calculate_trim_tank(dict_ac_data, dict_tank_design, runtime_output)
    #                 total_tank_energy += dict_tank_design['tmp_energies']['trim_tank']
    #                 if total_tank_energy >= mission_energy_amount:
    #                     energy_demand_covered = True
    #                     runtime_output.print('Energy check: Energy demand covered.')
    #                     runtime_output.print('Additional center tank is generated but unnecessary and remains empty.')
    #                 else:
    #                     runtime_output.print('Energy check: Demand not covered yet. Continue with design of '
    #                                          + 'additional center tank.')
    #                     dict_tank_design = calculate_additional_center_tank(dict_ac_data, dict_tank_design,
    #                                                                         runtime_output)
    #                     additional_tank_energy += dict_tank_design['tmp_energies']['additional_center_tank']
    #                     total_tank_energy += additional_tank_energy
    #                     if total_tank_energy >= mission_energy_amount:
    #                         energy_demand_covered = True

        # # Final check if tanks can store mission energy amount.
        # if energy_demand_covered:
        #     runtime_output.print('Tank design successful.')
        # else:
        #     energy_missing = mission_energy_amount - total_tank_energy
        #     volume_missing = energy_missing/kerosene_volumetric_energy_density
        #     runtime_output.print('Attention: Necessary amount of fuel cannot be stored in selected tanks! '
        #                          + str(int(round(energy_missing,0))) + 'J (' + str(round(volume_missing,2)) +
        #                          'L) missing.')

    """ Calculate total tank parameters. """
    dict_tank_design['total_tank_parameters'] = {}
    # Tank mass (integral tank -> equals zero).
    dict_tank_design['total_tank_parameters']['mass'] = 0.0
    # Tank position.
    dict_tank_design['total_tank_parameters']['position'] = {}
    dict_tank_design['total_tank_parameters']['position']['x'] = (
        dict_tank_design['geometry']['wing']['position']['x'])
    dict_tank_design['total_tank_parameters']['position']['y'] = (
        dict_tank_design['geometry']['wing']['position']['y'])
    dict_tank_design['total_tank_parameters']['position']['z'] = (
        dict_tank_design['geometry']['wing']['position']['z'])
    # Inertia (are set to zero).
    inertia_list = ['j_xx', 'j_yy', 'j_zz', 'j_xy', 'j_xz', 'j_yx', 'j_yz', 'j_zx', 'j_zy']
    dict_tank_design['total_tank_parameters']['inertia'] = {}
    for inertia in inertia_list:
        dict_tank_design['total_tank_parameters']['inertia'][inertia] = 0.0
    # Center of gravity (equals zero since integral tank has no mass).
    dict_tank_design['total_tank_parameters']['center_of_gravity'] = {}
    dict_tank_design['total_tank_parameters']['center_of_gravity']['x'] = 0.0
    dict_tank_design['total_tank_parameters']['center_of_gravity']['y'] = 0.0
    dict_tank_design['total_tank_parameters']['center_of_gravity']['z'] = 0.0

    # Additional fuselage length equals zero for kerosene driven aircraft.
    dict_tank_design['total_tank_parameters']['additional_fuselage_length'] = 0.0

    # Prints.
    runtime_output.print('Debug: The "calculate_tanks" function was successfully executed.')

    return dict_tank_design


def calculate_wing_tanks(dict_ac_data, dict_tank_design, runtime_output, mission_energy_amount):
    """Calculate energy contained in wing tanks.
    
    In the first step, the volume of the obelisks (according to Torenbeek or Simpson) is calculated for every given
    tank entity. Afterwards, the actual tank volume is calculated, considering volume loss due to internal structure
    of integral tanks and buffer for temperature-dependent expansion of fuel. All the volumes are summed up
    ('volume_obelisks_without_center_tank'), except for the center tank volume that is saved in a separate parameter
    ('volume_obelisk_center_tank'). The energy contained in the tank can than be calculated by multiplying the actual
    volume with the volumetric energy density of kerosene.
    Finally it is checked if the wing center tank is necessary. If the left and right inner and outer wing tanks are
    big enough to store the energy required, the wing center tank values are set to zero.

    :param dict dict_design: Dictionary containing data from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :param float mission_energy_amount: Mission energy amount in Joule
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    # Extract necessary data from 'dict_ac_data'.
    factor_volume_usable = dict_ac_data['factor_volume_usable']
    # Since a vent tank is considered, no temperature expansion allowance has to be considered for wing tanks.
    temperature_expansion_allowance = 1.0
    # Extract kerosene energy density from library.
    kerosene_volumetric_energy_density = constants.KEROSENE_VOLUMETRIC_ENERGY_DENSITY
    
    # Print.
    runtime_output.print('Wing tank design started...')

    wing_geometry_dict = dict_tank_design['geometry']['wing']
    wing_section_dict = dict_tank_design['geometry']['wing']['sections']
    wing_symmetry = dict_tank_design['geometry']['wing']['information']['wing_symmetry']

    # Get all tank entities that are located at the wing.
    wing_tank_entity_list = [k for k, v in dict_tank_design.items()
                            if k.startswith('tank_') and k[5:].isdigit() and v['tank_location'] == 'wing']
    wing_center_tank_entity = [k for k in wing_tank_entity_list
                               if dict_tank_design[k]['tank_position'].endswith('center')][0]
    # Calculate distance from wing root to wing tip.
    distance_wing_root_to_tip = (wing_section_dict['wing_tip']['position']['y']
                                  - wing_section_dict['wing_root']['position']['y'])
    # Calculate span offset for tank sections.
    center_tank_span_offset = -(wing_section_dict['wing_root']['position']['y']*
                                dict_ac_data['buffer_center_tank_segment'])
    inner_wing_tank_span_offset = distance_wing_root_to_tip*dict_ac_data['buffer_inner_tank_segment']
    kink_section_span_offset = 0.0
    outer_vent_tank_span_offset = -(distance_wing_root_to_tip*dict_ac_data['buffer_outer_tank_segment'])

    # Set tank sections according to number of wing sections.
    number_of_wing_sections = len(wing_section_dict)
    if number_of_wing_sections < 3:
        runtime_output.critical('Error: Not enough wing sections given. '
                                + 'Program aborted.')
        sys.exit(1)
    elif number_of_wing_sections == 3:
        sections = {'outer_center_tank_section': 'wing_root',
                    'section_0': 'wing_root',
                    'section_1': 'wing_tip',
                    'outer_vent_tank_section': 'wing_tip'}
        vent_tank_inner_surface = 'section_1'
        # Span offsets for 'center_tank', 'section_0', 'section_1', and 'outer_vent_tank_section'.
        span_offset_list = [center_tank_span_offset, inner_wing_tank_span_offset, kink_section_span_offset,
                            outer_vent_tank_span_offset]
    elif number_of_wing_sections == 4:
        sections = {'outer_center_tank_section': 'wing_root',
                    'section_0': 'wing_root',
                    'section_1': 'kink',
                    'section_2': 'wing_tip',
                    'outer_vent_tank_section': 'wing_tip'}
        vent_tank_inner_surface = 'section_2'
        # Span offsets for 'center_tank', 'section_0', 'section_1', 'section_2', and 'outer_vent_tank_section'.
        span_offset_list = [center_tank_span_offset, inner_wing_tank_span_offset, kink_section_span_offset,
                            outer_vent_tank_span_offset, outer_vent_tank_span_offset]
    else:
        runtime_output.critical('Error: Too many wing sections given. '
                                + 'Program aborted.')
        sys.exit(1)

    # Prepare tank sections.
    i = 0
    for sec, template in sections.items():
        wing_section_dict[sec] = {}
        wing_section_dict[sec]['position'] = {}
        wing_section_dict[sec]['position']['y'] = (
            wing_section_dict[template]['position']['y'] + span_offset_list[i])
        wing_section_dict[sec]['span_index'] = abs(int(round(
            wing_section_dict[sec]['position']['y']*100, 0)))
        wing_section_dict[sec]['position']['x'] = (
            wing_geometry_dict['interpolated_values']['x_coordinate_leading_edge']
            [wing_section_dict[sec]['span_index']])
        wing_section_dict[sec]['position']['z'] = (
            #wing_section_dict[template]['position']['z'])
            wing_geometry_dict['interpolated_values']['z_coordinate_leading_edge']
            [wing_section_dict[sec]['span_index']])
        # span_index = wing_section_dict[sec]['span_index']
        i += 1

    # -------------------------- START OF NEW VENT TANK CALCULATION -------------------------
    def calculate_wing_tanks_volume():
        """Calculate tank volumes (wing geometry does not change from here)."""
        # Initialize parameter.
        actual_volume_center_tank = 0
        actual_volume_without_center_tank = 0

        # Iterate through tank entities.
        wing_tank_entity_list = [key for key, value in dict_tank_design.items() if key.startswith('tank_')
                                and key[5:].isdigit() and value['tank_location'] == 'wing']
        for wing_tank_entity in wing_tank_entity_list:
            # Initialize empty dictionary and lists.
            dict_tank_design[wing_tank_entity]['geometry'] = {}
            dict_tank_design[wing_tank_entity]['geometry']['inner_surface'] = {}
            dict_tank_design[wing_tank_entity]['geometry']['outer_surface'] = {}
            inner_surface_str = str()
            outer_surface_str = str()

            if wing_symmetry:
                if number_of_wing_sections == 4:
                    # Set sections depending on tank position.
                    match dict_tank_design[wing_tank_entity]['tank_position']:
                        case 'inner_left' | 'inner_right':
                            inner_surface_str = 'section_0'
                            outer_surface_str = 'section_1'
                        case 'outer_left' | 'outer_right':
                            inner_surface_str = 'section_1'
                            outer_surface_str = 'section_2'
                        case 'center':
                            inner_surface_str = 'fuselage_center_line'
                            outer_surface_str = 'outer_center_tank_section'
                elif number_of_wing_sections == 3:
                    # Set sections depending on tank position.
                    match dict_tank_design[wing_tank_entity]['tank_position']:
                        case 'inner_left' | 'inner_right':
                            inner_surface_str = 'section_0'
                            outer_surface_str = 'section_1'
                        # case 'outer_left' | 'outer_right':
                        #     inner_surface_str = 'section_1'
                        #     outer_surface_str = 'section_2'
                        case 'center':
                            inner_surface_str = 'fuselage_center_line'
                            outer_surface_str = 'outer_center_tank_section'
                else:
                    runtime_output.critical('Error: Unknown wing geometry! Program aborted.')
                    sys.exit(1)
            else:
                runtime_output.critical('Error: Wing not symmetric! '
                                        + 'Program aborted.')
                sys.exit(1)

            # Calculate obelisk volume.
            obelisk_volume = calculate_obelisk_volume(dict_ac_data['obelisk_volume_calculation_method'],
                                                    wing_geometry_dict, inner_surface_str, outer_surface_str)
            # Calculate actual tank volume (considering volume loss due to internal structure of integral tanks and buffer
            # for temperature-dependent expansion of fuel).
            if dict_tank_design[wing_tank_entity]['tank_position'] == 'center':
                obelisk_volume *= 2
            actual_volume = factor_volume_usable*temperature_expansion_allowance*obelisk_volume
            # Sum up volume of all tanks except wing center tank.
            if dict_tank_design[wing_tank_entity]['tank_position'] != 'center':
                actual_volume_without_center_tank += actual_volume
            # Set volume of wing center tank.
            else:
                actual_volume_center_tank = actual_volume

            # Write data to 'dict_tank_design'.
            dict_tank_design[wing_tank_entity]['volume_available'] = actual_volume
            dict_tank_design[wing_tank_entity]['energy_available'] = (
                actual_volume*kerosene_volumetric_energy_density)
            # Print.
            # inner left wing tank (ID="0") calculated. Volume (energy) available: 3.51 l ().
            # f'{a*1000:,.2f}'
            # Transform volume to liter for print.
            actual_volume_l = actual_volume*1e3
            if dict_tank_design[wing_tank_entity]['tank_position'] == 'center' and print_outputs:
                runtime_output.print('Wing center tank (' + wing_tank_entity + ') calculated. '
                                    + 'Volume (energy) available: ' + f"{actual_volume_l:,.2f}" + ' L ('
                                    + f"{dict_tank_design[wing_tank_entity]['energy_available']/1e6:,.2f}" + ' MJ)')
            else:
                if print_outputs:
                    print_position = (dict_tank_design[wing_tank_entity]['tank_position'].split('_')[0].capitalize()
                                    + ' ' + dict_tank_design[wing_tank_entity]['tank_position'].split('_')[1] + ' ')
                    runtime_output.print(print_position + 'wing tank (' + wing_tank_entity + ') calculated. '
                                        + 'Volume (energy) available: ' + f"{actual_volume_l:,.2f}" + ' L ('
                                        + f"{dict_tank_design[wing_tank_entity]['energy_available']/1e6:,.2f}"
                                        + ' MJ)')

            # Calculate tank mass (equals zero due to integral tank).
            dict_tank_design[wing_tank_entity]['geometry']['total'] = {}
            dict_tank_design[wing_tank_entity]['geometry']['total']['mass'] = 0.0
            # Set inertia to zero.
            inertia_list = ['j_xx', 'j_yy', 'j_zz', 'j_xy', 'j_xz', 'j_yx', 'j_yz', 'j_zx', 'j_zy']
            dict_tank_design[wing_tank_entity]['geometry']['total']['inertia'] = {}
            for inertia in inertia_list:
                dict_tank_design[wing_tank_entity]['geometry']['total']['inertia'][inertia] = 0.0
            # Center of gravity equals zero since integral tank has no mass.
            dict_tank_design[wing_tank_entity]['geometry']['total']['center_of_gravity'] = {}
            dict_tank_design[wing_tank_entity]['geometry']['total']['center_of_gravity']['x'] = 0.0
            dict_tank_design[wing_tank_entity]['geometry']['total']['center_of_gravity']['y'] = 0.0
            dict_tank_design[wing_tank_entity]['geometry']['total']['center_of_gravity']['z'] = 0.0

            # Calculate position of tank entity (foremost point of tank).
            dict_tank_design[wing_tank_entity]['geometry']['total']['position'] = {}
            dict_tank_design[wing_tank_entity]['geometry']['total']['position']['x'] = (
                wing_geometry_dict['interpolated_values']['x_coordinate_front_spar'][
                    wing_section_dict[inner_surface_str]['span_index']
                ])
            if (dict_tank_design[wing_tank_entity]['tank_position'] == 'inner_left') or (
                dict_tank_design[wing_tank_entity]['tank_position'] == 'outer_left'):
                dict_tank_design[wing_tank_entity]['geometry']['total']['position']['y'] = -abs(
                    wing_section_dict[inner_surface_str]['position']['y'])
            else:
                dict_tank_design[wing_tank_entity]['geometry']['total']['position']['y'] = abs(
                    wing_section_dict[inner_surface_str]['position']['y'])
            dict_tank_design[wing_tank_entity]['geometry']['total']['position']['z'] = (
                wing_section_dict[inner_surface_str]['position']['z'])
            # Direction.
            dict_tank_design[wing_tank_entity]['geometry']['total']['direction'] = {}
            dict_tank_design[wing_tank_entity]['geometry']['total']['direction']['x'] = 0
            dict_tank_design[wing_tank_entity]['geometry']['total']['direction']['y'] = 1
            dict_tank_design[wing_tank_entity]['geometry']['total']['direction']['z'] = 0
            # Set required energy.
            dict_tank_design[wing_tank_entity]['energy_required_for_mission'] = True

            # Add required output data: x, y, and z position, shape, height, width, and length.
            for surface_str, surface_name in zip([inner_surface_str, outer_surface_str],
                                                ['inner_surface', 'outer_surface']):
                surface_data = wing_section_dict[surface_str]
                span_idx = surface_data['span_index']
                # Inner surface equals tank entity position.
                if surface_name == 'inner_surface':
                    x_coordinate = 0.0
                    y_coordinate = 0.0
                    z_coordinate = 0.0
                else:
                    # Adjust y coordinate for left hand tanks.
                    y_coordinate_entity = dict_tank_design[wing_tank_entity]['geometry']['total']['position']['y']
                    if 'left' in dict_tank_design[wing_tank_entity]['tank_position']:
                        y_coordinate_section = -surface_data['position']['y']
                        y_coordinate = -(abs(y_coordinate_section) - abs(y_coordinate_entity))
                    else:
                        y_coordinate_section = surface_data['position']['y']
                        y_coordinate = (abs(y_coordinate_section) - abs(y_coordinate_entity))
                    # old
                    # y_coordinate = y_coordinate_section - y_coordinate_entity
                    # NEW:
                    # y_coordinate = -(abs(y_coordinate_section) - abs(y_coordinate_entity))
                    # if 'left' in dict_tank_design[wing_tank_entity]['tank_position']:
                    #     y_coordinate = y_coordinate*(-1)
                    # x coordinate.
                    x_coordinate_entity = dict_tank_design[wing_tank_entity]['geometry']['total']['position']['x']
                    x_coordinate_section = wing_geometry_dict['interpolated_values']['x_coordinate_front_spar'][span_idx]
                    if x_coordinate_entity < 0 and x_coordinate_section > 0 and x_coordinate_section > x_coordinate_entity:
                        delta_x = x_coordinate_entity - x_coordinate_section
                    else:
                        delta_x = x_coordinate_section - x_coordinate_entity
                    x_coordinate = delta_x
                    # z coordinate.
                    z_coordinate_entity = dict_tank_design[wing_tank_entity]['geometry']['total']['position']['z']
                    z_coordinate_section = surface_data['position']['z']
                    z_coordinate = z_coordinate_section - z_coordinate_entity

                dict_tank_design[wing_tank_entity]['geometry'][surface_name] = {
                    'section_name': surface_str,
                    'position': {
                        'x': x_coordinate,
                        'y': y_coordinate,
                        'z': z_coordinate,
                    },
                    'shape': 'rectangular',
                    'height': wing_geometry_dict['interpolated_values']['tank_heights'][span_idx],
                    'width': wing_geometry_dict['interpolated_values']['tank_lengths'][span_idx],
                    'length': 0.0
                }

            # Correct wing center tank information.
            if 'center' in dict_tank_design[wing_tank_entity]['tank_position']:
                dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['position']['y'] = (
                    dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['position']['y']/2
                )
                dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['position']['y'] = -(
                    dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['position']['y']
                )

            # Centroid.
            # Assumption: inner surface area always greater than outer surface area.
            inner_surface_area = (dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
                                *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
            outer_surface_area = (dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
                            *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])
            y_position_outer_surface = dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['position']['y']
            y_position_inner_surface = dict_tank_design[wing_tank_entity]['geometry']['total']['position']['y']
            tank_span = y_position_outer_surface - y_position_inner_surface
            y_position_center_of_mass = y_position_inner_surface + (tank_span/4*((inner_surface_area
                                                    + 2*math.sqrt(inner_surface_area*outer_surface_area)
                                                    + outer_surface_area)/(inner_surface_area + outer_surface_area)))
            centroid_y = y_position_inner_surface + ((tank_span
                        *((dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
                            *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
                            +(dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
                            *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])
                            +(dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
                            *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
                            +(3*dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
                            *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])))
                        /(2
                            *(2*dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
                            *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
                            + (dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
                            *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])
                            + (dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
                            *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
                            + (2*dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
                            *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])))
            x_position_cog_inner_surface = dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width']/2
            x_position_cog_outer_surface = dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width']/2
            z_position_cog_inner_surface = dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']/2
            z_position_cog_outer_surface = dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']/2
            x_position_center_of_mass = ((inner_surface_area*x_position_cog_inner_surface
                                        + outer_surface_area*x_position_cog_outer_surface)
                                        /(inner_surface_area + outer_surface_area))
            z_position_center_of_mass = ((inner_surface_area*z_position_cog_inner_surface
                                        + outer_surface_area*z_position_cog_outer_surface)
                                        /(inner_surface_area + outer_surface_area))

            dict_tank_design[wing_tank_entity]['geometry']['total']['centroid'] = {}
            if wing_tank_entity == wing_center_tank_entity:
                dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['x'] = x_position_center_of_mass
                # dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['x'] = (
                #     dict_tank_design['tank_4']['geometry']['inner_surface']['position']['x']
                #     + dict_tank_design['tank_4']['geometry']['inner_surface']['width']/2
                # )
                dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['y'] = 0
                dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['z'] = 0
            else:
                dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['x'] = x_position_center_of_mass
                dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['y'] = y_position_center_of_mass
                dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['z'] = z_position_center_of_mass

        # Sum up.
        volume_wing_tank_without_center_tank = actual_volume_without_center_tank
        volume_wing_tank_with_center_tank = actual_volume_without_center_tank + actual_volume_center_tank

        """Calculate fuel mass and energy."""
        # Convert wing fuel volume to energy.
        energy_wing_tank_without_center_tank = (volume_wing_tank_without_center_tank
                                                *kerosene_volumetric_energy_density)
        energy_wing_tank_with_center_tank = (volume_wing_tank_with_center_tank
                                            *kerosene_volumetric_energy_density)
            
        return energy_wing_tank_without_center_tank, energy_wing_tank_with_center_tank


    def calculate_vent_tank_volume(wing_tank_energy_amount):
        """Calculate vent tank."""
        # § 25.969 Fuel tank expansion space. Each fuel tank must have an expansion space of not less than 2 percent
        # of the tank capacity. It must be impossible to fill the expansion space inadvertently with the airplane in the
        # normal ground attitude. For pressure fueling systems, compliance with this section may be shown with the means
        # provided to comply with §25.979(b).
        # Add vent tank to tank design dict.
        dict_tank_design['vent_tank'] = {}
        dict_tank_design['vent_tank']['tank_location'] = 'wing'
        dict_tank_design['vent_tank']['tank_position'] = 'vent'
        # CS 25.969: Vent tank capacity must at least be equal to 2 percent of tank capacity (1 percent per side).
        dict_tank_design['vent_tank']['energy_required'] = 0.01*wing_tank_energy_amount
        vent_tank_outer_surface = 'outer_vent_tank_section'

        # Calculate energy contained in vent tank iteratively. Position of inner surface is adjusted in 1 mm steps.
        vent_tank_energy = 0
        while vent_tank_energy < dict_tank_design['vent_tank']['energy_required']:
            vent_tank_volume = calculate_obelisk_volume(dict_ac_data['obelisk_volume_calculation_method'],
                                                        wing_geometry_dict,
                                                        vent_tank_inner_surface, vent_tank_outer_surface)
            vent_tank_volume_actual = factor_volume_usable*temperature_expansion_allowance*vent_tank_volume
            vent_tank_energy = vent_tank_volume_actual*kerosene_volumetric_energy_density
            wing_section_dict[vent_tank_inner_surface]['span_index'] -= 1

        # Update x position of 'section_2' aka. 'vent_tank_inner_surface'.
        span_idx = wing_section_dict[vent_tank_inner_surface]['span_index']
        wing_section_dict[vent_tank_inner_surface]['position']['x'] = (
            wing_geometry_dict['interpolated_values']['x_coordinate_leading_edge'][span_idx])
        wing_section_dict[vent_tank_inner_surface]['position']['y'] = span_idx/100

        # TODO: Eventuell löschen, wenn plot nicht mehr benötigt.
        dict_tank_design['vent_tank']['geometry'] = {}
        dict_tank_design['vent_tank']['geometry']['total'] = {}
        dict_tank_design['vent_tank']['geometry']['total']['position'] = {}
        dict_tank_design['vent_tank']['geometry']['total']['position']['x'] = (
            wing_geometry_dict['interpolated_values']['x_coordinate_front_spar'][span_idx])
        dict_tank_design['vent_tank']['geometry']['total']['position']['y'] = (
            wing_section_dict[vent_tank_inner_surface]['position']['y'])
        dict_tank_design['vent_tank']['geometry']['inner_surface'] = {}
        dict_tank_design['vent_tank']['geometry']['inner_surface']['width'] = (
            wing_geometry_dict['interpolated_values']['tank_lengths'][span_idx])
        dict_tank_design['vent_tank']['volume_available'] = vent_tank_volume_actual
        dict_tank_design['vent_tank']['energy_available'] = vent_tank_energy
        # Print.
        runtime_output.print('Vent tank (located near each wing tip) calculated.')
    
    print_outputs = False
    energy_wing_tank_without_center_tank, energy_wing_tank_with_center_tank = calculate_wing_tanks_volume()
    print_outputs = True
    calculate_vent_tank_volume(energy_wing_tank_with_center_tank)
    energy_wing_tank_without_center_tank, energy_wing_tank_with_center_tank = calculate_wing_tanks_volume()  
    # -------------------------- END OF NEW VENT TANK CALCULATION -------------------------

    # -------------------------- START OF OLD VENT TANK CALCULATION -------------------------
    # """Calculate vent tank."""
    # # § 25.969 Fuel tank expansion space. Each fuel tank must have an expansion space of not less than 2 percent
    # # of the tank capacity. It must be impossible to fill the expansion space inadvertently with the airplane in the
    # # normal ground attitude. For pressure fueling systems, compliance with this section may be shown with the means
    # # provided to comply with §25.979(b).
    # # Add vent tank to tank design dict.
    # dict_tank_design['vent_tank'] = {}
    # dict_tank_design['vent_tank']['tank_location'] = 'wing'
    # dict_tank_design['vent_tank']['tank_position'] = 'vent'
    # # CS 25.969: Vent tank capacity must at least be equal to 2 percent of tank capacity (1 percent per side).
    # dict_tank_design['vent_tank']['energy_required'] = 0.01*mission_energy_amount
    # vent_tank_outer_surface = 'outer_vent_tank_section'

    # # Calculate energy contained in vent tank iteratively. Position of inner surface is adjusted in 1 mm steps.
    # vent_tank_energy = 0
    # while vent_tank_energy < dict_tank_design['vent_tank']['energy_required']:
    #     vent_tank_volume = calculate_obelisk_volume(dict_ac_data['obelisk_volume_calculation_method'],
    #                                                 wing_geometry_dict,
    #                                                 vent_tank_inner_surface, vent_tank_outer_surface)
    #     vent_tank_volume_actual = factor_volume_usable*temperature_expansion_allowance*vent_tank_volume
    #     vent_tank_energy = vent_tank_volume_actual*kerosene_volumetric_energy_density
    #     wing_section_dict[vent_tank_inner_surface]['span_index'] -= 1

    # # Update x position of 'section_2' aka. 'vent_tank_inner_surface'.
    # span_idx = wing_section_dict[vent_tank_inner_surface]['span_index']
    # wing_section_dict[vent_tank_inner_surface]['position']['x'] = (
    #     wing_geometry_dict['interpolated_values']['x_coordinate_leading_edge'][span_idx])
    # wing_section_dict[vent_tank_inner_surface]['position']['y'] = span_idx/100

    # # TODO: Eventuell löschen, wenn plot nicht mehr benötigt.
    # dict_tank_design['vent_tank']['geometry'] = {}
    # dict_tank_design['vent_tank']['geometry']['total'] = {}
    # dict_tank_design['vent_tank']['geometry']['total']['position'] = {}
    # dict_tank_design['vent_tank']['geometry']['total']['position']['x'] = (
    #     wing_geometry_dict['interpolated_values']['x_coordinate_front_spar'][span_idx])
    # dict_tank_design['vent_tank']['geometry']['total']['position']['y'] = (
    #     wing_section_dict[vent_tank_inner_surface]['position']['y'])
    # dict_tank_design['vent_tank']['geometry']['inner_surface'] = {}
    # dict_tank_design['vent_tank']['geometry']['inner_surface']['width'] = (
    #     wing_geometry_dict['interpolated_values']['tank_lengths'][span_idx])
    # dict_tank_design['vent_tank']['volume_available'] = vent_tank_volume_actual
    # dict_tank_design['vent_tank']['energy_available'] = vent_tank_energy
    # # Print.
    # runtime_output.print('Vent tank (located near each wing tip) calculated.')

    # """Calculate tank volumes (wing geometry does not change from here)."""
    # # Initialize parameter.
    # actual_volume_center_tank = 0
    # actual_volume_without_center_tank = 0

    # # Iterate through tank entities.
    # wing_tank_entity_list = [key for key, value in dict_tank_design.items() if key.startswith('tank_')
    #                          and key[5:].isdigit() and value['tank_location'] == 'wing']
    # for wing_tank_entity in wing_tank_entity_list:
    #     # Initialize empty dictionary and lists.
    #     dict_tank_design[wing_tank_entity]['geometry'] = {}
    #     dict_tank_design[wing_tank_entity]['geometry']['inner_surface'] = {}
    #     dict_tank_design[wing_tank_entity]['geometry']['outer_surface'] = {}
    #     inner_surface_str = str()
    #     outer_surface_str = str()

    #     if wing_symmetry:
    #         if number_of_wing_sections == 4:
    #             # Set sections depending on tank position.
    #             match dict_tank_design[wing_tank_entity]['tank_position']:
    #                 case 'inner_left' | 'inner_right':
    #                     inner_surface_str = 'section_0'
    #                     outer_surface_str = 'section_1'
    #                 case 'outer_left' | 'outer_right':
    #                     inner_surface_str = 'section_1'
    #                     outer_surface_str = 'section_2'
    #                 case 'center':
    #                     inner_surface_str = 'fuselage_center_line'
    #                     outer_surface_str = 'outer_center_tank_section'
    #         else:
    #             runtime_output.critical('Error: Wing without kink not implemented yet! Program aborted.')
    #             sys.exit(1)
    #     else:
    #         runtime_output.critical('Error: Wing not symmetric! '
    #                                 + 'Program aborted.')
    #         sys.exit(1)

    #     # Calculate obelisk volume.
    #     obelisk_volume = calculate_obelisk_volume(dict_ac_data['obelisk_volume_calculation_method'],
    #                                               wing_geometry_dict, inner_surface_str, outer_surface_str)
    #     # Calculate actual tank volume (considering volume loss due to internal structure of integral tanks and buffer
    #     # for temperature-dependent expansion of fuel).
    #     if dict_tank_design[wing_tank_entity]['tank_position'] == 'center':
    #         obelisk_volume *= 2
    #     actual_volume = factor_volume_usable*temperature_expansion_allowance*obelisk_volume
    #     # Sum up volume of all tanks except wing center tank.
    #     if dict_tank_design[wing_tank_entity]['tank_position'] != 'center':
    #         actual_volume_without_center_tank += actual_volume
    #     # Set volume of wing center tank.
    #     else:
    #         actual_volume_center_tank = actual_volume

    #     # Write data to 'dict_tank_design'.
    #     dict_tank_design[wing_tank_entity]['volume_available'] = actual_volume
    #     dict_tank_design[wing_tank_entity]['energy_available'] = (
    #         actual_volume*kerosene_volumetric_energy_density)
    #     # Print.
    #     # inner left wing tank (ID="0") calculated. Volume (energy) available: 3.51 l ().
    #     # f'{a*1000:,.2f}'
    #     if dict_tank_design[wing_tank_entity]['tank_position'] == 'center':
    #         runtime_output.print('Wing center tank (' + wing_tank_entity + ') calculated. '
    #                             + 'Volume (energy) available: ' + f"{actual_volume:,.2f}" + ' L ('
    #                             + f"{dict_tank_design[wing_tank_entity]['energy_available']/1e6:,.2f}" + ' MJ)')
    #     else:
    #         print_position = (dict_tank_design[wing_tank_entity]['tank_position'].split('_')[0].capitalize()
    #                         + ' ' + dict_tank_design[wing_tank_entity]['tank_position'].split('_')[1] + ' ')
    #         runtime_output.print(print_position + 'wing tank (' + wing_tank_entity + ') calculated. '
    #                             + 'Volume (energy) available: ' + f"{actual_volume:,.2f}" + ' L ('
    #                             + f"{dict_tank_design[wing_tank_entity]['energy_available']/1e6:,.2f}" + ' MJ)')

    #     # Calculate tank mass (equals zero due to integral tank).
    #     dict_tank_design[wing_tank_entity]['geometry']['total'] = {}
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['mass'] = 0.0
    #     # Set inertia to zero.
    #     inertia_list = ['j_xx', 'j_yy', 'j_zz', 'j_xy', 'j_xz', 'j_yx', 'j_yz', 'j_zx', 'j_zy']
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['inertia'] = {}
    #     for inertia in inertia_list:
    #         dict_tank_design[wing_tank_entity]['geometry']['total']['inertia'][inertia] = 0.0
    #     # Center of gravity equals zero since integral tank has no mass.
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['center_of_gravity'] = {}
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['center_of_gravity']['x'] = 0.0
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['center_of_gravity']['y'] = 0.0
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['center_of_gravity']['z'] = 0.0

    #     # Calculate position of tank entity (foremost point of tank).
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['position'] = {}
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['position']['x'] = (
    #         wing_geometry_dict['interpolated_values']['x_coordinate_front_spar'][
    #             wing_section_dict[inner_surface_str]['span_index']
    #         ])
    #     if (dict_tank_design[wing_tank_entity]['tank_position'] == 'inner_left') or (
    #         dict_tank_design[wing_tank_entity]['tank_position'] == 'outer_left'):
    #         dict_tank_design[wing_tank_entity]['geometry']['total']['position']['y'] = -abs(
    #             wing_section_dict[inner_surface_str]['position']['y'])
    #     else:
    #         dict_tank_design[wing_tank_entity]['geometry']['total']['position']['y'] = abs(
    #             wing_section_dict[inner_surface_str]['position']['y'])
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['position']['z'] = (
    #         wing_section_dict[inner_surface_str]['position']['z'])
    #     # Direction.
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['direction'] = {}
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['direction']['x'] = 0
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['direction']['y'] = 1
    #     dict_tank_design[wing_tank_entity]['geometry']['total']['direction']['z'] = 0
    #     # Set required energy.
    #     dict_tank_design[wing_tank_entity]['energy_required_for_mission'] = True

    #     # Add required output data: x, y, and z position, shape, height, width, and length.
    #     for surface_str, surface_name in zip([inner_surface_str, outer_surface_str],
    #                                          ['inner_surface', 'outer_surface']):
    #         surface_data = wing_section_dict[surface_str]
    #         span_idx = surface_data['span_index']
    #         # Inner surface equals tank entity position.
    #         if surface_name == 'inner_surface':
    #             x_coordinate = 0.0
    #             y_coordinate = 0.0
    #             z_coordinate = 0.0
    #         else:
    #             # Adjust y coordinate for left hand tanks.
    #             y_coordinate_entity = dict_tank_design[wing_tank_entity]['geometry']['total']['position']['y']
    #             if 'left' in dict_tank_design[wing_tank_entity]['tank_position']:
    #                 y_coordinate_section = -surface_data['position']['y']
    #                 y_coordinate = -(abs(y_coordinate_section) - abs(y_coordinate_entity))
    #             else:
    #                 y_coordinate_section = surface_data['position']['y']
    #                 y_coordinate = (abs(y_coordinate_section) - abs(y_coordinate_entity))
    #             # old
    #             # y_coordinate = y_coordinate_section - y_coordinate_entity
    #             # NEW:
    #             # y_coordinate = -(abs(y_coordinate_section) - abs(y_coordinate_entity))
    #             # if 'left' in dict_tank_design[wing_tank_entity]['tank_position']:
    #             #     y_coordinate = y_coordinate*(-1)
    #             # x coordinate.
    #             x_coordinate_entity = dict_tank_design[wing_tank_entity]['geometry']['total']['position']['x']
    #             x_coordinate_section = wing_geometry_dict['interpolated_values']['x_coordinate_front_spar'][span_idx]
    #             if x_coordinate_entity < 0 and x_coordinate_section > 0 and x_coordinate_section > x_coordinate_entity:
    #                 delta_x = x_coordinate_entity - x_coordinate_section
    #             else:
    #                 delta_x = x_coordinate_section - x_coordinate_entity
    #             x_coordinate = delta_x
    #             # z coordinate.
    #             z_coordinate_entity = dict_tank_design[wing_tank_entity]['geometry']['total']['position']['z']
    #             z_coordinate_section = surface_data['position']['z']
    #             z_coordinate = z_coordinate_section - z_coordinate_entity

    #         dict_tank_design[wing_tank_entity]['geometry'][surface_name] = {
    #             'section_name': surface_str,
    #             'position': {
    #                 'x': x_coordinate,
    #                 'y': y_coordinate,
    #                 'z': z_coordinate,
    #             },
    #             'shape': 'rectangular',
    #             'height': wing_geometry_dict['interpolated_values']['tank_heights'][span_idx],
    #             'width': wing_geometry_dict['interpolated_values']['tank_lengths'][span_idx],
    #             'length': 0.0
    #         }

    #     # Correct wing center tank information.
    #     if 'center' in dict_tank_design[wing_tank_entity]['tank_position']:
    #         dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['position']['y'] = (
    #             dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['position']['y']/2
    #         )
    #         dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['position']['y'] = -(
    #             dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['position']['y']
    #         )

    #     # Centroid.
    #     # Assumption: inner surface area always greater than outer surface area.
    #     inner_surface_area = (dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
    #                           *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
    #     outer_surface_area = (dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
    #                     *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])
    #     y_position_outer_surface = dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['position']['y']
    #     y_position_inner_surface = dict_tank_design[wing_tank_entity]['geometry']['total']['position']['y']
    #     tank_span = y_position_outer_surface - y_position_inner_surface
    #     y_position_center_of_mass = y_position_inner_surface + (tank_span/4*((inner_surface_area
    #                                               + 2*math.sqrt(inner_surface_area*outer_surface_area)
    #                                               + outer_surface_area)/(inner_surface_area + outer_surface_area)))
    #     centroid_y = y_position_inner_surface + ((tank_span
    #                    *((dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
    #                       *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
    #                      +(dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
    #                        *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])
    #                      +(dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
    #                        *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
    #                      +(3*dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
    #                        *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])))
    #                   /(2
    #                     *(2*dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
    #                       *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
    #                     + (dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']
    #                        *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])
    #                     + (dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
    #                        *dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width'])
    #                     + (2*dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']
    #                        *dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width'])))
    #     x_position_cog_inner_surface = dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['width']/2
    #     x_position_cog_outer_surface = dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['width']/2
    #     z_position_cog_inner_surface = dict_tank_design[wing_tank_entity]['geometry']['inner_surface']['height']/2
    #     z_position_cog_outer_surface = dict_tank_design[wing_tank_entity]['geometry']['outer_surface']['height']/2
    #     x_position_center_of_mass = ((inner_surface_area*x_position_cog_inner_surface
    #                                  + outer_surface_area*x_position_cog_outer_surface)
    #                                  /(inner_surface_area + outer_surface_area))
    #     z_position_center_of_mass = ((inner_surface_area*z_position_cog_inner_surface
    #                                  + outer_surface_area*z_position_cog_outer_surface)
    #                                  /(inner_surface_area + outer_surface_area))

    #     dict_tank_design[wing_tank_entity]['geometry']['total']['centroid'] = {}
    #     if wing_tank_entity == wing_center_tank_entity:
    #         dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['x'] = x_position_center_of_mass
    #         # dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['x'] = (
    #         #     dict_tank_design['tank_4']['geometry']['inner_surface']['position']['x']
    #         #     + dict_tank_design['tank_4']['geometry']['inner_surface']['width']/2
    #         # )
    #         dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['y'] = 0
    #         dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['z'] = 0
    #     else:
    #         dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['x'] = x_position_center_of_mass
    #         dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['y'] = y_position_center_of_mass
    #         dict_tank_design[wing_tank_entity]['geometry']['total']['centroid']['z'] = z_position_center_of_mass

    # # Sum up.
    # volume_wing_tank_without_center_tank = actual_volume_without_center_tank
    # volume_wing_tank_with_center_tank = actual_volume_without_center_tank + actual_volume_center_tank

    # """Calculate fuel mass and energy."""
    # # Convert wing fuel volume to energy.
    # energy_wing_tank_without_center_tank = (volume_wing_tank_without_center_tank
    #                                         *kerosene_volumetric_energy_density)
    # energy_wing_tank_with_center_tank = (volume_wing_tank_with_center_tank
    #                                      *kerosene_volumetric_energy_density)
    
    # -------------------------- END OF OLD VENT TANK CALCULATION -------------------------

    """Check if tanks can store required energy."""
    # If left and right inner and outer wing tanks are big enough to store required amount of energy, wing center tank
    # remains empty.
    if energy_wing_tank_without_center_tank >= mission_energy_amount:
        dict_tank_design[wing_center_tank_entity]['volume_available'] = 0
        dict_tank_design[wing_center_tank_entity]['energy_available'] = 0

    # Prints.
    runtime_output.print('Wing tanks successfully calculated.')
    runtime_output.debug('Debug: The "calculate_wing_tanks" function was successfully executed.')

    # Prepare data output.
    dict_tank_design['tmp_energies'] = {}
    dict_tank_design['tmp_energies']['wing_tank_without_center_tank'] = float(energy_wing_tank_without_center_tank)
    dict_tank_design['tmp_energies']['wing_tank_with_center_tank'] = float(energy_wing_tank_with_center_tank)

    return dict_tank_design


def calculate_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output, energy_demand_covered):
    """Calculate energy contained in additional center tank.
    
    This code enables the installation of an additional center tank as LD3-45 container. Therefore, it is checked
    initially whether the existing cargo compartment provides sufficient height. If this is not the case, all export
    values are set to zero. Otherwise, the LD3-45 container is installed 10 cm behind the end of the landing gear bay.
    It is assumed that this is approximately in line with the trailing edge of the wing. The volume and dimensions of
    the container are taken from [1].

    Attention: Please note that it is not possible to calculate more than one additional center tank at the moment!

    [1] https://www.lufthansa-cargo.com/de/fleet-ulds/ulds/containers

    :param dict dict_design: Dictionary containing data from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :param energy_demand_covered: Information if mission energy demand already covered
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    # Extract necessary data from 'dict_ac_data'.
    factor_volume_usable = dict_ac_data['factor_volume_usable']
    temperature_expansion_allowance = dict_ac_data['temperature_expansion_allowance']
    
    # Extract kerosene energy density from library.
    kerosene_volumetric_energy_density = constants.KEROSENE_VOLUMETRIC_ENERGY_DENSITY
    
    # Print.
    runtime_output.print('Additional center tank design started...')

    # Get all tank entities that are located at the fuselage (additional center tanks).
    additional_center_tank_entity_list = [k for k, v in dict_tank_design.items() if k.startswith('tank_')
                                          and k[5:].isdigit() and v['tank_location'] == 'fuselage']
    number_of_additional_center_tanks = len(additional_center_tank_entity_list)
    if number_of_additional_center_tanks > 1:
        runtime_output.critical('Error: Tank design not possible for more than one additional center tank! '
                                + 'Program aborted.')
        sys.exit('Exit information: Tank design not possible for more than one additional center tank!')
    # Extract identifier of center tank.
    center_tank_entity = additional_center_tank_entity_list[0]

    """Prepare data of LD3-45 container."""
    # Assumption: Angle at base equals 45 degree and usable volume of container taken from [1] (not calculated).
    height_container = 1.15
    width_base_container = 1.53
    width_top_container = 2.44
    length_container = 1.56
    height_base_part_container = (width_top_container - width_base_container)*0.5
    height_top_part_container = height_container - height_base_part_container
    volume_container = 3.4
    # Calculate areas and volumes for later use.
    area_base_part = width_base_container*length_container
    area_top_part = width_top_container*length_container
    volume_base_part = (
        (height_base_part_container*width_top_container*length_container)
        - height_base_part_container*height_base_part_container*length_container)
    volume_top_part = width_top_container*length_container*height_top_part_container

    # Check, if cargo compartment height sufficient to store LD3-45 container.
    if dict_tank_design['geometry']['fuselage']['cargo_compartment_height'] < height_container:
        additional_center_tank_design_possible = False
        runtime_output.warning('Warning: Cargo compartment height too small. Storage of LD3-45 container not '
                               + 'possible. All values are set to 0.')
    else:
        additional_center_tank_design_possible = True

    """Calculate additional center tank."""
    # Calculate usable volume and convert to energy.
    center_tank_volume = volume_container*factor_volume_usable*temperature_expansion_allowance
    energy_kerosene_available = center_tank_volume*kerosene_volumetric_energy_density
    center_tank_volume_l = center_tank_volume*1e3
    # Calculate ACT and necessary values if possible.
    if additional_center_tank_design_possible:
        # Write data to 'dict_tank_design'.
        dict_tank_design[center_tank_entity]['volume_available'] = center_tank_volume
        dict_tank_design[center_tank_entity]['energy_available'] = energy_kerosene_available
        if energy_demand_covered:
            dict_tank_design[center_tank_entity]['energy_required_for_mission'] = False
        else:
            dict_tank_design[center_tank_entity]['energy_required_for_mission'] = True
        
        # Print.
        runtime_output.print('Additional center tank (' + center_tank_entity + ') calculated. '
                             + 'Volume (energy) available: ' + f"{center_tank_volume_l:,.2f}" + ' L ('
                                + f"{dict_tank_design[center_tank_entity]['energy_available']/1e6:,.2f}" + ' MJ)')

        # Prepare geometrical data for additional center tank.
        # Assumption: Landing gear bay ends at wing trailing edge position.
        # x_coordinate_wing_leading_edge = dict_ac_data['wing_position_x']
        chord_length_at_fuselage_center_line = (
            dict_tank_design['geometry']['wing']['sections']['fuselage_center_line']['chord_length'])
        # x_coordinate_wing_trailing_edge = x_coordinate_wing_leading_edge + chord_length_at_fuselage_center_line
        buffer = 0.1
        x_coordinate = chord_length_at_fuselage_center_line + buffer
        y_coordinate = [width_top_container*0.5, width_base_container*0.5, 0.0,
                        -width_base_container*0.5, -width_top_container*0.5]
        cargo_floor_below_wing = (dict_tank_design['geometry']['fuselage']['z_coordinate_cargo_floor']
                                  < dict_tank_design['geometry']['wing']['position']['z'])
        z_offset_wing_to_cargo_floor = abs(dict_tank_design['geometry']['wing']['position']['z']
                                           - dict_tank_design['geometry']['fuselage']['z_coordinate_cargo_floor'])
        if cargo_floor_below_wing:
            relative_z_position_cargo_floor_to_wing = -z_offset_wing_to_cargo_floor
        else:
            relative_z_position_cargo_floor_to_wing = z_offset_wing_to_cargo_floor
        z_coordinate = relative_z_position_cargo_floor_to_wing + np.array([
            # plane_0
            height_base_part_container + height_top_part_container*0.5,
            # plane_1, plane_2, plane_3
            height_container*0.5, height_container*0.5, height_container*0.5,
            # plane_4
            height_base_part_container + height_top_part_container*0.5])
        heights = [height_top_part_container, height_container, height_container,
                height_container, height_top_part_container]
        plane_names = ['cross_section_0', 'cross_section_1', 'cross_section_2', 'cross_section_3', 'cross_section_4']

        # Create the dictionary for center tank geometry.
        dict_tank_design[center_tank_entity]['geometry'] = {
            plane: {
                'section_name': plane,
                'position': {
                    'x': x_coordinate,
                    'y': y_coordinate[i],
                    'z': z_coordinate[i]
                },
                'shape': 'rectangular',
                'height': heights[i],
                'width': length_container,
                'length': 0.0
            }
            for i, plane in enumerate(plane_names)
        }

        # Define total tank geometry values.
        # Calculate tank mass (equals zero due to integral tank).
        dict_tank_design[center_tank_entity]['geometry']['total'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['mass'] = 300
        # Set inertia to zero.
        inertia_list = ['j_xx', 'j_yy', 'j_zz', 'j_xy', 'j_xz', 'j_yx', 'j_yz', 'j_zx', 'j_zy']
        dict_tank_design[center_tank_entity]['geometry']['total']['inertia'] = {}
        for inertia in inertia_list:
            dict_tank_design[center_tank_entity]['geometry']['total']['inertia'][inertia] = 0.0
        # Center of gravity.
        # Prepare data for z coordinate calculation.
        h = height_base_part_container
        a_1 = area_top_part
        a_2 = area_base_part
        y_0 = ((h*(a_1 + 2*math.sqrt(a_1*a_2) + 3*a_2))
            /(4*(a_1 + math.sqrt(a_1*a_2)) + a_2))
        z_coordinate_cog_base_part = (dict_tank_design['geometry']['fuselage']['z_coordinate_cargo_floor']
                                    + (h - y_0))
        z_coordinate_cog_top_part = (dict_tank_design['geometry']['fuselage']['z_coordinate_cargo_floor']
                                    + height_base_part_container + height_top_part_container*0.5)
        z_cog = ((z_coordinate_cog_base_part*volume_base_part + z_coordinate_cog_top_part*volume_top_part)
                /(volume_base_part + volume_top_part))

        dict_tank_design[center_tank_entity]['geometry']['total']['center_of_gravity'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['center_of_gravity']['x'] = (
            x_coordinate + length_container*0.5)
        dict_tank_design[center_tank_entity]['geometry']['total']['center_of_gravity']['y'] = 0.0
        dict_tank_design[center_tank_entity]['geometry']['total']['center_of_gravity']['z'] = z_cog

        # Calculate position of tank entity (foremost point of tank).
        dict_tank_design[center_tank_entity]['geometry']['total']['position'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['position']['x'] = x_coordinate
        dict_tank_design[center_tank_entity]['geometry']['total']['position']['y'] = 0.0
        dict_tank_design[center_tank_entity]['geometry']['total']['position']['z'] = z_coordinate[2]
        # Direction.
        dict_tank_design[center_tank_entity]['geometry']['total']['direction'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['direction']['x'] = 0
        dict_tank_design[center_tank_entity]['geometry']['total']['direction']['y'] = 1
        dict_tank_design[center_tank_entity]['geometry']['total']['direction']['z'] = 0

        # Centroid.
        dict_tank_design[center_tank_entity]['geometry']['total']['centroid'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['centroid']['x'] = length_container/2
        dict_tank_design[center_tank_entity]['geometry']['total']['centroid']['y'] = 0
        dict_tank_design[center_tank_entity]['geometry']['total']['centroid']['z'] = z_cog

    else:
        # Set all export values to 0 if tank design not possible.
        dict_tank_design[center_tank_entity]['geometry'] = {
            'cross_section_0': {
                'section_name': 'cross_section_0',
                'position': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                },
                'shape': None,
                'height': 0.0,
                'width': 0.0,
                'length': 0.0
            }
        }
        # Tank mass.
        dict_tank_design[center_tank_entity]['geometry']['total'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['mass'] = 0.0
        # Inertia.
        inertia_list = ['j_xx', 'j_yy', 'j_zz', 'j_xy', 'j_xz', 'j_yx', 'j_yz', 'j_zx', 'j_zy']
        dict_tank_design[center_tank_entity]['geometry']['total']['inertia'] = {}
        for inertia in inertia_list:
            dict_tank_design[center_tank_entity]['geometry']['total']['inertia'][inertia] = 0.0
        # Center of gravity.
        dict_tank_design[center_tank_entity]['geometry']['total']['center_of_gravity'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['center_of_gravity']['x'] = 0.0
        dict_tank_design[center_tank_entity]['geometry']['total']['center_of_gravity']['y'] = 0.0
        dict_tank_design[center_tank_entity]['geometry']['total']['center_of_gravity']['z'] = 0.0
        # Position of tank entity (foremost point of tank).
        dict_tank_design[center_tank_entity]['geometry']['total']['position'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['position']['x'] = 0.0
        dict_tank_design[center_tank_entity]['geometry']['total']['position']['y'] = 0.0
        dict_tank_design[center_tank_entity]['geometry']['total']['position']['z'] = 0.0
        # Direction.
        dict_tank_design[center_tank_entity]['geometry']['total']['direction'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['direction']['x'] = 0
        dict_tank_design[center_tank_entity]['geometry']['total']['direction']['y'] = 1
        dict_tank_design[center_tank_entity]['geometry']['total']['direction']['z'] = 0
        # Centroid.
        dict_tank_design[center_tank_entity]['geometry']['total']['centroid'] = {}
        dict_tank_design[center_tank_entity]['geometry']['total']['centroid']['x'] = 0
        dict_tank_design[center_tank_entity]['geometry']['total']['centroid']['y'] = 0
        dict_tank_design[center_tank_entity]['geometry']['total']['centroid']['z'] = 0

    # Prints.
    runtime_output.print('Additional center tank design completed.')
    runtime_output.debug('Debug: The "calculate_additional_center_tank" function was successfully executed.')

    # Prepare data output.
    dict_tank_design['tmp_energies'] = {}
    dict_tank_design['tmp_energies']['additional_center_tank'] = float(energy_kerosene_available)

    return dict_tank_design


def calculate_trim_tank(dict_ac_data, dict_tank_design, runtime_output, energy_demand_covered):
    """Calculate energy contained in trim tank.
    
    [Add some more information here...]

    :param dict dict_design: Dictionary containing data from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :param energy_demand_covered: Information if mission energy demand already covered
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    #TODO: Implement vent tank design from wing tank.
    # Extract values from 'dict_ac_data'.
    factor_volume_usable = dict_ac_data['factor_volume_usable']
    temperature_expansion_allowance = dict_ac_data['temperature_expansion_allowance']
    # Extract kerosene energy density from library.
    kerosene_volumetric_energy_density = constants.KEROSENE_VOLUMETRIC_ENERGY_DENSITY
    
    # Print.
    runtime_output.print('Trim tank design started...')

    geometry_dict = dict_tank_design['geometry']['horizontal_stabilizer']
    section_dict = dict_tank_design['geometry']['horizontal_stabilizer']['sections']
    symmetry = dict_tank_design['geometry']['horizontal_stabilizer']['information']['symmetry']

    # Get all tank entities that are located at the horizontal stabilizer (trim tank).
    tank_entity = [k for k, v in dict_tank_design.items() if k.startswith('tank_') and k[5:].isdigit()
                   and v['tank_location'] == 'horizontal_stabilizer'][0]

    # Calculate distance from wing root to wing tip.
    span = section_dict['tip']['position']['y'] - section_dict['fuselage_center_line']['position']['y']
    # Calculate span offset for tank.
    outer_tank_span_offset = -(span*dict_ac_data['buffer_outer_tank_segment'])

    # Set tank sections according to number of wing sections.
    number_of_sections = len(section_dict)
    if number_of_sections < 2:
        runtime_output.critical('Error: Not enough horizontal stabilizer sections given. Program aborted.')
        sys.exit(1)
    elif number_of_sections ==2:
        sections = {'section_0': 'fuselage_center_line',
                    'section_1': 'tip'}
        # Span offsets for 'section_0' and 'section_1'.
        span_offset_list = [0.0, outer_tank_span_offset]
    elif number_of_sections == 3:
        sections = {'section_0': 'fuselage_center_line',
                    'section_1': 'tip'}
        # Span offsets for 'section_0' and 'section_1'.
        span_offset_list = [0.0, outer_tank_span_offset]
    else:
        runtime_output.critical('Error: Too many horizontal stabilizer sections given. '
                                + 'Program aborted.')
        sys.exit(1)

    # Prepare tank sections.
    i = 0
    for sec, template in sections.items():
        section_dict[sec] = {}
        section_dict[sec]['position'] = {}
        section_dict[sec]['position']['y'] = section_dict[template]['position']['y'] + span_offset_list[i]
        section_dict[sec]['span_index'] = int(round(section_dict[sec]['position']['y']*100, 0))
        section_dict[sec]['position']['x'] = (
            geometry_dict['interpolated_values']['x_coordinate_leading_edge'][section_dict[sec]['span_index']])
        section_dict[sec]['position']['z'] = (
            geometry_dict['interpolated_values']['z_coordinate_leading_edge'][section_dict[sec]['span_index']])
        i += 1

    """Calculate tank volumes (geometry does not change from here)."""
    # Initialize empty dictionary and lists.
    dict_tank_design[tank_entity]['geometry'] = {}
    dict_tank_design[tank_entity]['geometry']['inner_surface'] = {}
    dict_tank_design[tank_entity]['geometry']['outer_surface'] = {}
    inner_surface_str = str()
    outer_surface_str = str()

    if symmetry:
        if number_of_sections == 2:
            inner_surface_str = 'section_0'
            outer_surface_str = 'section_1'
        elif number_of_sections == 3:
            inner_surface_str = 'section_0'
            outer_surface_str = 'section_1'
    else:
        runtime_output.critical('Error: Horizontal stabilizer not symmetric! Program aborted.')
        sys.exit(1)

    # Calculate obelisk volume.
    obelisk_volume = calculate_obelisk_volume(dict_ac_data['obelisk_volume_calculation_method'],
                                              geometry_dict, inner_surface_str, outer_surface_str)
    obelisk_volume *= 2

    # Calculate actual tank volume (considering volume loss due to internal structure of integral tanks and buffer for
    # temperature-dependent expansion of fuel).
    volume_trim_tank = factor_volume_usable*temperature_expansion_allowance*obelisk_volume

    # Calculate energy and write data to 'dict_tank_design'.
    dict_tank_design[tank_entity]['volume_available'] = volume_trim_tank
    dict_tank_design[tank_entity]['energy_available'] = (
        volume_trim_tank*kerosene_volumetric_energy_density)
    if energy_demand_covered:
        dict_tank_design[tank_entity]['energy_required_for_mission'] = False
    else:
        dict_tank_design[tank_entity]['energy_required_for_mission'] = True

    # Add required output data: x, y, and z position, shape, height, width, and length.
    for surface_str, surface_name in zip([inner_surface_str, outer_surface_str], ['inner_surface', 'outer_surface']):
        surface_data = section_dict[surface_str]
        span_idx = surface_data['span_index']

        # Extract x and z coordinates.
        x_coordinate = geometry_dict['interpolated_values']['x_coordinate_front_spar'][span_idx]
        y_coordinate = surface_data['position']['y']
        z_coordinate = surface_data['position']['z']

        dict_tank_design[tank_entity]['geometry'][surface_name] = {
            'section_name': surface_str,
            'position': {
                'x': x_coordinate,
                'y': y_coordinate,
                'z': z_coordinate,
            },
            'shape': 'rectangular',
            'height': geometry_dict['interpolated_values']['tank_heights'][span_idx],
            'width': geometry_dict['interpolated_values']['tank_lengths'][span_idx],
            'length': 0.0
        }
    # Rename and restructure dict.
    dict_tank_design[tank_entity]['geometry']['cross_section_1'] = dict_tank_design[tank_entity]['geometry'].pop(
        'inner_surface')
    dict_tank_design[tank_entity]['geometry']['cross_section_2'] = dict_tank_design[tank_entity]['geometry'].pop(
        'outer_surface')

    def deep_copy_dict(d):
        if isinstance(d, dict):
            return {key: deep_copy_dict(value) for key, value in d.items()}
        elif isinstance(d, list):
            return [deep_copy_dict(item) for item in d]
        else:
            return d

    independent_copy = deep_copy_dict(dict_tank_design[tank_entity]['geometry']['cross_section_2'])
    independent_copy['position']['y'] *= -1
    dict_tank_design[tank_entity]['geometry']['cross_section_0'] = independent_copy

    """Set or calculate total tank parameters."""
    # Set tank mass to zero (integral tank).
    dict_tank_design[tank_entity]['geometry']['total'] = {}
    dict_tank_design[tank_entity]['geometry']['total']['mass'] = 0.0
    # Set inertia to zero.
    inertia_list = ['j_xx', 'j_yy', 'j_zz', 'j_xy', 'j_xz', 'j_yx', 'j_yz', 'j_zx', 'j_zy']
    dict_tank_design[tank_entity]['geometry']['total']['inertia'] = {}
    for inertia in inertia_list:
        dict_tank_design[tank_entity]['geometry']['total']['inertia'][inertia] = 0.0
    # Center of gravity equals zero since integral tank has no mass.
    dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity'] = {}
    dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity']['x'] = 0.0
    dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity']['y'] = 0.0
    dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity']['z'] = 0.0
    # Calculate position of tank entity (foremost point of tank).
    distance_wing_to_horizontal_stabilizer = (dict_tank_design['geometry']['horizontal_stabilizer']['position']['x']
                                              - dict_tank_design['geometry']['wing']['position']['x'])
    wing_lower_than_horizontal_stabilizer = (dict_tank_design['geometry']['wing']['position']['z']
                                             < dict_tank_design['geometry']['horizontal_stabilizer']['position']['z'])
    z_offset_wing_to_horizontal_stabilizer = abs(dict_tank_design['geometry']['horizontal_stabilizer']['position']['z']
                                              - dict_tank_design['geometry']['wing']['position']['z'])
    if wing_lower_than_horizontal_stabilizer:
        relative_z_position_horizontal_stabilizer_to_wing = z_offset_wing_to_horizontal_stabilizer
    else:
        relative_z_position_horizontal_stabilizer_to_wing = -z_offset_wing_to_horizontal_stabilizer
    dict_tank_design[tank_entity]['geometry']['total']['position'] = {}
    dict_tank_design[tank_entity]['geometry']['total']['position']['x'] = (
        distance_wing_to_horizontal_stabilizer
        + geometry_dict['interpolated_values']['x_coordinate_front_spar'][section_dict[inner_surface_str]
                                                                          ['span_index']])
    dict_tank_design[tank_entity]['geometry']['total']['position']['y'] = (
        section_dict[inner_surface_str]['position']['y'])
    dict_tank_design[tank_entity]['geometry']['total']['position']['z'] = (
        relative_z_position_horizontal_stabilizer_to_wing)
    # Direction.
    dict_tank_design[tank_entity]['geometry']['total']['direction'] = {}
    dict_tank_design[tank_entity]['geometry']['total']['direction']['x'] = 0
    dict_tank_design[tank_entity]['geometry']['total']['direction']['y'] = 1
    dict_tank_design[tank_entity]['geometry']['total']['direction']['z'] = 0
    
    # Centroid.
    # Assumption: inner surface area always greater than outer surface area.
    inner_surface = dict_tank_design[tank_entity]['geometry']['cross_section_1']
    outer_surface = dict_tank_design[tank_entity]['geometry']['cross_section_2']
    inner_surface_area = (inner_surface['height']
                            *inner_surface['width'])
    outer_surface_area = (outer_surface['height']
                    *outer_surface['width'])
    y_position_outer_surface = outer_surface['position']['y']
    y_position_inner_surface = inner_surface['position']['y']
    tank_span = y_position_outer_surface - y_position_inner_surface
    y_position_center_of_mass = 0
    centroid_y = 0
    x_position_cog_inner_surface = inner_surface['width']/2
    x_position_cog_outer_surface = outer_surface['width']/2
    z_position_cog_inner_surface = inner_surface['height']/2
    z_position_cog_outer_surface = outer_surface['height']/2
    x_position_center_of_mass = ((inner_surface_area*x_position_cog_inner_surface
                                    + outer_surface_area*x_position_cog_outer_surface)
                                    /(inner_surface_area + outer_surface_area))
    z_position_center_of_mass = ((inner_surface_area*z_position_cog_inner_surface
                                    + outer_surface_area*z_position_cog_outer_surface)
                                    /(inner_surface_area + outer_surface_area))

    dict_tank_design[tank_entity]['geometry']['total']['centroid'] = {}
    dict_tank_design[tank_entity]['geometry']['total']['centroid']['x'] = x_position_center_of_mass
    dict_tank_design[tank_entity]['geometry']['total']['centroid']['y'] = y_position_center_of_mass
    dict_tank_design[tank_entity]['geometry']['total']['centroid']['z'] = z_position_center_of_mass

    # Transform volume to liter for print.
    volume_trim_tank_l = volume_trim_tank*1e3
    # Print.
    runtime_output.print('Trim tank (' + tank_entity + ') calculated. '
                         + 'Volume (energy) available: ' + f"{volume_trim_tank_l:,.2f}" + ' L ('
                         + f"{dict_tank_design[tank_entity]['energy_available']/1e6:,.2f}" + ' MJ)')

    # Prints.
    runtime_output.print('Trim tank design completed.')
    runtime_output.debug('Debug: The "calculate_trim_tank" function was successfully executed.')

    # Prepare data output.
    dict_tank_design['tmp_energies'] = {}
    dict_tank_design['tmp_energies']['trim_tank'] = float(dict_tank_design[tank_entity]['energy_available'])

    return dict_tank_design
