"""Module providing calculation functions provided by the user."""
# Import standard modules.
import sys

# Import own modules.
from src.tube_and_wing.empirical.tank_design_tu_berlin.general.tankdesigntuberlin import tank_design_tu_berlin
from src.tube_and_wing.empirical.tank_design_tu_berlin.kerosene.calculatetanks import calculate_kerosene_tanks


def method_kerosene(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """Tank sizing method for kerosene-powered aircraft.

    This function performs the tank sizing for kerosene-powered aircraft configurations.
    The following steps are performed:

    1. Data Merging: The `dict_ac_exchange` and `dict_mod_config` dictionaries are merged into a single dictionary,
    `dict_complete_data`, which holds all relevant design and configuration parameters.
    2. Data Extraction: General parameters are extracted from `dict_complete_data` and stored in `dict_general_data`. 
    Parameters related to the design mission are filtered and saved in `dict_design_data`.
    3. Validation of Design Data: The design data is checked for completeness. If any key parameters are missing (i.e.,
    `None`), the tank design process is deemed impossible, and the program is aborted.
    4. Fuselage Check: The number of fuselages is determined from the design data. If no fuselage or more than one 
    fuselage is found, the tank design cannot proceed, and the program is aborted.
    5. Tank Design Preparation: The general and design parameters are merged into `dict_design`, and a new tank design
    is prepared using the `tank_design_tu_berlin` function. If the tank definition is invalid, the program is
    terminated.
    6. Tank Calculation: Kerosene tank volumes and configurations are calculated using the `calculate_kerosene_tanks` 
    function, that takes the design parameters and performs necessary computations.
    7. Debugging and Final Output: A debug message is logged indicating successful execution, and a final message
    confirms that the tank design has been completed.

    The output dictionary 'kerosene_output_dict' contains the results of the tank design and is structured
    according to the following scheme:
        kerosene_output_dict = {
            'tank_0': {...},
            'tank_1': {...},
            ...,
            'number_of_tanks': (),
            'tank_configuration': ()}

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict dict_ac_exchange: Dict containing parameters and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameters and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict kerosene_output_dict: Dictionary containing results from calculation for kerosene-powered aircraft
    """

    """ Preparation. """
    # Merge both data dictionaries.
    dict_design = dict_ac_exchange | dict_mod_config
    dict_design['energy_carrier'] = routing_dict['user_layer']

    """Calculate tanks."""
    # Tank design preparation.
    kerosene_output_dict = tank_design_tu_berlin(paths_and_names, routing_dict, dict_design, runtime_output)
    # Tank calculation.
    kerosene_output_dict = calculate_kerosene_tanks(dict_design, kerosene_output_dict, runtime_output)

    # Debug print.
    runtime_output.debug('Debug: The "method_kerosene" function was successfully executed.')
    runtime_output.print('Tank design completed.')

    return kerosene_output_dict
