/**
 * \file aerodynamicEfficiencyMethods.cpp
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief Aerodynamic efficiency methods - oswald factor e
 * \version 0.1
 * \date 2023-12-07
 *
 * @copyright Copyright (c) 2023
 *
 */
#ifndef AERODYNAMIC_EFFICIENCY_METHODS_H_
#define AERODYNAMIC_EFFICIENCY_METHODS_H_

#include <cmath>
// Methods from snorri gudmunsson - general aviaition aircraft design - basic equations applicable -> see chapter "The
// Anatomy of the Wing"
namespace brandt {
double aerodynamic_efficiency(const double aspect_ratio_corr, const double sweep_angle_maximum_wing_thickness_line) {
  return (2. - aspect_ratio_corr +
          sqrt((4 + pow(aspect_ratio_corr, 2.0) * (1 + pow(tan(sweep_angle_maximum_wing_thickness_line), 2)))));
}
}  // namespace brandt
namespace howe {
double aerodynamic_efficiency(const double aspect_ratio_corr, const double thickness_to_chord_ratio,
                              const double number_of_engines, const double quarter_chord_sweep,
                              const double taper_ratio, const double mach_number_free_stream) {
  if (aspect_ratio_corr < 5. || mach_number_free_stream > 0.95) {
    throw("Error")
  }
  const double taper_ratio_factor = 0.005 * (1 + 1.5 * pow(taper_ratio - 0.6, 2.));
  return 1 /
         ((1 + 0.12 * mach_number_free_stream) *
          (1 + 0.142 + taper_ratio_factor * aspect_ratio_corr * pow(10 * thickness_to_chord_ratio, 0.33)) /
          (pow(cos(quarter_chord_sweep), 2.) + 0.1 * (3 * number_of_engines + 1) / (pow(4 + aspect_ratio_corr, 0.8))));
}
}  // namespace howe

#endif  // AERODYNAMIC_EFFICIENCY_METHODS_H_
